/* Generated by Cython 0.25.2 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__catalyst___protocol
#define __PYX_HAVE_API__catalyst___protocol
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "catalyst/_protocol.pyx",
  "python2/lib/python2.7/site-packages/Cython/Includes/cpython/type.pxd",
  "python2/lib/python2.7/site-packages/Cython/Includes/cpython/bool.pxd",
  "python2/lib/python2.7/site-packages/Cython/Includes/cpython/complex.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_8catalyst_9_protocol_check_parameters;
struct __pyx_obj_8catalyst_9_protocol_BarData;
struct __pyx_obj_8catalyst_9_protocol_SidView;
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__;
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes;
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__;
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems;

/* "catalyst/_protocol.pyx":50
 * 
 * 
 * cdef class check_parameters(object):             # <<<<<<<<<<<<<<
 *     """
 *     Asserts that the keywords passed into the wrapped function are included
 */
struct __pyx_obj_8catalyst_9_protocol_check_parameters {
  PyObject_HEAD
  PyObject *keyword_names;
  PyObject *types;
  PyObject *keys_to_types;
};


/* "catalyst/_protocol.pyx":134
 * 
 * 
 * cdef class BarData:             # <<<<<<<<<<<<<<
 *     """
 *     Provides methods to access spot value or history windows of price data.
 */
struct __pyx_obj_8catalyst_9_protocol_BarData {
  PyObject_HEAD
  struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *__pyx_vtab;
  PyObject *data_portal;
  PyObject *simulation_dt_func;
  PyObject *data_frequency;
  PyObject *restrictions;
  PyObject *_views;
  PyObject *_universe_func;
  PyObject *_last_calculated_universe;
  PyObject *_universe_last_updated_at;
  PyBoolObject *_daily_mode;
  PyObject *_trading_calendar;
  PyObject *_is_restricted;
  PyBoolObject *_adjust_minutes;
};


/* "catalyst/_protocol.pyx":826
 *         )
 * 
 * cdef class SidView:             # <<<<<<<<<<<<<<
 *     cdef object asset
 *     cdef object data_portal
 */
struct __pyx_obj_8catalyst_9_protocol_SidView {
  PyObject_HEAD
  struct __pyx_vtabstruct_8catalyst_9_protocol_SidView *__pyx_vtab;
  PyObject *asset;
  PyObject *data_portal;
  PyObject *simulation_dt_func;
  PyObject *data_frequency;
};


/* "catalyst/_protocol.pyx":70
 *         self.keys_to_types = dict(zip(keyword_names, types))
 * 
 *     def __call__(self, func):             # <<<<<<<<<<<<<<
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):
 */
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ {
  PyObject_HEAD
  PyObject *__pyx_v_func;
  struct __pyx_obj_8catalyst_9_protocol_check_parameters *__pyx_v_self;
};


/* "catalyst/_protocol.pyx":126
 * 
 * @contextmanager
 * def handle_non_market_minutes(bar_data):             # <<<<<<<<<<<<<<
 *     try:
 *         bar_data._handle_non_market_minutes = True
 */
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes {
  PyObject_HEAD
  PyObject *__pyx_v_bar_data;
};


/* "catalyst/_protocol.pyx":778
 *         return self._last_calculated_universe
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ {
  PyObject_HEAD
  PyObject *__pyx_v_asset;
  struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "catalyst/_protocol.pyx":795
 *         return [(asset, self[asset]) for asset in self._calculate_universe()]
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */
struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems {
  PyObject_HEAD
  PyObject *__pyx_v_asset;
  struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};



/* "catalyst/_protocol.pyx":134
 * 
 * 
 * cdef class BarData:             # <<<<<<<<<<<<<<
 *     """
 *     Provides methods to access spot value or history windows of price data.
 */

struct __pyx_vtabstruct_8catalyst_9_protocol_BarData {
  PyObject *(*_get_equity_price_view)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *);
  PyObject *(*_create_sid_view)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *);
  PyObject *(*_get_current_minute)(struct __pyx_obj_8catalyst_9_protocol_BarData *);
  PyBoolObject *(*_can_trade_for_asset)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *, PyObject *, PyObject *, PyObject *);
  PyBoolObject *(*_is_stale_for_asset)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_calculate_universe)(struct __pyx_obj_8catalyst_9_protocol_BarData *);
  PyObject *(*_warn_deprecated)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *);
};
static struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *__pyx_vtabptr_8catalyst_9_protocol_BarData;


/* "catalyst/_protocol.pyx":826
 *         )
 * 
 * cdef class SidView:             # <<<<<<<<<<<<<<
 *     cdef object asset
 *     cdef object data_portal
 */

struct __pyx_vtabstruct_8catalyst_9_protocol_SidView {
  PyObject *(*_warn_deprecated)(struct __pyx_obj_8catalyst_9_protocol_SidView *, PyObject *);
};
static struct __pyx_vtabstruct_8catalyst_9_protocol_SidView *__pyx_vtabptr_8catalyst_9_protocol_SidView;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
static CYTHON_INLINE int __Pyx_unpack_tuple2(PyObject* tuple, PyObject** value1, PyObject** value2,
                                             int is_tuple, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#else
    #define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetNameInClass.proto */
static PyObject *__Pyx_GetNameInClass(PyObject *nmspace, PyObject *name);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_short(short value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE short __Pyx_PyInt_As_short(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyObject *);
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static int __Pyx_Coroutine_clear(PyObject *self);
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue);
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue) PyGen_FetchStopIterationValue(pvalue)
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__get_equity_price_view(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset); /* proto*/
static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__create_sid_view(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset); /* proto*/
static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__get_current_minute(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto*/
static PyBoolObject *__pyx_f_8catalyst_9_protocol_7BarData__can_trade_for_asset(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset, PyObject *__pyx_v_dt, PyObject *__pyx_v_adjusted_dt, PyObject *__pyx_v_data_portal); /* proto*/
static PyBoolObject *__pyx_f_8catalyst_9_protocol_7BarData__is_stale_for_asset(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset, PyObject *__pyx_v_dt, PyObject *__pyx_v_adjusted_dt, PyObject *__pyx_v_data_portal); /* proto*/
static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__calculate_universe(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__warn_deprecated(CYTHON_UNUSED struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_msg); /* proto*/
static PyObject *__pyx_f_8catalyst_9_protocol_7SidView__warn_deprecated(CYTHON_UNUSED struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_msg); /* proto*/

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'catalyst._protocol' */
static PyTypeObject *__pyx_ptype_8catalyst_9_protocol_check_parameters = 0;
static PyTypeObject *__pyx_ptype_8catalyst_9_protocol_BarData = 0;
static PyTypeObject *__pyx_ptype_8catalyst_9_protocol_SidView = 0;
static PyTypeObject *__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct____call__ = 0;
static PyTypeObject *__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes = 0;
static PyTypeObject *__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_2___iter__ = 0;
static PyTypeObject *__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_3_iteritems = 0;
static PyBoolObject *__pyx_f_8catalyst_9_protocol__is_iterable(PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "catalyst._protocol"
int __pyx_module_is_main_catalyst___protocol = 0;

/* Implementation of 'catalyst._protocol' */
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_f[] = "f";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_i[] = "i";
static const char __pyx_k__3[] = ", ";
static const char __pyx_k__4[] = "";
static const char __pyx_k_dt[] = "dt";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_pd[] = "pd";
static const char __pyx_k_NaT[] = "NaT";
static const char __pyx_k_PY2[] = "PY2";
static const char __pyx_k_arg[] = "arg";
static const char __pyx_k_dec[] = "dec";
static const char __pyx_k_sid[] = "sid";
static const char __pyx_k_six[] = "six";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bars[] = "bars";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_func[] = "func";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mavg[] = "mavg";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_next[] = "next";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type_";
static const char __pyx_k_vwap[] = "vwap";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_Asset[] = "Asset";
static const char __pyx_k_Panel[] = "Panel";
static const char __pyx_k_asset[] = "asset";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_daily[] = "daily";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_field[] = "field";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_isnan[] = "isnan";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_price[] = "price";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_wraps[] = "wraps";
static const char __pyx_k_Future[] = "Future";
static const char __pyx_k_Series[] = "Series";
static const char __pyx_k_assets[] = "assets";
static const char __pyx_k_fields[] = "fields";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_pandas[] = "pandas";
static const char __pyx_k_stddev[] = "stddev";
static const char __pyx_k_symbol[] = "symbol";
static const char __pyx_k_volume[] = "volume";
static const char __pyx_k_current[] = "current";
static const char __pyx_k_getattr[] = "__getattr__";
static const char __pyx_k_history[] = "history";
static const char __pyx_k_keyword[] = "keyword";
static const char __pyx_k_returns[] = "returns";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_bar_data[] = "bar_data";
static const char __pyx_k_category[] = "category";
static const char __pyx_k_contains[] = "contains";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_is_stale[] = "is_stale";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_DataFrame[] = "DataFrame";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_bar_count[] = "bar_count";
static const char __pyx_k_can_trade[] = "can_trade";
static const char __pyx_k_direction[] = "direction";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_frequency[] = "frequency";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_next_open[] = "next_open";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_contextlib[] = "contextlib";
static const char __pyx_k_open_price[] = "open_price";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_stacklevel[] = "stacklevel";
static const char __pyx_k_close_price[] = "close_price";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_data_portal[] = "data_portal";
static const char __pyx_k_last_traded[] = "last_traded";
static const char __pyx_k_asset_finder[] = "asset_finder";
static const char __pyx_k_no_wraps_py2[] = "no_wraps_py2";
static const char __pyx_k_pandas_tslib[] = "pandas.tslib";
static const char __pyx_k_restrictions[] = "restrictions";
static const char __pyx_k_string_types[] = "string_types";
static const char __pyx_k_current_chain[] = "current_chain";
static const char __pyx_k_expected_type[] = "expected_type";
static const char __pyx_k_is_restricted[] = "is_restricted";
static const char __pyx_k_keyword_names[] = "keyword_names";
static const char __pyx_k_universe_func[] = "universe_func";
static const char __pyx_k_BarData___iter[] = "BarData.__iter__";
static const char __pyx_k_contextmanager[] = "contextmanager";
static const char __pyx_k_data_frequency[] = "data_frequency";
static const char __pyx_k_get_spot_value[] = "get_spot_value";
static const char __pyx_k_normalize_date[] = "normalize_date";
static const char __pyx_k_retrieve_asset[] = "retrieve_asset";
static const char __pyx_k_auto_close_date[] = "auto_close_date";
static const char __pyx_k_catalyst_assets[] = "catalyst.assets";
static const char __pyx_k_current_session[] = "current_session";
static const char __pyx_k_get_adjustments[] = "get_adjustments";
static const char __pyx_k_previous_minute[] = "previous_minute";
static const char __pyx_k_AssetConvertible[] = "AssetConvertible";
static const char __pyx_k_ContinuousFuture[] = "ContinuousFuture";
static const char __pyx_k_is_exchange_open[] = "is_exchange_open";
static const char __pyx_k_trading_calendar[] = "trading_calendar";
static const char __pyx_k_BarData_iteritems[] = "BarData.iteritems";
static const char __pyx_k_continuous_future[] = "continuous_future";
static const char __pyx_k_is_open_on_minute[] = "is_open_on_minute";
static const char __pyx_k_catalyst__protocol[] = "catalyst._protocol";
static const char __pyx_k_expected_type_name[] = "expected_type_name";
static const char __pyx_k_get_adjusted_value[] = "get_adjusted_value";
static const char __pyx_k_get_fetcher_assets[] = "get_fetcher_assets";
static const char __pyx_k_get_history_window[] = "get_history_window";
static const char __pyx_k_get_last_traded_dt[] = "get_last_traded_dt";
static const char __pyx_k_simulation_dt_func[] = "simulation_dt_func";
static const char __pyx_k_minutes_for_session[] = "minutes_for_session";
static const char __pyx_k_or_iterable_of_type[] = "or iterable of type ";
static const char __pyx_k_get_simple_transform[] = "get_simple_transform";
static const char __pyx_k_is_alive_for_session[] = "is_alive_for_session";
static const char __pyx_k_PricingDataAssociable[] = "PricingDataAssociable";
static const char __pyx_k_minute_to_session_label[] = "minute_to_session_label";
static const char __pyx_k_no_wraps_py2_locals_dec[] = "no_wraps_py2.<locals>.dec";
static const char __pyx_k_assert_keywords_and_call[] = "assert_keywords_and_call";
static const char __pyx_k_get_current_future_chain[] = "get_current_future_chain";
static const char __pyx_k_ZiplineDeprecationWarning[] = "ZiplineDeprecationWarning";
static const char __pyx_k_handle_non_market_minutes[] = "handle_non_market_minutes";
static const char __pyx_k_catalyst_catalyst_warnings[] = "catalyst.catalyst_warnings";
static const char __pyx_k_handle_non_market_minutes_2[] = "_handle_non_market_minutes";
static const char __pyx_k_The_mavg_method_is_deprecated[] = "The `mavg` method is deprecated.";
static const char __pyx_k_The_vwap_method_is_deprecated[] = "The `vwap` method is deprecated.";
static const char __pyx_k_call___locals_assert_keywords[] = "__call__.<locals>.assert_keywords_and_call";
static const char __pyx_k_Checking_whether_an_asset_is_in[] = "Checking whether an asset is in data is deprecated.";
static const char __pyx_k_The_stddev_method_is_deprecated[] = "The `stddev` method is deprecated.";
static const char __pyx_k_Users_fredfortier_Code_enigma_c[] = "/Users/fredfortier/Code/enigma/catalyst/catalyst/_protocol.pyx";
static const char __pyx_k_data_sid_N_is_deprecated_Use_da[] = "`data[sid(N)]` is deprecated. Use `data.current`.";
static const char __pyx_k_s_got_an_unexpected_keyword_arg[] = "%s() got an unexpected keyword argument '%s'";
static const char __pyx_k_Expected_s_argument_to_be_of_typ[] = "Expected %s argument to be of type %s%s";
static const char __pyx_k_Iterating_over_the_assets_in_dat[] = "Iterating over the assets in `data` is deprecated.";
static const char __pyx_k_The_returns_method_is_deprecated[] = "The `returns` method is deprecated.";
static const char __pyx_k_catalyst_assets_continuous_futur[] = "catalyst.assets.continuous_futures";
static PyObject *__pyx_n_s_Asset;
static PyObject *__pyx_n_s_AssetConvertible;
static PyObject *__pyx_n_s_BarData___iter;
static PyObject *__pyx_n_s_BarData_iteritems;
static PyObject *__pyx_kp_s_Checking_whether_an_asset_is_in;
static PyObject *__pyx_n_s_ContinuousFuture;
static PyObject *__pyx_n_s_DataFrame;
static PyObject *__pyx_kp_s_Expected_s_argument_to_be_of_typ;
static PyObject *__pyx_n_s_Future;
static PyObject *__pyx_n_s_Iterable;
static PyObject *__pyx_kp_s_Iterating_over_the_assets_in_dat;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_NaT;
static PyObject *__pyx_n_s_PY2;
static PyObject *__pyx_n_s_Panel;
static PyObject *__pyx_n_s_PricingDataAssociable;
static PyObject *__pyx_n_s_Series;
static PyObject *__pyx_kp_s_The_mavg_method_is_deprecated;
static PyObject *__pyx_kp_s_The_returns_method_is_deprecated;
static PyObject *__pyx_kp_s_The_stddev_method_is_deprecated;
static PyObject *__pyx_kp_s_The_vwap_method_is_deprecated;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Users_fredfortier_Code_enigma_c;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_ZiplineDeprecationWarning;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_n_s_arg;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_assert_keywords_and_call;
static PyObject *__pyx_n_s_asset;
static PyObject *__pyx_n_s_asset_finder;
static PyObject *__pyx_n_s_assets;
static PyObject *__pyx_n_s_auto_close_date;
static PyObject *__pyx_n_s_bar_count;
static PyObject *__pyx_n_s_bar_data;
static PyObject *__pyx_n_s_bars;
static PyObject *__pyx_n_s_call___locals_assert_keywords;
static PyObject *__pyx_n_s_can_trade;
static PyObject *__pyx_n_s_catalyst__protocol;
static PyObject *__pyx_n_s_catalyst_assets;
static PyObject *__pyx_n_s_catalyst_assets_continuous_futur;
static PyObject *__pyx_n_s_catalyst_catalyst_warnings;
static PyObject *__pyx_n_s_category;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_close_price;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_contains;
static PyObject *__pyx_n_s_contextlib;
static PyObject *__pyx_n_s_contextmanager;
static PyObject *__pyx_n_s_continuous_future;
static PyObject *__pyx_n_s_current;
static PyObject *__pyx_n_s_current_chain;
static PyObject *__pyx_n_s_current_session;
static PyObject *__pyx_n_s_daily;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_frequency;
static PyObject *__pyx_n_s_data_portal;
static PyObject *__pyx_kp_s_data_sid_N_is_deprecated_Use_da;
static PyObject *__pyx_n_s_datetime;
static PyObject *__pyx_n_s_dec;
static PyObject *__pyx_n_s_direction;
static PyObject *__pyx_n_s_dt;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_expected_type;
static PyObject *__pyx_n_s_expected_type_name;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_field;
static PyObject *__pyx_n_s_fields;
static PyObject *__pyx_n_s_frequency;
static PyObject *__pyx_n_s_func;
static PyObject *__pyx_n_s_functools;
static PyObject *__pyx_n_s_g;
static PyObject *__pyx_n_s_get_adjusted_value;
static PyObject *__pyx_n_s_get_adjustments;
static PyObject *__pyx_n_s_get_current_future_chain;
static PyObject *__pyx_n_s_get_fetcher_assets;
static PyObject *__pyx_n_s_get_history_window;
static PyObject *__pyx_n_s_get_last_traded_dt;
static PyObject *__pyx_n_s_get_simple_transform;
static PyObject *__pyx_n_s_get_spot_value;
static PyObject *__pyx_n_s_getattr;
static PyObject *__pyx_n_s_handle_non_market_minutes;
static PyObject *__pyx_n_s_handle_non_market_minutes_2;
static PyObject *__pyx_n_s_history;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_is_alive_for_session;
static PyObject *__pyx_n_s_is_exchange_open;
static PyObject *__pyx_n_s_is_open_on_minute;
static PyObject *__pyx_n_s_is_restricted;
static PyObject *__pyx_n_s_is_stale;
static PyObject *__pyx_n_s_isnan;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_iteritems;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_keyword;
static PyObject *__pyx_n_s_keyword_names;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_last_traded;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mavg;
static PyObject *__pyx_n_s_minute_to_session_label;
static PyObject *__pyx_n_s_minutes_for_session;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_next;
static PyObject *__pyx_n_s_next_open;
static PyObject *__pyx_n_s_no_wraps_py2;
static PyObject *__pyx_n_s_no_wraps_py2_locals_dec;
static PyObject *__pyx_n_s_normalize_date;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_open_price;
static PyObject *__pyx_kp_s_or_iterable_of_type;
static PyObject *__pyx_n_s_pandas;
static PyObject *__pyx_n_s_pandas_tslib;
static PyObject *__pyx_n_s_pd;
static PyObject *__pyx_n_s_previous_minute;
static PyObject *__pyx_n_s_price;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_restrictions;
static PyObject *__pyx_n_s_retrieve_asset;
static PyObject *__pyx_n_s_returns;
static PyObject *__pyx_kp_s_s_got_an_unexpected_keyword_arg;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_sid;
static PyObject *__pyx_n_s_simulation_dt_func;
static PyObject *__pyx_n_s_six;
static PyObject *__pyx_n_s_stacklevel;
static PyObject *__pyx_n_s_stddev;
static PyObject *__pyx_n_s_string_types;
static PyObject *__pyx_n_s_symbol;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_trading_calendar;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_types;
static PyObject *__pyx_n_s_universe_func;
static PyObject *__pyx_n_s_volume;
static PyObject *__pyx_n_s_vwap;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_s_wraps;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_8catalyst_9_protocol_12no_wraps_py2_dec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_g); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_no_wraps_py2(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_f); /* proto */
static int __pyx_pf_8catalyst_9_protocol_16check_parameters___init__(struct __pyx_obj_8catalyst_9_protocol_check_parameters *__pyx_v_self, PyObject *__pyx_v_keyword_names, PyObject *__pyx_v_types); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_16check_parameters_8__call___assert_keywords_and_call(PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_16check_parameters_2__call__(struct __pyx_obj_8catalyst_9_protocol_check_parameters *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_2handle_non_market_minutes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bar_data); /* proto */
static int __pyx_pf_8catalyst_9_protocol_7BarData___init__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_data_portal, PyObject *__pyx_v_simulation_dt_func, PyObject *__pyx_v_data_frequency, PyObject *__pyx_v_trading_calendar, PyObject *__pyx_v_restrictions, PyObject *__pyx_v_universe_func); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_2current(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets, PyObject *__pyx_v_fields); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_4current_chain(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_continuous_future); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_6can_trade(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_8is_stale(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_10history(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets, PyObject *__pyx_v_fields, PyObject *__pyx_v_bar_count, PyObject *__pyx_v_frequency); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_10current_dt___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_14fetcher_assets___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static int __pyx_pf_8catalyst_9_protocol_7BarData_26_handle_non_market_minutes___set__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_15current_session___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_23current_session_minutes___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_12__iter__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static int __pyx_pf_8catalyst_9_protocol_7BarData_15__contains__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_17items(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_19iteritems(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_8catalyst_9_protocol_7BarData_22__len__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_24keys(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_26iterkeys(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_28__getitem__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static int __pyx_pf_8catalyst_9_protocol_7SidView___init__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_asset, PyObject *__pyx_v_data_portal, PyObject *__pyx_v_simulation_dt_func, PyObject *__pyx_v_data_frequency); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_2__getattr__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_column); /* proto */
static int __pyx_pf_8catalyst_9_protocol_7SidView_4__contains__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_column); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_6__getitem__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_column); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_3sid___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_2dt___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_8datetime___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_10current_dt___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_8mavg(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_num_minutes); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_10stddev(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_num_minutes); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_12vwap(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_num_minutes); /* proto */
static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_14returns(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self); /* proto */
static PyObject *__pyx_tp_new_8catalyst_9_protocol_check_parameters(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8catalyst_9_protocol_BarData(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8catalyst_9_protocol_SidView(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct____call__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_2___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_3_iteritems(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_codeobj__2;
static PyObject *__pyx_codeobj__6;
static PyObject *__pyx_codeobj__8;
static PyObject *__pyx_codeobj__10;

/* "catalyst/_protocol.pyx":35
 * 
 * 
 * cdef bool _is_iterable(obj):             # <<<<<<<<<<<<<<
 *     return isinstance(obj, Iterable) and not isinstance(obj, string_types)
 * 
 */

static PyBoolObject *__pyx_f_8catalyst_9_protocol__is_iterable(PyObject *__pyx_v_obj) {
  PyBoolObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("_is_iterable", 0);

  /* "catalyst/_protocol.pyx":36
 * 
 * cdef bool _is_iterable(obj):
 *     return isinstance(obj, Iterable) and not isinstance(obj, string_types)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_Iterable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_2); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7cpython_4bool_bool)))) __PYX_ERR(0, 36, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_string_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_2); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (!(__pyx_t_3 != 0));
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7cpython_4bool_bool)))) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = ((PyBoolObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":35
 * 
 * 
 * cdef bool _is_iterable(obj):             # <<<<<<<<<<<<<<
 *     return isinstance(obj, Iterable) and not isinstance(obj, string_types)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("catalyst._protocol._is_iterable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":42
 * # with python3 so it is not a big deal.
 * if PY2:
 *     def no_wraps_py2(f):             # <<<<<<<<<<<<<<
 *         def dec(g):
 *             return g
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_1no_wraps_py2(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static PyMethodDef __pyx_mdef_8catalyst_9_protocol_1no_wraps_py2 = {"no_wraps_py2", (PyCFunction)__pyx_pw_8catalyst_9_protocol_1no_wraps_py2, METH_O, 0};
static PyObject *__pyx_pw_8catalyst_9_protocol_1no_wraps_py2(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("no_wraps_py2 (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_no_wraps_py2(__pyx_self, ((PyObject *)__pyx_v_f));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":43
 * if PY2:
 *     def no_wraps_py2(f):
 *         def dec(g):             # <<<<<<<<<<<<<<
 *             return g
 *         return dec
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_12no_wraps_py2_1dec(PyObject *__pyx_self, PyObject *__pyx_v_g); /*proto*/
static PyMethodDef __pyx_mdef_8catalyst_9_protocol_12no_wraps_py2_1dec = {"dec", (PyCFunction)__pyx_pw_8catalyst_9_protocol_12no_wraps_py2_1dec, METH_O, 0};
static PyObject *__pyx_pw_8catalyst_9_protocol_12no_wraps_py2_1dec(PyObject *__pyx_self, PyObject *__pyx_v_g) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dec (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_12no_wraps_py2_dec(__pyx_self, ((PyObject *)__pyx_v_g));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_12no_wraps_py2_dec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_g) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dec", 0);

  /* "catalyst/_protocol.pyx":44
 *     def no_wraps_py2(f):
 *         def dec(g):
 *             return g             # <<<<<<<<<<<<<<
 *         return dec
 * else:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_g);
  __pyx_r = __pyx_v_g;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":43
 * if PY2:
 *     def no_wraps_py2(f):
 *         def dec(g):             # <<<<<<<<<<<<<<
 *             return g
 *         return dec
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":42
 * # with python3 so it is not a big deal.
 * if PY2:
 *     def no_wraps_py2(f):             # <<<<<<<<<<<<<<
 *         def dec(g):
 *             return g
 */

static PyObject *__pyx_pf_8catalyst_9_protocol_no_wraps_py2(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_f) {
  PyObject *__pyx_v_dec = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("no_wraps_py2", 0);

  /* "catalyst/_protocol.pyx":43
 * if PY2:
 *     def no_wraps_py2(f):
 *         def dec(g):             # <<<<<<<<<<<<<<
 *             return g
 *         return dec
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8catalyst_9_protocol_12no_wraps_py2_1dec, 0, __pyx_n_s_no_wraps_py2_locals_dec, NULL, __pyx_n_s_catalyst__protocol, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dec = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":45
 *         def dec(g):
 *             return g
 *         return dec             # <<<<<<<<<<<<<<
 * else:
 *     no_wraps_py2 = wraps
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dec);
  __pyx_r = __pyx_v_dec;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":42
 * # with python3 so it is not a big deal.
 * if PY2:
 *     def no_wraps_py2(f):             # <<<<<<<<<<<<<<
 *         def dec(g):
 *             return g
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("catalyst._protocol.no_wraps_py2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":64
 *     cdef dict keys_to_types
 * 
 *     def __init__(self, keyword_names, types):             # <<<<<<<<<<<<<<
 *         self.keyword_names = keyword_names
 *         self.types = types
 */

/* Python wrapper */
static int __pyx_pw_8catalyst_9_protocol_16check_parameters_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8catalyst_9_protocol_16check_parameters_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_keyword_names = 0;
  PyObject *__pyx_v_types = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_keyword_names,&__pyx_n_s_types,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_keyword_names)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_types)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 64, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 64, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_keyword_names = values[0];
    __pyx_v_types = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 64, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("catalyst._protocol.check_parameters.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8catalyst_9_protocol_16check_parameters___init__(((struct __pyx_obj_8catalyst_9_protocol_check_parameters *)__pyx_v_self), __pyx_v_keyword_names, __pyx_v_types);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8catalyst_9_protocol_16check_parameters___init__(struct __pyx_obj_8catalyst_9_protocol_check_parameters *__pyx_v_self, PyObject *__pyx_v_keyword_names, PyObject *__pyx_v_types) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "catalyst/_protocol.pyx":65
 * 
 *     def __init__(self, keyword_names, types):
 *         self.keyword_names = keyword_names             # <<<<<<<<<<<<<<
 *         self.types = types
 * 
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_keyword_names))||((__pyx_v_keyword_names) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_keyword_names)->tp_name), 0))) __PYX_ERR(0, 65, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_keyword_names;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->keyword_names);
  __Pyx_DECREF(__pyx_v_self->keyword_names);
  __pyx_v_self->keyword_names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":66
 *     def __init__(self, keyword_names, types):
 *         self.keyword_names = keyword_names
 *         self.types = types             # <<<<<<<<<<<<<<
 * 
 *         self.keys_to_types = dict(zip(keyword_names, types))
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_types))||((__pyx_v_types) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_types)->tp_name), 0))) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_types;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->types);
  __Pyx_DECREF(__pyx_v_self->types);
  __pyx_v_self->types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":68
 *         self.types = types
 * 
 *         self.keys_to_types = dict(zip(keyword_names, types))             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, func):
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_keyword_names);
  __Pyx_GIVEREF(__pyx_v_keyword_names);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_keyword_names);
  __Pyx_INCREF(__pyx_v_types);
  __Pyx_GIVEREF(__pyx_v_types);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_types);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyDict_Type)), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->keys_to_types);
  __Pyx_DECREF(__pyx_v_self->keys_to_types);
  __pyx_v_self->keys_to_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":64
 *     cdef dict keys_to_types
 * 
 *     def __init__(self, keyword_names, types):             # <<<<<<<<<<<<<<
 *         self.keyword_names = keyword_names
 *         self.types = types
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("catalyst._protocol.check_parameters.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":70
 *         self.keys_to_types = dict(zip(keyword_names, types))
 * 
 *     def __call__(self, func):             # <<<<<<<<<<<<<<
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_16check_parameters_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_16check_parameters_3__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_func)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(0, 70, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 70, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("catalyst._protocol.check_parameters.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8catalyst_9_protocol_16check_parameters_2__call__(((struct __pyx_obj_8catalyst_9_protocol_check_parameters *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":72
 *     def __call__(self, func):
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):             # <<<<<<<<<<<<<<
 *             cdef short i
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_16check_parameters_8__call___1assert_keywords_and_call(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8catalyst_9_protocol_16check_parameters_8__call___1assert_keywords_and_call = {"assert_keywords_and_call", (PyCFunction)__pyx_pw_8catalyst_9_protocol_16check_parameters_8__call___1assert_keywords_and_call, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8catalyst_9_protocol_16check_parameters_8__call___1assert_keywords_and_call(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("assert_keywords_and_call (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "assert_keywords_and_call", 1))) return NULL;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_8catalyst_9_protocol_16check_parameters_8__call___assert_keywords_and_call(__pyx_self, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_16check_parameters_8__call___assert_keywords_and_call(PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *__pyx_cur_scope;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *__pyx_outer_scope;
  short __pyx_v_i;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_expected_type = NULL;
  PyObject *__pyx_v_expected_type_name = NULL;
  PyObject *__pyx_v_keyword = NULL;
  PyObject *__pyx_v_type_ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  short __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  int __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  __Pyx_RefNannySetupContext("assert_keywords_and_call", 0);
  __pyx_outer_scope = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "catalyst/_protocol.pyx":76
 * 
 *             # verify all the keyword arguments
 *             for field in kwargs:             # <<<<<<<<<<<<<<
 *                 if field not in self.keyword_names:
 *                     raise TypeError("%s() got an unexpected keyword argument"
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, ((PyObject *)NULL), (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "catalyst/_protocol.pyx":77
 *             # verify all the keyword arguments
 *             for field in kwargs:
 *                 if field not in self.keyword_names:             # <<<<<<<<<<<<<<
 *                     raise TypeError("%s() got an unexpected keyword argument"
 *                                     " '%s'" % (func.__name__, field))
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 77, __pyx_L1_error) }
    __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_v_field, __pyx_cur_scope->__pyx_v_self->keyword_names, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 77, __pyx_L1_error)
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "catalyst/_protocol.pyx":79
 *                 if field not in self.keyword_names:
 *                     raise TypeError("%s() got an unexpected keyword argument"
 *                                     " '%s'" % (func.__name__, field))             # <<<<<<<<<<<<<<
 * 
 *             # verify type of each argument
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_func)) { __Pyx_RaiseClosureNameError("func"); __PYX_ERR(0, 79, __pyx_L1_error) }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_func, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_field);
      __Pyx_GIVEREF(__pyx_v_field);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_field);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_got_an_unexpected_keyword_arg, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "catalyst/_protocol.pyx":78
 *             for field in kwargs:
 *                 if field not in self.keyword_names:
 *                     raise TypeError("%s() got an unexpected keyword argument"             # <<<<<<<<<<<<<<
 *                                     " '%s'" % (func.__name__, field))
 * 
 */
      __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 78, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 78, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 78, __pyx_L1_error)

      /* "catalyst/_protocol.pyx":77
 *             # verify all the keyword arguments
 *             for field in kwargs:
 *                 if field not in self.keyword_names:             # <<<<<<<<<<<<<<
 *                     raise TypeError("%s() got an unexpected keyword argument"
 *                                     " '%s'" % (func.__name__, field))
 */
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":82
 * 
 *             # verify type of each argument
 *             for i, arg in enumerate(args[1:]):             # <<<<<<<<<<<<<<
 *                 expected_type = self.types[i]
 * 
 */
  __pyx_t_10 = 0;
  __pyx_t_1 = __Pyx_PyTuple_GetSlice(__pyx_v_args, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __pyx_t_1; __Pyx_INCREF(__pyx_t_5); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 82, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_i = __pyx_t_10;
    __pyx_t_10 = (__pyx_t_10 + 1);

    /* "catalyst/_protocol.pyx":83
 *             # verify type of each argument
 *             for i, arg in enumerate(args[1:]):
 *                 expected_type = self.types[i]             # <<<<<<<<<<<<<<
 * 
 *                 if (i == 0 or i == 1) and _is_iterable(arg):
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 83, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_self->types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 83, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_cur_scope->__pyx_v_self->types, __pyx_v_i, short, 1, __Pyx_PyInt_From_short, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_expected_type, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":85
 *                 expected_type = self.types[i]
 * 
 *                 if (i == 0 or i == 1) and _is_iterable(arg):             # <<<<<<<<<<<<<<
 *                     if len(arg) == 0:
 *                         continue
 */
    switch (__pyx_v_i) {
      case 0:
      case 1:
      __pyx_t_7 = 1;
      break;
      default:
      __pyx_t_7 = 0;
      break;
    }
    if (__pyx_t_7) {
    } else {
      __pyx_t_8 = __pyx_t_7;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((PyObject *)__pyx_f_8catalyst_9_protocol__is_iterable(__pyx_v_arg)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __pyx_t_7;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_8) {

      /* "catalyst/_protocol.pyx":86
 * 
 *                 if (i == 0 or i == 1) and _is_iterable(arg):
 *                     if len(arg) == 0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     arg = arg[0]
 */
      __pyx_t_2 = PyObject_Length(__pyx_v_arg); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 86, __pyx_L1_error)
      __pyx_t_8 = ((__pyx_t_2 == 0) != 0);
      if (__pyx_t_8) {

        /* "catalyst/_protocol.pyx":87
 *                 if (i == 0 or i == 1) and _is_iterable(arg):
 *                     if len(arg) == 0:
 *                         continue             # <<<<<<<<<<<<<<
 *                     arg = arg[0]
 * 
 */
        goto __pyx_L6_continue;

        /* "catalyst/_protocol.pyx":86
 * 
 *                 if (i == 0 or i == 1) and _is_iterable(arg):
 *                     if len(arg) == 0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     arg = arg[0]
 */
      }

      /* "catalyst/_protocol.pyx":88
 *                     if len(arg) == 0:
 *                         continue
 *                     arg = arg[0]             # <<<<<<<<<<<<<<
 * 
 *                 if not isinstance(arg, expected_type):
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_arg, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_arg, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "catalyst/_protocol.pyx":85
 *                 expected_type = self.types[i]
 * 
 *                 if (i == 0 or i == 1) and _is_iterable(arg):             # <<<<<<<<<<<<<<
 *                     if len(arg) == 0:
 *                         continue
 */
    }

    /* "catalyst/_protocol.pyx":90
 *                     arg = arg[0]
 * 
 *                 if not isinstance(arg, expected_type):             # <<<<<<<<<<<<<<
 *                     expected_type_name = expected_type.__name__ \
 *                         if not _is_iterable(expected_type) \
 */
    __pyx_t_8 = PyObject_IsInstance(__pyx_v_arg, __pyx_v_expected_type); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __pyx_t_7 = ((!(__pyx_t_8 != 0)) != 0);
    if (__pyx_t_7) {

      /* "catalyst/_protocol.pyx":92
 *                 if not isinstance(arg, expected_type):
 *                     expected_type_name = expected_type.__name__ \
 *                         if not _is_iterable(expected_type) \             # <<<<<<<<<<<<<<
 *                         else ', '.join([type_.__name__ for type_ in expected_type])
 * 
 */
      __pyx_t_9 = ((PyObject *)__pyx_f_8catalyst_9_protocol__is_iterable(__pyx_v_expected_type)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 92, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 92, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (((!__pyx_t_7) != 0)) {

        /* "catalyst/_protocol.pyx":91
 * 
 *                 if not isinstance(arg, expected_type):
 *                     expected_type_name = expected_type.__name__ \             # <<<<<<<<<<<<<<
 *                         if not _is_iterable(expected_type) \
 *                         else ', '.join([type_.__name__ for type_ in expected_type])
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_expected_type, __pyx_n_s_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 91, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_1 = __pyx_t_9;
        __pyx_t_9 = 0;
      } else {

        /* "catalyst/_protocol.pyx":93
 *                     expected_type_name = expected_type.__name__ \
 *                         if not _is_iterable(expected_type) \
 *                         else ', '.join([type_.__name__ for type_ in expected_type])             # <<<<<<<<<<<<<<
 * 
 *                     raise TypeError("Expected %s argument to be of type %s%s" %
 */
        __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (likely(PyList_CheckExact(__pyx_v_expected_type)) || PyTuple_CheckExact(__pyx_v_expected_type)) {
          __pyx_t_11 = __pyx_v_expected_type; __Pyx_INCREF(__pyx_t_11); __pyx_t_2 = 0;
          __pyx_t_12 = NULL;
        } else {
          __pyx_t_2 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_v_expected_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 93, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_12)) {
            if (likely(PyList_CheckExact(__pyx_t_11))) {
              if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_11)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_13 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_2); __Pyx_INCREF(__pyx_t_13); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 93, __pyx_L1_error)
              #else
              __pyx_t_13 = PySequence_ITEM(__pyx_t_11, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 93, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
            } else {
              if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_2); __Pyx_INCREF(__pyx_t_13); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 93, __pyx_L1_error)
              #else
              __pyx_t_13 = PySequence_ITEM(__pyx_t_11, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 93, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
            }
          } else {
            __pyx_t_13 = __pyx_t_12(__pyx_t_11);
            if (unlikely(!__pyx_t_13)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 93, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_13);
          }
          __Pyx_XDECREF_SET(__pyx_v_type_, __pyx_t_13);
          __pyx_t_13 = 0;
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_type_, __pyx_n_s_name); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_t_13))) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyString_Join(__pyx_kp_s__3, __pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_1 = __pyx_t_11;
        __pyx_t_11 = 0;
      }
      __pyx_v_expected_type_name = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "catalyst/_protocol.pyx":96
 * 
 *                     raise TypeError("Expected %s argument to be of type %s%s" %
 *                         (self.keyword_names[i],             # <<<<<<<<<<<<<<
 *                          'or iterable of type ' if i in (0, 1) else '',
 *                          expected_type_name)
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 96, __pyx_L1_error) }
      if (unlikely(__pyx_cur_scope->__pyx_v_self->keyword_names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 96, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_cur_scope->__pyx_v_self->keyword_names, __pyx_v_i, short, 1, __Pyx_PyInt_From_short, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "catalyst/_protocol.pyx":97
 *                     raise TypeError("Expected %s argument to be of type %s%s" %
 *                         (self.keyword_names[i],
 *                          'or iterable of type ' if i in (0, 1) else '',             # <<<<<<<<<<<<<<
 *                          expected_type_name)
 *                     )
 */
      switch (__pyx_v_i) {
        case 0:
        case 1:
        __Pyx_INCREF(__pyx_kp_s_or_iterable_of_type);
        __pyx_t_11 = __pyx_kp_s_or_iterable_of_type;
        break;
        default:
        __Pyx_INCREF(__pyx_kp_s__4);
        __pyx_t_11 = __pyx_kp_s__4;
        break;
      }

      /* "catalyst/_protocol.pyx":96
 * 
 *                     raise TypeError("Expected %s argument to be of type %s%s" %
 *                         (self.keyword_names[i],             # <<<<<<<<<<<<<<
 *                          'or iterable of type ' if i in (0, 1) else '',
 *                          expected_type_name)
 */
      __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_11);
      __Pyx_INCREF(__pyx_v_expected_type_name);
      __Pyx_GIVEREF(__pyx_v_expected_type_name);
      PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_expected_type_name);
      __pyx_t_1 = 0;
      __pyx_t_11 = 0;

      /* "catalyst/_protocol.pyx":95
 *                         else ', '.join([type_.__name__ for type_ in expected_type])
 * 
 *                     raise TypeError("Expected %s argument to be of type %s%s" %             # <<<<<<<<<<<<<<
 *                         (self.keyword_names[i],
 *                          'or iterable of type ' if i in (0, 1) else '',
 */
      __pyx_t_11 = __Pyx_PyString_Format(__pyx_kp_s_Expected_s_argument_to_be_of_typ, __pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_11);
      __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_9, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_11, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __PYX_ERR(0, 95, __pyx_L1_error)

      /* "catalyst/_protocol.pyx":90
 *                     arg = arg[0]
 * 
 *                 if not isinstance(arg, expected_type):             # <<<<<<<<<<<<<<
 *                     expected_type_name = expected_type.__name__ \
 *                         if not _is_iterable(expected_type) \
 */
    }

    /* "catalyst/_protocol.pyx":82
 * 
 *             # verify type of each argument
 *             for i, arg in enumerate(args[1:]):             # <<<<<<<<<<<<<<
 *                 expected_type = self.types[i]
 * 
 */
    __pyx_L6_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "catalyst/_protocol.pyx":102
 * 
 *             # verify type of each kwarg
 *             for keyword, arg in iteritems(kwargs):             # <<<<<<<<<<<<<<
 *                 if keyword in ('assets', 'fields') and _is_iterable(arg):
 *                     if len(arg) == 0:
 */
  __pyx_t_11 = __Pyx_GetModuleGlobalName(__pyx_n_s_iteritems); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_11, function);
    }
  }
  if (!__pyx_t_9) {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_v_kwargs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_v_kwargs};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_v_kwargs};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9); __pyx_t_9 = NULL;
      __Pyx_INCREF(__pyx_v_kwargs);
      __Pyx_GIVEREF(__pyx_v_kwargs);
      PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_kwargs);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
    __pyx_t_11 = __pyx_t_5; __Pyx_INCREF(__pyx_t_11); __pyx_t_3 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 102, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_11))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_11)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 102, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_11, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 102, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_11, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_12(__pyx_t_11);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 102, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 102, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_13 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_13)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_1)) goto __pyx_L17_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_9 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_9)) goto __pyx_L17_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_13), 2) < 0) __PYX_ERR(0, 102, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L18_unpacking_done;
      __pyx_L17_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 102, __pyx_L1_error)
      __pyx_L18_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_keyword, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "catalyst/_protocol.pyx":103
 *             # verify type of each kwarg
 *             for keyword, arg in iteritems(kwargs):
 *                 if keyword in ('assets', 'fields') and _is_iterable(arg):             # <<<<<<<<<<<<<<
 *                     if len(arg) == 0:
 *                         continue
 */
    __Pyx_INCREF(__pyx_v_keyword);
    __pyx_t_5 = __pyx_v_keyword;
    __pyx_t_15 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_assets, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 103, __pyx_L1_error)
    if (!__pyx_t_15) {
    } else {
      __pyx_t_8 = __pyx_t_15;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_15 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_fields, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 103, __pyx_L1_error)
    __pyx_t_8 = __pyx_t_15;
    __pyx_L22_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_15 = (__pyx_t_8 != 0);
    if (__pyx_t_15) {
    } else {
      __pyx_t_7 = __pyx_t_15;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_5 = ((PyObject *)__pyx_f_8catalyst_9_protocol__is_iterable(__pyx_v_arg)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __pyx_t_15;
    __pyx_L20_bool_binop_done:;
    if (__pyx_t_7) {

      /* "catalyst/_protocol.pyx":104
 *             for keyword, arg in iteritems(kwargs):
 *                 if keyword in ('assets', 'fields') and _is_iterable(arg):
 *                     if len(arg) == 0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     arg = arg[0]
 */
      __pyx_t_2 = PyObject_Length(__pyx_v_arg); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 104, __pyx_L1_error)
      __pyx_t_7 = ((__pyx_t_2 == 0) != 0);
      if (__pyx_t_7) {

        /* "catalyst/_protocol.pyx":105
 *                 if keyword in ('assets', 'fields') and _is_iterable(arg):
 *                     if len(arg) == 0:
 *                         continue             # <<<<<<<<<<<<<<
 *                     arg = arg[0]
 *                 if not isinstance(arg, self.keys_to_types[keyword]):
 */
        goto __pyx_L15_continue;

        /* "catalyst/_protocol.pyx":104
 *             for keyword, arg in iteritems(kwargs):
 *                 if keyword in ('assets', 'fields') and _is_iterable(arg):
 *                     if len(arg) == 0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     arg = arg[0]
 */
      }

      /* "catalyst/_protocol.pyx":106
 *                     if len(arg) == 0:
 *                         continue
 *                     arg = arg[0]             # <<<<<<<<<<<<<<
 *                 if not isinstance(arg, self.keys_to_types[keyword]):
 *                     expected_type = self.keys_to_types[keyword].__name__ \
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_arg, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_arg, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "catalyst/_protocol.pyx":103
 *             # verify type of each kwarg
 *             for keyword, arg in iteritems(kwargs):
 *                 if keyword in ('assets', 'fields') and _is_iterable(arg):             # <<<<<<<<<<<<<<
 *                     if len(arg) == 0:
 *                         continue
 */
    }

    /* "catalyst/_protocol.pyx":107
 *                         continue
 *                     arg = arg[0]
 *                 if not isinstance(arg, self.keys_to_types[keyword]):             # <<<<<<<<<<<<<<
 *                     expected_type = self.keys_to_types[keyword].__name__ \
 *                         if not _is_iterable(self.keys_to_types[keyword]) \
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 107, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_self->keys_to_types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 107, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_self->keys_to_types, __pyx_v_keyword); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyObject_IsInstance(__pyx_v_arg, __pyx_t_5); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_15 = ((!(__pyx_t_7 != 0)) != 0);
    if (__pyx_t_15) {

      /* "catalyst/_protocol.pyx":109
 *                 if not isinstance(arg, self.keys_to_types[keyword]):
 *                     expected_type = self.keys_to_types[keyword].__name__ \
 *                         if not _is_iterable(self.keys_to_types[keyword]) \             # <<<<<<<<<<<<<<
 *                         else ', '.join([type_.__name__ for type_ in
 *                             self.keys_to_types[keyword]])
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 109, __pyx_L1_error) }
      if (unlikely(__pyx_cur_scope->__pyx_v_self->keys_to_types == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 109, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_self->keys_to_types, __pyx_v_keyword); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = ((PyObject *)__pyx_f_8catalyst_9_protocol__is_iterable(__pyx_t_9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 109, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (((!__pyx_t_15) != 0)) {

        /* "catalyst/_protocol.pyx":108
 *                     arg = arg[0]
 *                 if not isinstance(arg, self.keys_to_types[keyword]):
 *                     expected_type = self.keys_to_types[keyword].__name__ \             # <<<<<<<<<<<<<<
 *                         if not _is_iterable(self.keys_to_types[keyword]) \
 *                         else ', '.join([type_.__name__ for type_ in
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 108, __pyx_L1_error) }
        if (unlikely(__pyx_cur_scope->__pyx_v_self->keys_to_types == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 108, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_self->keys_to_types, __pyx_v_keyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_5 = __pyx_t_9;
        __pyx_t_9 = 0;
      } else {

        /* "catalyst/_protocol.pyx":110
 *                     expected_type = self.keys_to_types[keyword].__name__ \
 *                         if not _is_iterable(self.keys_to_types[keyword]) \
 *                         else ', '.join([type_.__name__ for type_ in             # <<<<<<<<<<<<<<
 *                             self.keys_to_types[keyword]])
 * 
 */
        __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 110, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "catalyst/_protocol.pyx":111
 *                         if not _is_iterable(self.keys_to_types[keyword]) \
 *                         else ', '.join([type_.__name__ for type_ in
 *                             self.keys_to_types[keyword]])             # <<<<<<<<<<<<<<
 * 
 *                     raise TypeError("Expected %s argument to be of type %s%s" %
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 111, __pyx_L1_error) }
        if (unlikely(__pyx_cur_scope->__pyx_v_self->keys_to_types == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 111, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_self->keys_to_types, __pyx_v_keyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
          __pyx_t_13 = __pyx_t_1; __Pyx_INCREF(__pyx_t_13); __pyx_t_2 = 0;
          __pyx_t_16 = NULL;
        } else {
          __pyx_t_2 = -1; __pyx_t_13 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_16 = Py_TYPE(__pyx_t_13)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 111, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        for (;;) {
          if (likely(!__pyx_t_16)) {
            if (likely(PyList_CheckExact(__pyx_t_13))) {
              if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_13)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyList_GET_ITEM(__pyx_t_13, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 111, __pyx_L1_error)
              #else
              __pyx_t_1 = PySequence_ITEM(__pyx_t_13, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            } else {
              if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_13)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_13, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 111, __pyx_L1_error)
              #else
              __pyx_t_1 = PySequence_ITEM(__pyx_t_13, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            }
          } else {
            __pyx_t_1 = __pyx_t_16(__pyx_t_13);
            if (unlikely(!__pyx_t_1)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 111, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_1);
          }
          __Pyx_XDECREF_SET(__pyx_v_type_, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "catalyst/_protocol.pyx":110
 *                     expected_type = self.keys_to_types[keyword].__name__ \
 *                         if not _is_iterable(self.keys_to_types[keyword]) \
 *                         else ', '.join([type_.__name__ for type_ in             # <<<<<<<<<<<<<<
 *                             self.keys_to_types[keyword]])
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_type_, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 110, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_13 = __Pyx_PyString_Join(__pyx_kp_s__3, __pyx_t_9); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 110, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_5 = __pyx_t_13;
        __pyx_t_13 = 0;
      }
      __Pyx_XDECREF_SET(__pyx_v_expected_type, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "catalyst/_protocol.pyx":115
 *                     raise TypeError("Expected %s argument to be of type %s%s" %
 *                                     (keyword,
 *                                      'or iterable of type ' if keyword in             # <<<<<<<<<<<<<<
 *                                      ('assets', 'fields') else '',
 *                                      expected_type)
 */
      __Pyx_INCREF(__pyx_v_keyword);
      __pyx_t_13 = __pyx_v_keyword;
      __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_13, __pyx_n_s_assets, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 115, __pyx_L1_error)
      if (!__pyx_t_7) {
      } else {
        __pyx_t_15 = __pyx_t_7;
        goto __pyx_L28_bool_binop_done;
      }
      __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_13, __pyx_n_s_fields, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 115, __pyx_L1_error)
      __pyx_t_15 = __pyx_t_7;
      __pyx_L28_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if ((__pyx_t_15 != 0)) {
        __Pyx_INCREF(__pyx_kp_s_or_iterable_of_type);
        __pyx_t_5 = __pyx_kp_s_or_iterable_of_type;
      } else {
        __Pyx_INCREF(__pyx_kp_s__4);
        __pyx_t_5 = __pyx_kp_s__4;
      }

      /* "catalyst/_protocol.pyx":114
 * 
 *                     raise TypeError("Expected %s argument to be of type %s%s" %
 *                                     (keyword,             # <<<<<<<<<<<<<<
 *                                      'or iterable of type ' if keyword in
 *                                      ('assets', 'fields') else '',
 */
      __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 114, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_INCREF(__pyx_v_keyword);
      __Pyx_GIVEREF(__pyx_v_keyword);
      PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_keyword);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_expected_type);
      __Pyx_GIVEREF(__pyx_v_expected_type);
      PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_expected_type);
      __pyx_t_5 = 0;

      /* "catalyst/_protocol.pyx":113
 *                             self.keys_to_types[keyword]])
 * 
 *                     raise TypeError("Expected %s argument to be of type %s%s" %             # <<<<<<<<<<<<<<
 *                                     (keyword,
 *                                      'or iterable of type ' if keyword in
 */
      __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Expected_s_argument_to_be_of_typ, __pyx_t_13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 113, __pyx_L1_error)

      /* "catalyst/_protocol.pyx":107
 *                         continue
 *                     arg = arg[0]
 *                 if not isinstance(arg, self.keys_to_types[keyword]):             # <<<<<<<<<<<<<<
 *                     expected_type = self.keys_to_types[keyword].__name__ \
 *                         if not _is_iterable(self.keys_to_types[keyword]) \
 */
    }

    /* "catalyst/_protocol.pyx":102
 * 
 *             # verify type of each kwarg
 *             for keyword, arg in iteritems(kwargs):             # <<<<<<<<<<<<<<
 *                 if keyword in ('assets', 'fields') and _is_iterable(arg):
 *                     if len(arg) == 0:
 */
    __pyx_L15_continue:;
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "catalyst/_protocol.pyx":120
 *                     )
 * 
 *             return func(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *         return assert_keywords_and_call
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_func)) { __Pyx_RaiseClosureNameError("func"); __PYX_ERR(0, 120, __pyx_L1_error) }
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_v_func, __pyx_v_args, __pyx_v_kwargs); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":72
 *     def __call__(self, func):
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):             # <<<<<<<<<<<<<<
 *             cdef short i
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("catalyst._protocol.check_parameters.__call__.assert_keywords_and_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_expected_type);
  __Pyx_XDECREF(__pyx_v_expected_type_name);
  __Pyx_XDECREF(__pyx_v_keyword);
  __Pyx_XDECREF(__pyx_v_type_);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":70
 *         self.keys_to_types = dict(zip(keyword_names, types))
 * 
 *     def __call__(self, func):             # <<<<<<<<<<<<<<
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):
 */

static PyObject *__pyx_pf_8catalyst_9_protocol_16check_parameters_2__call__(struct __pyx_obj_8catalyst_9_protocol_check_parameters *__pyx_v_self, PyObject *__pyx_v_func) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *__pyx_cur_scope;
  PyObject *__pyx_v_assert_keywords_and_call = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__call__", 0);
  __pyx_cur_scope = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *)__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct____call__(__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct____call__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 70, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_func = __pyx_v_func;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_func);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_func);

  /* "catalyst/_protocol.pyx":71
 * 
 *     def __call__(self, func):
 *         @no_wraps_py2(func)             # <<<<<<<<<<<<<<
 *         def assert_keywords_and_call(*args, **kwargs):
 *             cdef short i
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_no_wraps_py2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_cur_scope->__pyx_v_func); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_func};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_func};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_func);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_func);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_cur_scope->__pyx_v_func);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "catalyst/_protocol.pyx":72
 *     def __call__(self, func):
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):             # <<<<<<<<<<<<<<
 *             cdef short i
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8catalyst_9_protocol_16check_parameters_8__call___1assert_keywords_and_call, 0, __pyx_n_s_call___locals_assert_keywords, ((PyObject*)__pyx_cur_scope), __pyx_n_s_catalyst__protocol, __pyx_d, ((PyObject *)__pyx_codeobj__6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_assert_keywords_and_call = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":122
 *             return func(*args, **kwargs)
 * 
 *         return assert_keywords_and_call             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_assert_keywords_and_call);
  __pyx_r = __pyx_v_assert_keywords_and_call;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":70
 *         self.keys_to_types = dict(zip(keyword_names, types))
 * 
 *     def __call__(self, func):             # <<<<<<<<<<<<<<
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.check_parameters.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_assert_keywords_and_call);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8catalyst_9_protocol_4generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "catalyst/_protocol.pyx":126
 * 
 * @contextmanager
 * def handle_non_market_minutes(bar_data):             # <<<<<<<<<<<<<<
 *     try:
 *         bar_data._handle_non_market_minutes = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_3handle_non_market_minutes(PyObject *__pyx_self, PyObject *__pyx_v_bar_data); /*proto*/
static PyMethodDef __pyx_mdef_8catalyst_9_protocol_3handle_non_market_minutes = {"handle_non_market_minutes", (PyCFunction)__pyx_pw_8catalyst_9_protocol_3handle_non_market_minutes, METH_O, 0};
static PyObject *__pyx_pw_8catalyst_9_protocol_3handle_non_market_minutes(PyObject *__pyx_self, PyObject *__pyx_v_bar_data) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("handle_non_market_minutes (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_2handle_non_market_minutes(__pyx_self, ((PyObject *)__pyx_v_bar_data));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_2handle_non_market_minutes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bar_data) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("handle_non_market_minutes", 0);
  __pyx_cur_scope = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *)__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes(__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 126, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_bar_data = __pyx_v_bar_data;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_bar_data);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_bar_data);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8catalyst_9_protocol_4generator, (PyObject *) __pyx_cur_scope, __pyx_n_s_handle_non_market_minutes, __pyx_n_s_handle_non_market_minutes, __pyx_n_s_catalyst__protocol); if (unlikely(!gen)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("catalyst._protocol.handle_non_market_minutes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8catalyst_9_protocol_4generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *__pyx_cur_scope = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  int __pyx_t_2;
  char const *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 126, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":127
 * @contextmanager
 * def handle_non_market_minutes(bar_data):
 *     try:             # <<<<<<<<<<<<<<
 *         bar_data._handle_non_market_minutes = True
 *         yield
 */
  /*try:*/ {

    /* "catalyst/_protocol.pyx":128
 * def handle_non_market_minutes(bar_data):
 *     try:
 *         bar_data._handle_non_market_minutes = True             # <<<<<<<<<<<<<<
 *         yield
 *     finally:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_bar_data, __pyx_n_s_handle_non_market_minutes_2, Py_True) < 0) __PYX_ERR(0, 128, __pyx_L5_error)

    /* "catalyst/_protocol.pyx":129
 *     try:
 *         bar_data._handle_non_market_minutes = True
 *         yield             # <<<<<<<<<<<<<<
 *     finally:
 *         bar_data._handle_non_market_minutes = False
 */
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 129, __pyx_L5_error)
  }

  /* "catalyst/_protocol.pyx":131
 *         yield
 *     finally:
 *         bar_data._handle_non_market_minutes = False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_bar_data, __pyx_n_s_handle_non_market_minutes_2, Py_False) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
      goto __pyx_L6;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L5_error:;
      __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0)) __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __pyx_t_1 = __pyx_lineno; __pyx_t_2 = __pyx_clineno; __pyx_t_3 = __pyx_filename;
      {
        if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_bar_data, __pyx_n_s_handle_non_market_minutes_2, Py_False) < 0) __PYX_ERR(0, 131, __pyx_L9_error)
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0;
      __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_2; __pyx_filename = __pyx_t_3;
      goto __pyx_L1_error;
      __pyx_L9_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "catalyst/_protocol.pyx":126
 * 
 * @contextmanager
 * def handle_non_market_minutes(bar_data):             # <<<<<<<<<<<<<<
 *     try:
 *         bar_data._handle_non_market_minutes = True
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("handle_non_market_minutes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":173
 *     cdef bool _adjust_minutes
 * 
 *     def __init__(self, data_portal, simulation_dt_func, data_frequency,             # <<<<<<<<<<<<<<
 *                  trading_calendar, restrictions, universe_func=None):
 *         self.data_portal = data_portal
 */

/* Python wrapper */
static int __pyx_pw_8catalyst_9_protocol_7BarData_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8catalyst_9_protocol_7BarData_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data_portal = 0;
  PyObject *__pyx_v_simulation_dt_func = 0;
  PyObject *__pyx_v_data_frequency = 0;
  PyObject *__pyx_v_trading_calendar = 0;
  PyObject *__pyx_v_restrictions = 0;
  PyObject *__pyx_v_universe_func = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data_portal,&__pyx_n_s_simulation_dt_func,&__pyx_n_s_data_frequency,&__pyx_n_s_trading_calendar,&__pyx_n_s_restrictions,&__pyx_n_s_universe_func,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "catalyst/_protocol.pyx":174
 * 
 *     def __init__(self, data_portal, simulation_dt_func, data_frequency,
 *                  trading_calendar, restrictions, universe_func=None):             # <<<<<<<<<<<<<<
 *         self.data_portal = data_portal
 *         self.simulation_dt_func = simulation_dt_func
 */
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data_portal)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_simulation_dt_func)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 6, 1); __PYX_ERR(0, 173, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data_frequency)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 6, 2); __PYX_ERR(0, 173, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_trading_calendar)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 6, 3); __PYX_ERR(0, 173, __pyx_L3_error)
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_restrictions)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 6, 4); __PYX_ERR(0, 173, __pyx_L3_error)
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_universe_func);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 173, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_data_portal = values[0];
    __pyx_v_simulation_dt_func = values[1];
    __pyx_v_data_frequency = values[2];
    __pyx_v_trading_calendar = values[3];
    __pyx_v_restrictions = values[4];
    __pyx_v_universe_func = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 173, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("catalyst._protocol.BarData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData___init__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), __pyx_v_data_portal, __pyx_v_simulation_dt_func, __pyx_v_data_frequency, __pyx_v_trading_calendar, __pyx_v_restrictions, __pyx_v_universe_func);

  /* "catalyst/_protocol.pyx":173
 *     cdef bool _adjust_minutes
 * 
 *     def __init__(self, data_portal, simulation_dt_func, data_frequency,             # <<<<<<<<<<<<<<
 *                  trading_calendar, restrictions, universe_func=None):
 *         self.data_portal = data_portal
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8catalyst_9_protocol_7BarData___init__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_data_portal, PyObject *__pyx_v_simulation_dt_func, PyObject *__pyx_v_data_frequency, PyObject *__pyx_v_trading_calendar, PyObject *__pyx_v_restrictions, PyObject *__pyx_v_universe_func) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "catalyst/_protocol.pyx":175
 *     def __init__(self, data_portal, simulation_dt_func, data_frequency,
 *                  trading_calendar, restrictions, universe_func=None):
 *         self.data_portal = data_portal             # <<<<<<<<<<<<<<
 *         self.simulation_dt_func = simulation_dt_func
 *         self.data_frequency = data_frequency
 */
  __Pyx_INCREF(__pyx_v_data_portal);
  __Pyx_GIVEREF(__pyx_v_data_portal);
  __Pyx_GOTREF(__pyx_v_self->data_portal);
  __Pyx_DECREF(__pyx_v_self->data_portal);
  __pyx_v_self->data_portal = __pyx_v_data_portal;

  /* "catalyst/_protocol.pyx":176
 *                  trading_calendar, restrictions, universe_func=None):
 *         self.data_portal = data_portal
 *         self.simulation_dt_func = simulation_dt_func             # <<<<<<<<<<<<<<
 *         self.data_frequency = data_frequency
 *         self._views = {}
 */
  __Pyx_INCREF(__pyx_v_simulation_dt_func);
  __Pyx_GIVEREF(__pyx_v_simulation_dt_func);
  __Pyx_GOTREF(__pyx_v_self->simulation_dt_func);
  __Pyx_DECREF(__pyx_v_self->simulation_dt_func);
  __pyx_v_self->simulation_dt_func = __pyx_v_simulation_dt_func;

  /* "catalyst/_protocol.pyx":177
 *         self.data_portal = data_portal
 *         self.simulation_dt_func = simulation_dt_func
 *         self.data_frequency = data_frequency             # <<<<<<<<<<<<<<
 *         self._views = {}
 * 
 */
  __Pyx_INCREF(__pyx_v_data_frequency);
  __Pyx_GIVEREF(__pyx_v_data_frequency);
  __Pyx_GOTREF(__pyx_v_self->data_frequency);
  __Pyx_DECREF(__pyx_v_self->data_frequency);
  __pyx_v_self->data_frequency = __pyx_v_data_frequency;

  /* "catalyst/_protocol.pyx":178
 *         self.simulation_dt_func = simulation_dt_func
 *         self.data_frequency = data_frequency
 *         self._views = {}             # <<<<<<<<<<<<<<
 * 
 *         self._daily_mode = (self.data_frequency == "daily")
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_views);
  __Pyx_DECREF(__pyx_v_self->_views);
  __pyx_v_self->_views = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":180
 *         self._views = {}
 * 
 *         self._daily_mode = (self.data_frequency == "daily")             # <<<<<<<<<<<<<<
 * 
 *         self._universe_func = universe_func
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->data_frequency, __pyx_n_s_daily, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7cpython_4bool_bool))))) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_daily_mode);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_daily_mode));
  __pyx_v_self->_daily_mode = ((PyBoolObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":182
 *         self._daily_mode = (self.data_frequency == "daily")
 * 
 *         self._universe_func = universe_func             # <<<<<<<<<<<<<<
 *         self._last_calculated_universe = None
 *         self._universe_last_updated_at = None
 */
  __Pyx_INCREF(__pyx_v_universe_func);
  __Pyx_GIVEREF(__pyx_v_universe_func);
  __Pyx_GOTREF(__pyx_v_self->_universe_func);
  __Pyx_DECREF(__pyx_v_self->_universe_func);
  __pyx_v_self->_universe_func = __pyx_v_universe_func;

  /* "catalyst/_protocol.pyx":183
 * 
 *         self._universe_func = universe_func
 *         self._last_calculated_universe = None             # <<<<<<<<<<<<<<
 *         self._universe_last_updated_at = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_calculated_universe);
  __Pyx_DECREF(__pyx_v_self->_last_calculated_universe);
  __pyx_v_self->_last_calculated_universe = Py_None;

  /* "catalyst/_protocol.pyx":184
 *         self._universe_func = universe_func
 *         self._last_calculated_universe = None
 *         self._universe_last_updated_at = None             # <<<<<<<<<<<<<<
 * 
 *         self._adjust_minutes = False
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_universe_last_updated_at);
  __Pyx_DECREF(__pyx_v_self->_universe_last_updated_at);
  __pyx_v_self->_universe_last_updated_at = Py_None;

  /* "catalyst/_protocol.pyx":186
 *         self._universe_last_updated_at = None
 * 
 *         self._adjust_minutes = False             # <<<<<<<<<<<<<<
 * 
 *         self._trading_calendar = trading_calendar
 */
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  __Pyx_GOTREF(__pyx_v_self->_adjust_minutes);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_adjust_minutes));
  __pyx_v_self->_adjust_minutes = ((PyBoolObject *)Py_False);

  /* "catalyst/_protocol.pyx":188
 *         self._adjust_minutes = False
 * 
 *         self._trading_calendar = trading_calendar             # <<<<<<<<<<<<<<
 *         self._is_restricted = restrictions.is_restricted
 * 
 */
  __Pyx_INCREF(__pyx_v_trading_calendar);
  __Pyx_GIVEREF(__pyx_v_trading_calendar);
  __Pyx_GOTREF(__pyx_v_self->_trading_calendar);
  __Pyx_DECREF(__pyx_v_self->_trading_calendar);
  __pyx_v_self->_trading_calendar = __pyx_v_trading_calendar;

  /* "catalyst/_protocol.pyx":189
 * 
 *         self._trading_calendar = trading_calendar
 *         self._is_restricted = restrictions.is_restricted             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_equity_price_view(self, asset):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_restrictions, __pyx_n_s_is_restricted); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_is_restricted);
  __Pyx_DECREF(__pyx_v_self->_is_restricted);
  __pyx_v_self->_is_restricted = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":173
 *     cdef bool _adjust_minutes
 * 
 *     def __init__(self, data_portal, simulation_dt_func, data_frequency,             # <<<<<<<<<<<<<<
 *                  trading_calendar, restrictions, universe_func=None):
 *         self.data_portal = data_portal
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("catalyst._protocol.BarData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":191
 *         self._is_restricted = restrictions.is_restricted
 * 
 *     cdef _get_equity_price_view(self, asset):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a DataPortalSidView for the given asset.  Used to support the
 */

static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__get_equity_price_view(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset) {
  PyObject *__pyx_v_view = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannySetupContext("_get_equity_price_view", 0);
  __Pyx_INCREF(__pyx_v_asset);

  /* "catalyst/_protocol.pyx":205
 *         SidView : Accessor into the given asset's data.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self._warn_deprecated("`data[sid(N)]` is deprecated. Use "
 *                             "`data.current`.")
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "catalyst/_protocol.pyx":206
 *         """
 *         try:
 *             self._warn_deprecated("`data[sid(N)]` is deprecated. Use "             # <<<<<<<<<<<<<<
 *                             "`data.current`.")
 *             view = self._views[asset]
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_data_sid_N_is_deprecated_Use_da); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "catalyst/_protocol.pyx":208
 *             self._warn_deprecated("`data[sid(N)]` is deprecated. Use "
 *                             "`data.current`.")
 *             view = self._views[asset]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             try:
 */
      if (unlikely(__pyx_v_self->_views == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 208, __pyx_L3_error)
      }
      __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_self->_views, __pyx_v_asset); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 208, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_view = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "catalyst/_protocol.pyx":205
 *         SidView : Accessor into the given asset's data.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self._warn_deprecated("`data[sid(N)]` is deprecated. Use "
 *                             "`data.current`.")
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "catalyst/_protocol.pyx":209
 *                             "`data.current`.")
 *             view = self._views[asset]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             try:
 *                 asset = self.data_portal.asset_finder.retrieve_asset(asset)
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("catalyst._protocol.BarData._get_equity_price_view", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 209, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "catalyst/_protocol.pyx":210
 *             view = self._views[asset]
 *         except KeyError:
 *             try:             # <<<<<<<<<<<<<<
 *                 asset = self.data_portal.asset_finder.retrieve_asset(asset)
 *             except ValueError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        /*try:*/ {

          /* "catalyst/_protocol.pyx":211
 *         except KeyError:
 *             try:
 *                 asset = self.data_portal.asset_finder.retrieve_asset(asset)             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 # assume fetcher
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_asset_finder); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 211, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_retrieve_asset); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 211, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
            }
          }
          if (!__pyx_t_12) {
            __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_asset); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 211, __pyx_L13_error)
            __Pyx_GOTREF(__pyx_t_11);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_13)) {
              PyObject *__pyx_temp[2] = {__pyx_t_12, __pyx_v_asset};
              __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 211, __pyx_L13_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_11);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
              PyObject *__pyx_temp[2] = {__pyx_t_12, __pyx_v_asset};
              __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 211, __pyx_L13_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_11);
            } else
            #endif
            {
              __pyx_t_14 = PyTuple_New(1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 211, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
              __Pyx_INCREF(__pyx_v_asset);
              __Pyx_GIVEREF(__pyx_v_asset);
              PyTuple_SET_ITEM(__pyx_t_14, 0+1, __pyx_v_asset);
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_14, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 211, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF_SET(__pyx_v_asset, __pyx_t_11);
          __pyx_t_11 = 0;

          /* "catalyst/_protocol.pyx":210
 *             view = self._views[asset]
 *         except KeyError:
 *             try:             # <<<<<<<<<<<<<<
 *                 asset = self.data_portal.asset_finder.retrieve_asset(asset)
 *             except ValueError:
 */
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L20_try_end;
        __pyx_L13_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "catalyst/_protocol.pyx":212
 *             try:
 *                 asset = self.data_portal.asset_finder.retrieve_asset(asset)
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 # assume fetcher
 *                 pass
 */
        __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
        if (__pyx_t_5) {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L14_exception_handled;
        }
        goto __pyx_L15_except_error;
        __pyx_L15_except_error:;

        /* "catalyst/_protocol.pyx":210
 *             view = self._views[asset]
 *         except KeyError:
 *             try:             # <<<<<<<<<<<<<<
 *                 asset = self.data_portal.asset_finder.retrieve_asset(asset)
 *             except ValueError:
 */
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        goto __pyx_L5_except_error;
        __pyx_L14_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        __pyx_L20_try_end:;
      }

      /* "catalyst/_protocol.pyx":215
 *                 # assume fetcher
 *                 pass
 *             view = self._views[asset] = self._create_sid_view(asset)             # <<<<<<<<<<<<<<
 * 
 *         return view
 */
      __pyx_t_11 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_create_sid_view(__pyx_v_self, __pyx_v_asset); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 215, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_XDECREF_SET(__pyx_v_view, __pyx_t_11);
      if (unlikely(__pyx_v_self->_views == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 215, __pyx_L5_except_error)
      }
      if (unlikely(PyDict_SetItem(__pyx_v_self->_views, __pyx_v_asset, __pyx_t_11) < 0)) __PYX_ERR(0, 215, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "catalyst/_protocol.pyx":205
 *         SidView : Accessor into the given asset's data.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self._warn_deprecated("`data[sid(N)]` is deprecated. Use "
 *                             "`data.current`.")
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L10_try_end:;
  }

  /* "catalyst/_protocol.pyx":217
 *             view = self._views[asset] = self._create_sid_view(asset)
 * 
 *         return view             # <<<<<<<<<<<<<<
 * 
 *     cdef _create_sid_view(self, asset):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_view);
  __pyx_r = __pyx_v_view;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":191
 *         self._is_restricted = restrictions.is_restricted
 * 
 *     cdef _get_equity_price_view(self, asset):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a DataPortalSidView for the given asset.  Used to support the
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("catalyst._protocol.BarData._get_equity_price_view", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_view);
  __Pyx_XDECREF(__pyx_v_asset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":219
 *         return view
 * 
 *     cdef _create_sid_view(self, asset):             # <<<<<<<<<<<<<<
 *         return SidView(
 *             asset,
 */

static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__create_sid_view(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_create_sid_view", 0);

  /* "catalyst/_protocol.pyx":220
 * 
 *     cdef _create_sid_view(self, asset):
 *         return SidView(             # <<<<<<<<<<<<<<
 *             asset,
 *             self.data_portal,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "catalyst/_protocol.pyx":224
 *             self.data_portal,
 *             self.simulation_dt_func,
 *             self.data_frequency             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_asset);
  __Pyx_GIVEREF(__pyx_v_asset);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_asset);
  __Pyx_INCREF(__pyx_v_self->data_portal);
  __Pyx_GIVEREF(__pyx_v_self->data_portal);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->data_portal);
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __Pyx_GIVEREF(__pyx_v_self->simulation_dt_func);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->simulation_dt_func);
  __Pyx_INCREF(__pyx_v_self->data_frequency);
  __Pyx_GIVEREF(__pyx_v_self->data_frequency);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->data_frequency);

  /* "catalyst/_protocol.pyx":220
 * 
 *     cdef _create_sid_view(self, asset):
 *         return SidView(             # <<<<<<<<<<<<<<
 *             asset,
 *             self.data_portal,
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8catalyst_9_protocol_SidView), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":219
 *         return view
 * 
 *     cdef _create_sid_view(self, asset):             # <<<<<<<<<<<<<<
 *         return SidView(
 *             asset,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("catalyst._protocol.BarData._create_sid_view", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":227
 *         )
 * 
 *     cdef _get_current_minute(self):             # <<<<<<<<<<<<<<
 *         """
 *         Internal utility method to get the current simulation time.
 */

static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__get_current_minute(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_v_dt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_get_current_minute", 0);

  /* "catalyst/_protocol.pyx":239
 *         - if we're in daily mode, get the session label for this minute.
 *         """
 *         dt = self.simulation_dt_func()             # <<<<<<<<<<<<<<
 * 
 *         if self._adjust_minutes:
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_2 = __pyx_v_self->simulation_dt_func; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_dt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":241
 *         dt = self.simulation_dt_func()
 * 
 *         if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *             dt = \
 *                 self.data_portal.trading_calendar.previous_minute(dt)
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 241, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "catalyst/_protocol.pyx":243
 *         if self._adjust_minutes:
 *             dt = \
 *                 self.data_portal.trading_calendar.previous_minute(dt)             # <<<<<<<<<<<<<<
 * 
 *         if self._daily_mode:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_trading_calendar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_previous_minute); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_2) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_dt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_dt};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_dt};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 243, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
        __Pyx_INCREF(__pyx_v_dt);
        __Pyx_GIVEREF(__pyx_v_dt);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_dt);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_dt, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":241
 *         dt = self.simulation_dt_func()
 * 
 *         if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *             dt = \
 *                 self.data_portal.trading_calendar.previous_minute(dt)
 */
  }

  /* "catalyst/_protocol.pyx":245
 *                 self.data_portal.trading_calendar.previous_minute(dt)
 * 
 *         if self._daily_mode:             # <<<<<<<<<<<<<<
 *             # if we're in daily mode, take the given dt (which is the last
 *             # minute of the session) and get the session label for it.
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_daily_mode)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "catalyst/_protocol.pyx":248
 *             # if we're in daily mode, take the given dt (which is the last
 *             # minute of the session) and get the session label for it.
 *             dt = self.data_portal.trading_calendar.minute_to_session_label(dt)             # <<<<<<<<<<<<<<
 * 
 *         return dt
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_trading_calendar); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_minute_to_session_label); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_dt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_dt};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_dt};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_v_dt);
        __Pyx_GIVEREF(__pyx_v_dt);
        PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_v_dt);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_dt, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":245
 *                 self.data_portal.trading_calendar.previous_minute(dt)
 * 
 *         if self._daily_mode:             # <<<<<<<<<<<<<<
 *             # if we're in daily mode, take the given dt (which is the last
 *             # minute of the session) and get the session label for it.
 */
  }

  /* "catalyst/_protocol.pyx":250
 *             dt = self.data_portal.trading_calendar.minute_to_session_label(dt)
 * 
 *         return dt             # <<<<<<<<<<<<<<
 * 
 *     @check_parameters(('assets', 'fields'),
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dt);
  __pyx_r = __pyx_v_dt;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":227
 *         )
 * 
 *     cdef _get_current_minute(self):             # <<<<<<<<<<<<<<
 *         """
 *         Internal utility method to get the current simulation time.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.BarData._get_current_minute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":254
 *     @check_parameters(('assets', 'fields'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types))
 *     def current(self, assets, fields):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the current value of the given assets for the given fields
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_3current(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8catalyst_9_protocol_7BarData_2current[] = "\n        Returns the current value of the given assets for the given fields\n        at the current simulation time.  Current values are the as-traded price\n        and are usually not adjusted for events like splits or dividends (see\n        notes for more information).\n\n        Parameters\n        ----------\n        assets : Asset or iterable of Assets\n        fields : str or iterable[str].\n            Valid values are: \"price\",\n            \"last_traded\", \"open\", \"high\", \"low\", \"close\", \"volume\", or column\n            names in files read by ``fetch_csv``.\n\n        Returns\n        -------\n        current_value : Scalar, pandas Series, or pandas DataFrame.\n            See notes below.\n\n        Notes\n        -----\n        If a single asset and a single field are passed in, a scalar float\n        value is returned.\n\n        If a single asset and a list of fields are passed in, a pandas Series\n        is returned whose indices are the fields, and whose values are scalar\n        values for this asset for each field.\n\n        If a list of assets and a single field are passed in, a pandas Series\n        is returned whose indices are the assets, and whose values are scalar\n        values for each asset for the given field.\n\n        If a list of assets and a list of fields are passed in, a pandas\n        DataFrame is returned, indexed by asset.  The columns are the requested\n        fields, filled with the scalar values for each asset for each field.\n\n        If the current simulation time is not a valid market time, we use the\n        last market close instead.\n\n        \"price\" returns the last known close price of the asset.  If there is\n        no last known value (either because the asset has never traded, or\n        because it has delisted) NaN is returned.  If a value is found, and we\n        had to cross an adjustment boundary (split, dividend, etc) to get it,\n        the value is adjusted before being returned"".\n\n        \"last_traded\" returns the date of the last trade event of the asset,\n        even if the asset has stopped trading. If there is no last known value,\n        pd.NaT is returned.\n\n        \"volume\" returns the trade volume for the current simulation time.  If\n        there is no trade this minute, 0 is returned.\n\n        \"open\", \"high\", \"low\", and \"close\" return the relevant information for\n        the current trade bar.  If there is no current trade bar, NaN is\n        returned.\n        ";
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_3current(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_assets = 0;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("current (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_assets,&__pyx_n_s_fields,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_assets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_fields)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("current", 1, 2, 2, 1); __PYX_ERR(0, 254, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "current") < 0)) __PYX_ERR(0, 254, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_assets = values[0];
    __pyx_v_fields = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("current", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 254, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("catalyst._protocol.BarData.current", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_2current(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), __pyx_v_assets, __pyx_v_fields);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_2current(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets, PyObject *__pyx_v_fields) {
  PyBoolObject *__pyx_v_multiple_assets = NULL;
  PyBoolObject *__pyx_v_multiple_fields = NULL;
  PyObject *__pyx_v_asset = NULL;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_series = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  PyObject *__pyx_t_18 = NULL;
  __Pyx_RefNannySetupContext("current", 0);

  /* "catalyst/_protocol.pyx":311
 *         returned.
 *         """
 *         multiple_assets = _is_iterable(assets)             # <<<<<<<<<<<<<<
 *         multiple_fields = _is_iterable(fields)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8catalyst_9_protocol__is_iterable(__pyx_v_assets)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_multiple_assets = ((PyBoolObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":312
 *         """
 *         multiple_assets = _is_iterable(assets)
 *         multiple_fields = _is_iterable(fields)             # <<<<<<<<<<<<<<
 * 
 *         # There's some overly verbose code in here, particularly around
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8catalyst_9_protocol__is_iterable(__pyx_v_fields)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_multiple_fields = ((PyBoolObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":323
 *         # verbose, but this is OK for now.
 * 
 *         if not multiple_assets:             # <<<<<<<<<<<<<<
 *             asset = assets
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_multiple_assets)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "catalyst/_protocol.pyx":324
 * 
 *         if not multiple_assets:
 *             asset = assets             # <<<<<<<<<<<<<<
 * 
 *             if not multiple_fields:
 */
    __Pyx_INCREF(__pyx_v_assets);
    __pyx_v_asset = __pyx_v_assets;

    /* "catalyst/_protocol.pyx":326
 *             asset = assets
 * 
 *             if not multiple_fields:             # <<<<<<<<<<<<<<
 *                 field = fields
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_multiple_fields)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 326, __pyx_L1_error)
    __pyx_t_2 = ((!__pyx_t_3) != 0);
    if (__pyx_t_2) {

      /* "catalyst/_protocol.pyx":327
 * 
 *             if not multiple_fields:
 *                 field = fields             # <<<<<<<<<<<<<<
 * 
 *                 # return scalar value
 */
      __Pyx_INCREF(__pyx_v_fields);
      __pyx_v_field = __pyx_v_fields;

      /* "catalyst/_protocol.pyx":330
 * 
 *                 # return scalar value
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     return self.data_portal.get_spot_value(
 *                         asset,
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 330, __pyx_L1_error)
      __pyx_t_3 = ((!__pyx_t_2) != 0);
      if (__pyx_t_3) {

        /* "catalyst/_protocol.pyx":331
 *                 # return scalar value
 *                 if not self._adjust_minutes:
 *                     return self.data_portal.get_spot_value(             # <<<<<<<<<<<<<<
 *                         asset,
 *                         field,
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "catalyst/_protocol.pyx":334
 *                         asset,
 *                         field,
 *                         self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                         self.data_frequency
 *                     )
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "catalyst/_protocol.pyx":335
 *                         field,
 *                         self._get_current_minute(),
 *                         self.data_frequency             # <<<<<<<<<<<<<<
 *                     )
 *                 else:
 */
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_asset, __pyx_v_field, __pyx_t_5, __pyx_v_self->data_frequency};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_asset, __pyx_v_field, __pyx_t_5, __pyx_v_self->data_frequency};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 331, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_v_asset);
          __Pyx_GIVEREF(__pyx_v_asset);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_asset);
          __Pyx_INCREF(__pyx_v_field);
          __Pyx_GIVEREF(__pyx_v_field);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_field);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
          __Pyx_INCREF(__pyx_v_self->data_frequency);
          __Pyx_GIVEREF(__pyx_v_self->data_frequency);
          PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_self->data_frequency);
          __pyx_t_5 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "catalyst/_protocol.pyx":330
 * 
 *                 # return scalar value
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     return self.data_portal.get_spot_value(
 *                         asset,
 */
      }

      /* "catalyst/_protocol.pyx":338
 *                     )
 *                 else:
 *                     return self.data_portal.get_adjusted_value(             # <<<<<<<<<<<<<<
 *                         asset,
 *                         field,
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_adjusted_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "catalyst/_protocol.pyx":341
 *                         asset,
 *                         field,
 *                         self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                         self.simulation_dt_func(),
 *                         self.data_frequency
 */
        __pyx_t_8 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 341, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);

        /* "catalyst/_protocol.pyx":342
 *                         field,
 *                         self._get_current_minute(),
 *                         self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *                         self.data_frequency
 *                     )
 */
        __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
        __pyx_t_6 = __pyx_v_self->simulation_dt_func; __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        if (__pyx_t_9) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 342, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else {
          __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 342, __pyx_L1_error)
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "catalyst/_protocol.pyx":343
 *                         self._get_current_minute(),
 *                         self.simulation_dt_func(),
 *                         self.data_frequency             # <<<<<<<<<<<<<<
 *                     )
 *             else:
 */
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[6] = {__pyx_t_6, __pyx_v_asset, __pyx_v_field, __pyx_t_8, __pyx_t_5, __pyx_v_self->data_frequency};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[6] = {__pyx_t_6, __pyx_v_asset, __pyx_v_field, __pyx_t_8, __pyx_t_5, __pyx_v_self->data_frequency};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        {
          __pyx_t_9 = PyTuple_New(5+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 338, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_v_asset);
          __Pyx_GIVEREF(__pyx_v_asset);
          PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_7, __pyx_v_asset);
          __Pyx_INCREF(__pyx_v_field);
          __Pyx_GIVEREF(__pyx_v_field);
          PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_7, __pyx_v_field);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_7, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_7, __pyx_t_5);
          __Pyx_INCREF(__pyx_v_self->data_frequency);
          __Pyx_GIVEREF(__pyx_v_self->data_frequency);
          PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_7, __pyx_v_self->data_frequency);
          __pyx_t_8 = 0;
          __pyx_t_5 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L0;
      }

      /* "catalyst/_protocol.pyx":326
 *             asset = assets
 * 
 *             if not multiple_fields:             # <<<<<<<<<<<<<<
 *                 field = fields
 * 
 */
    }

    /* "catalyst/_protocol.pyx":348
 *                 # assume fields is iterable
 *                 # return a Series indexed by field
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     return pd.Series(data={
 *                         field: self.data_portal.get_spot_value(
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 348, __pyx_L1_error)
      __pyx_t_2 = ((!__pyx_t_3) != 0);
      if (__pyx_t_2) {

        /* "catalyst/_protocol.pyx":349
 *                 # return a Series indexed by field
 *                 if not self._adjust_minutes:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         field: self.data_portal.get_spot_value(
 *                                     asset,
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Series); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        { /* enter inner scope */
          PyObject *__pyx_7genexpr__pyx_v_field = NULL;
          __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 349, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_9);

          /* "catalyst/_protocol.pyx":356
 *                                     self.data_frequency
 *                                )
 *                         for field in fields             # <<<<<<<<<<<<<<
 *                     }, index=fields, name=assets.symbol)
 *                 else:
 */
          if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
            __pyx_t_5 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 356, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 356, __pyx_L9_error)
          }
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_5))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 356, __pyx_L9_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 356, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 356, __pyx_L9_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 356, __pyx_L9_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              }
            } else {
              __pyx_t_8 = __pyx_t_11(__pyx_t_5);
              if (unlikely(!__pyx_t_8)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 356, __pyx_L9_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_8);
            }
            __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_field, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":350
 *                 if not self._adjust_minutes:
 *                     return pd.Series(data={
 *                         field: self.data_portal.get_spot_value(             # <<<<<<<<<<<<<<
 *                                     asset,
 *                                     field,
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 350, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "catalyst/_protocol.pyx":353
 *                                     asset,
 *                                     field,
 *                                     self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                                     self.data_frequency
 *                                )
 */
            __pyx_t_12 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 353, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "catalyst/_protocol.pyx":354
 *                                     field,
 *                                     self._get_current_minute(),
 *                                     self.data_frequency             # <<<<<<<<<<<<<<
 *                                )
 *                         for field in fields
 */
            __pyx_t_13 = NULL;
            __pyx_t_7 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
                __pyx_t_7 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[5] = {__pyx_t_13, __pyx_v_asset, __pyx_7genexpr__pyx_v_field, __pyx_t_12, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 350, __pyx_L9_error)
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[5] = {__pyx_t_13, __pyx_v_asset, __pyx_7genexpr__pyx_v_field, __pyx_t_12, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 350, __pyx_L9_error)
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            {
              __pyx_t_14 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 350, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_14);
              if (__pyx_t_13) {
                __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13); __pyx_t_13 = NULL;
              }
              __Pyx_INCREF(__pyx_v_asset);
              __Pyx_GIVEREF(__pyx_v_asset);
              PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_7, __pyx_v_asset);
              __Pyx_INCREF(__pyx_7genexpr__pyx_v_field);
              __Pyx_GIVEREF(__pyx_7genexpr__pyx_v_field);
              PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_7, __pyx_7genexpr__pyx_v_field);
              __Pyx_GIVEREF(__pyx_t_12);
              PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_7, __pyx_t_12);
              __Pyx_INCREF(__pyx_v_self->data_frequency);
              __Pyx_GIVEREF(__pyx_v_self->data_frequency);
              PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_7, __pyx_v_self->data_frequency);
              __pyx_t_12 = 0;
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_14, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 350, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_9, (PyObject*)__pyx_7genexpr__pyx_v_field, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 350, __pyx_L9_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":356
 *                                     self.data_frequency
 *                                )
 *                         for field in fields             # <<<<<<<<<<<<<<
 *                     }, index=fields, name=assets.symbol)
 *                 else:
 */
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_7genexpr__pyx_v_field);
          goto __pyx_L12_exit_scope;
          __pyx_L9_error:;
          __Pyx_XDECREF(__pyx_7genexpr__pyx_v_field);
          goto __pyx_L1_error;
          __pyx_L12_exit_scope:;
        } /* exit inner scope */
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_data, __pyx_t_9) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "catalyst/_protocol.pyx":357
 *                                )
 *                         for field in fields
 *                     }, index=fields, name=assets.symbol)             # <<<<<<<<<<<<<<
 *                 else:
 *                     return pd.Series(data={
 */
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_index, __pyx_v_fields) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_assets, __pyx_n_s_symbol); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_name_2, __pyx_t_9) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "catalyst/_protocol.pyx":349
 *                 # return a Series indexed by field
 *                 if not self._adjust_minutes:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         field: self.data_portal.get_spot_value(
 *                                     asset,
 */
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_r = __pyx_t_9;
        __pyx_t_9 = 0;
        goto __pyx_L0;

        /* "catalyst/_protocol.pyx":348
 *                 # assume fields is iterable
 *                 # return a Series indexed by field
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     return pd.Series(data={
 *                         field: self.data_portal.get_spot_value(
 */
      }

      /* "catalyst/_protocol.pyx":359
 *                     }, index=fields, name=assets.symbol)
 *                 else:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         field: self.data_portal.get_adjusted_value(
 *                                     asset,
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Series); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        { /* enter inner scope */
          PyObject *__pyx_8genexpr1__pyx_v_field = NULL;
          __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_4);

          /* "catalyst/_protocol.pyx":367
 *                                     self.data_frequency
 *                                )
 *                         for field in fields             # <<<<<<<<<<<<<<
 *                     }, index=fields, name=assets.symbol)
 *         else:
 */
          if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
            __pyx_t_5 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 367, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 367, __pyx_L15_error)
          }
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_5))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 367, __pyx_L15_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 367, __pyx_L15_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 367, __pyx_L15_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 367, __pyx_L15_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              }
            } else {
              __pyx_t_8 = __pyx_t_11(__pyx_t_5);
              if (unlikely(!__pyx_t_8)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 367, __pyx_L15_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_8);
            }
            __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_field, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":360
 *                 else:
 *                     return pd.Series(data={
 *                         field: self.data_portal.get_adjusted_value(             # <<<<<<<<<<<<<<
 *                                     asset,
 *                                     field,
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_adjusted_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 360, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "catalyst/_protocol.pyx":363
 *                                     asset,
 *                                     field,
 *                                     self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                                     self.simulation_dt_func(),
 *                                     self.data_frequency
 */
            __pyx_t_14 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 363, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "catalyst/_protocol.pyx":364
 *                                     field,
 *                                     self._get_current_minute(),
 *                                     self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *                                     self.data_frequency
 *                                )
 */
            __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
            __pyx_t_13 = __pyx_v_self->simulation_dt_func; __pyx_t_15 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_13);
              if (likely(__pyx_t_15)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                __Pyx_INCREF(__pyx_t_15);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_13, function);
              }
            }
            if (__pyx_t_15) {
              __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_15); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 364, __pyx_L15_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else {
              __pyx_t_12 = __Pyx_PyObject_CallNoArg(__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 364, __pyx_L15_error)
            }
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "catalyst/_protocol.pyx":365
 *                                     self._get_current_minute(),
 *                                     self.simulation_dt_func(),
 *                                     self.data_frequency             # <<<<<<<<<<<<<<
 *                                )
 *                         for field in fields
 */
            __pyx_t_13 = NULL;
            __pyx_t_7 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
                __pyx_t_7 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[6] = {__pyx_t_13, __pyx_v_asset, __pyx_8genexpr1__pyx_v_field, __pyx_t_14, __pyx_t_12, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 360, __pyx_L15_error)
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[6] = {__pyx_t_13, __pyx_v_asset, __pyx_8genexpr1__pyx_v_field, __pyx_t_14, __pyx_t_12, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 360, __pyx_L15_error)
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            {
              __pyx_t_15 = PyTuple_New(5+__pyx_t_7); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 360, __pyx_L15_error)
              __Pyx_GOTREF(__pyx_t_15);
              if (__pyx_t_13) {
                __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_13); __pyx_t_13 = NULL;
              }
              __Pyx_INCREF(__pyx_v_asset);
              __Pyx_GIVEREF(__pyx_v_asset);
              PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_7, __pyx_v_asset);
              __Pyx_INCREF(__pyx_8genexpr1__pyx_v_field);
              __Pyx_GIVEREF(__pyx_8genexpr1__pyx_v_field);
              PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_7, __pyx_8genexpr1__pyx_v_field);
              __Pyx_GIVEREF(__pyx_t_14);
              PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_7, __pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_12);
              PyTuple_SET_ITEM(__pyx_t_15, 3+__pyx_t_7, __pyx_t_12);
              __Pyx_INCREF(__pyx_v_self->data_frequency);
              __Pyx_GIVEREF(__pyx_v_self->data_frequency);
              PyTuple_SET_ITEM(__pyx_t_15, 4+__pyx_t_7, __pyx_v_self->data_frequency);
              __pyx_t_14 = 0;
              __pyx_t_12 = 0;
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 360, __pyx_L15_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_4, (PyObject*)__pyx_8genexpr1__pyx_v_field, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 360, __pyx_L15_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":367
 *                                     self.data_frequency
 *                                )
 *                         for field in fields             # <<<<<<<<<<<<<<
 *                     }, index=fields, name=assets.symbol)
 *         else:
 */
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_field);
          goto __pyx_L18_exit_scope;
          __pyx_L15_error:;
          __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_field);
          goto __pyx_L1_error;
          __pyx_L18_exit_scope:;
        } /* exit inner scope */
        if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_data, __pyx_t_4) < 0) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "catalyst/_protocol.pyx":368
 *                                )
 *                         for field in fields
 *                     }, index=fields, name=assets.symbol)             # <<<<<<<<<<<<<<
 *         else:
 *             if not multiple_fields:
 */
        if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_index, __pyx_v_fields) < 0) __PYX_ERR(0, 359, __pyx_L1_error)
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_assets, __pyx_n_s_symbol); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_name_2, __pyx_t_4) < 0) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "catalyst/_protocol.pyx":359
 *                     }, index=fields, name=assets.symbol)
 *                 else:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         field: self.data_portal.get_adjusted_value(
 *                                     asset,
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L0;
      }
    }

    /* "catalyst/_protocol.pyx":323
 *         # verbose, but this is OK for now.
 * 
 *         if not multiple_assets:             # <<<<<<<<<<<<<<
 *             asset = assets
 * 
 */
  }

  /* "catalyst/_protocol.pyx":370
 *                     }, index=fields, name=assets.symbol)
 *         else:
 *             if not multiple_fields:             # <<<<<<<<<<<<<<
 *                 field = fields
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_multiple_fields)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 370, __pyx_L1_error)
    __pyx_t_3 = ((!__pyx_t_2) != 0);
    if (__pyx_t_3) {

      /* "catalyst/_protocol.pyx":371
 *         else:
 *             if not multiple_fields:
 *                 field = fields             # <<<<<<<<<<<<<<
 * 
 *                 # assume assets is iterable
 */
      __Pyx_INCREF(__pyx_v_fields);
      __pyx_v_field = __pyx_v_fields;

      /* "catalyst/_protocol.pyx":375
 *                 # assume assets is iterable
 *                 # return a Series indexed by asset
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     return pd.Series(data={
 *                         asset: self.data_portal.get_spot_value(
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 375, __pyx_L1_error)
      __pyx_t_2 = ((!__pyx_t_3) != 0);
      if (__pyx_t_2) {

        /* "catalyst/_protocol.pyx":376
 *                 # return a Series indexed by asset
 *                 if not self._adjust_minutes:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         asset: self.data_portal.get_spot_value(
 *                                     asset,
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 376, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Series); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 376, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 376, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        { /* enter inner scope */
          PyObject *__pyx_8genexpr2__pyx_v_asset = NULL;
          __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_1);

          /* "catalyst/_protocol.pyx":383
 *                                     self.data_frequency
 *                                )
 *                         for asset in assets             # <<<<<<<<<<<<<<
 *                         }, index=assets, name=fields)
 *                 else:
 */
          if (likely(PyList_CheckExact(__pyx_v_assets)) || PyTuple_CheckExact(__pyx_v_assets)) {
            __pyx_t_5 = __pyx_v_assets; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_assets); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 383, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 383, __pyx_L23_error)
          }
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_5))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 383, __pyx_L23_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 383, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 383, __pyx_L23_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 383, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              }
            } else {
              __pyx_t_8 = __pyx_t_11(__pyx_t_5);
              if (unlikely(!__pyx_t_8)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 383, __pyx_L23_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_8);
            }
            __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_asset, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":377
 *                 if not self._adjust_minutes:
 *                     return pd.Series(data={
 *                         asset: self.data_portal.get_spot_value(             # <<<<<<<<<<<<<<
 *                                     asset,
 *                                     field,
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "catalyst/_protocol.pyx":380
 *                                     asset,
 *                                     field,
 *                                     self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                                     self.data_frequency
 *                                )
 */
            __pyx_t_15 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 380, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "catalyst/_protocol.pyx":381
 *                                     field,
 *                                     self._get_current_minute(),
 *                                     self.data_frequency             # <<<<<<<<<<<<<<
 *                                )
 *                         for asset in assets
 */
            __pyx_t_12 = NULL;
            __pyx_t_7 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
                __pyx_t_7 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_8genexpr2__pyx_v_asset, __pyx_v_field, __pyx_t_15, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 377, __pyx_L23_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_8genexpr2__pyx_v_asset, __pyx_v_field, __pyx_t_15, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 377, __pyx_L23_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            {
              __pyx_t_14 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 377, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_14);
              if (__pyx_t_12) {
                __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
              }
              __Pyx_INCREF(__pyx_8genexpr2__pyx_v_asset);
              __Pyx_GIVEREF(__pyx_8genexpr2__pyx_v_asset);
              PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_7, __pyx_8genexpr2__pyx_v_asset);
              __Pyx_INCREF(__pyx_v_field);
              __Pyx_GIVEREF(__pyx_v_field);
              PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_7, __pyx_v_field);
              __Pyx_GIVEREF(__pyx_t_15);
              PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_7, __pyx_t_15);
              __Pyx_INCREF(__pyx_v_self->data_frequency);
              __Pyx_GIVEREF(__pyx_v_self->data_frequency);
              PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_7, __pyx_v_self->data_frequency);
              __pyx_t_15 = 0;
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_14, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 377, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_asset, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 377, __pyx_L23_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":383
 *                                     self.data_frequency
 *                                )
 *                         for asset in assets             # <<<<<<<<<<<<<<
 *                         }, index=assets, name=fields)
 *                 else:
 */
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_asset);
          goto __pyx_L26_exit_scope;
          __pyx_L23_error:;
          __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_asset);
          goto __pyx_L1_error;
          __pyx_L26_exit_scope:;
        } /* exit inner scope */
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_data, __pyx_t_1) < 0) __PYX_ERR(0, 376, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "catalyst/_protocol.pyx":384
 *                                )
 *                         for asset in assets
 *                         }, index=assets, name=fields)             # <<<<<<<<<<<<<<
 *                 else:
 *                     return pd.Series(data={
 */
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_index, __pyx_v_assets) < 0) __PYX_ERR(0, 376, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_name_2, __pyx_v_fields) < 0) __PYX_ERR(0, 376, __pyx_L1_error)

        /* "catalyst/_protocol.pyx":376
 *                 # return a Series indexed by asset
 *                 if not self._adjust_minutes:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         asset: self.data_portal.get_spot_value(
 *                                     asset,
 */
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "catalyst/_protocol.pyx":375
 *                 # assume assets is iterable
 *                 # return a Series indexed by asset
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     return pd.Series(data={
 *                         asset: self.data_portal.get_spot_value(
 */
      }

      /* "catalyst/_protocol.pyx":386
 *                         }, index=assets, name=fields)
 *                 else:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         asset: self.data_portal.get_adjusted_value(
 *                                     asset,
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Series); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        { /* enter inner scope */
          PyObject *__pyx_8genexpr3__pyx_v_asset = NULL;
          __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 386, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_9);

          /* "catalyst/_protocol.pyx":394
 *                                     self.data_frequency
 *                                )
 *                         for asset in assets             # <<<<<<<<<<<<<<
 *                         }, index=assets, name=fields)
 * 
 */
          if (likely(PyList_CheckExact(__pyx_v_assets)) || PyTuple_CheckExact(__pyx_v_assets)) {
            __pyx_t_5 = __pyx_v_assets; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_assets); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 394, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 394, __pyx_L29_error)
          }
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_5))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 394, __pyx_L29_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L29_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 394, __pyx_L29_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L29_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              }
            } else {
              __pyx_t_8 = __pyx_t_11(__pyx_t_5);
              if (unlikely(!__pyx_t_8)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 394, __pyx_L29_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_8);
            }
            __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_asset, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":387
 *                 else:
 *                     return pd.Series(data={
 *                         asset: self.data_portal.get_adjusted_value(             # <<<<<<<<<<<<<<
 *                                     asset,
 *                                     field,
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_adjusted_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 387, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "catalyst/_protocol.pyx":390
 *                                     asset,
 *                                     field,
 *                                     self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                                     self.simulation_dt_func(),
 *                                     self.data_frequency
 */
            __pyx_t_14 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 390, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "catalyst/_protocol.pyx":391
 *                                     field,
 *                                     self._get_current_minute(),
 *                                     self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *                                     self.data_frequency
 *                                )
 */
            __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
            __pyx_t_12 = __pyx_v_self->simulation_dt_func; __pyx_t_13 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
              }
            }
            if (__pyx_t_13) {
              __pyx_t_15 = __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 391, __pyx_L29_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            } else {
              __pyx_t_15 = __Pyx_PyObject_CallNoArg(__pyx_t_12); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 391, __pyx_L29_error)
            }
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

            /* "catalyst/_protocol.pyx":392
 *                                     self._get_current_minute(),
 *                                     self.simulation_dt_func(),
 *                                     self.data_frequency             # <<<<<<<<<<<<<<
 *                                )
 *                         for asset in assets
 */
            __pyx_t_12 = NULL;
            __pyx_t_7 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
                __pyx_t_7 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[6] = {__pyx_t_12, __pyx_8genexpr3__pyx_v_asset, __pyx_v_field, __pyx_t_14, __pyx_t_15, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 387, __pyx_L29_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
              PyObject *__pyx_temp[6] = {__pyx_t_12, __pyx_8genexpr3__pyx_v_asset, __pyx_v_field, __pyx_t_14, __pyx_t_15, __pyx_v_self->data_frequency};
              __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 387, __pyx_L29_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            {
              __pyx_t_13 = PyTuple_New(5+__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 387, __pyx_L29_error)
              __Pyx_GOTREF(__pyx_t_13);
              if (__pyx_t_12) {
                __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_12); __pyx_t_12 = NULL;
              }
              __Pyx_INCREF(__pyx_8genexpr3__pyx_v_asset);
              __Pyx_GIVEREF(__pyx_8genexpr3__pyx_v_asset);
              PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_7, __pyx_8genexpr3__pyx_v_asset);
              __Pyx_INCREF(__pyx_v_field);
              __Pyx_GIVEREF(__pyx_v_field);
              PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_7, __pyx_v_field);
              __Pyx_GIVEREF(__pyx_t_14);
              PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_7, __pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_15);
              PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_7, __pyx_t_15);
              __Pyx_INCREF(__pyx_v_self->data_frequency);
              __Pyx_GIVEREF(__pyx_v_self->data_frequency);
              PyTuple_SET_ITEM(__pyx_t_13, 4+__pyx_t_7, __pyx_v_self->data_frequency);
              __pyx_t_14 = 0;
              __pyx_t_15 = 0;
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_13, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 387, __pyx_L29_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_9, (PyObject*)__pyx_8genexpr3__pyx_v_asset, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 387, __pyx_L29_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

            /* "catalyst/_protocol.pyx":394
 *                                     self.data_frequency
 *                                )
 *                         for asset in assets             # <<<<<<<<<<<<<<
 *                         }, index=assets, name=fields)
 * 
 */
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_asset);
          goto __pyx_L32_exit_scope;
          __pyx_L29_error:;
          __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_asset);
          goto __pyx_L1_error;
          __pyx_L32_exit_scope:;
        } /* exit inner scope */
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_data, __pyx_t_9) < 0) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "catalyst/_protocol.pyx":395
 *                                )
 *                         for asset in assets
 *                         }, index=assets, name=fields)             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_index, __pyx_v_assets) < 0) __PYX_ERR(0, 386, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_name_2, __pyx_v_fields) < 0) __PYX_ERR(0, 386, __pyx_L1_error)

        /* "catalyst/_protocol.pyx":386
 *                         }, index=assets, name=fields)
 *                 else:
 *                     return pd.Series(data={             # <<<<<<<<<<<<<<
 *                         asset: self.data_portal.get_adjusted_value(
 *                                     asset,
 */
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_r = __pyx_t_9;
        __pyx_t_9 = 0;
        goto __pyx_L0;
      }

      /* "catalyst/_protocol.pyx":370
 *                     }, index=fields, name=assets.symbol)
 *         else:
 *             if not multiple_fields:             # <<<<<<<<<<<<<<
 *                 field = fields
 * 
 */
    }

    /* "catalyst/_protocol.pyx":399
 *             else:
 *                 # both assets and fields are iterable
 *                 data = {}             # <<<<<<<<<<<<<<
 * 
 *                 if not self._adjust_minutes:
 */
    /*else*/ {
      __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_v_data = ((PyObject*)__pyx_t_9);
      __pyx_t_9 = 0;

      /* "catalyst/_protocol.pyx":401
 *                 data = {}
 * 
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     for field in fields:
 *                         series = pd.Series(data={
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 401, __pyx_L1_error)
      __pyx_t_3 = ((!__pyx_t_2) != 0);
      if (__pyx_t_3) {

        /* "catalyst/_protocol.pyx":402
 * 
 *                 if not self._adjust_minutes:
 *                     for field in fields:             # <<<<<<<<<<<<<<
 *                         series = pd.Series(data={
 *                             asset: self.data_portal.get_spot_value(
 */
        if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
          __pyx_t_9 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_9); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
        } else {
          __pyx_t_10 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 402, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_11 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 402, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_11)) {
            if (likely(PyList_CheckExact(__pyx_t_9))) {
              if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_9)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
              #else
              __pyx_t_1 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            } else {
              if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
              #else
              __pyx_t_1 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            }
          } else {
            __pyx_t_1 = __pyx_t_11(__pyx_t_9);
            if (unlikely(!__pyx_t_1)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 402, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_1);
          }
          __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "catalyst/_protocol.pyx":403
 *                 if not self._adjust_minutes:
 *                     for field in fields:
 *                         series = pd.Series(data={             # <<<<<<<<<<<<<<
 *                             asset: self.data_portal.get_spot_value(
 *                                         asset,
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Series); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 403, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          { /* enter inner scope */
            PyObject *__pyx_8genexpr4__pyx_v_asset = NULL;
            __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 403, __pyx_L38_error)
            __Pyx_GOTREF(__pyx_t_5);

            /* "catalyst/_protocol.pyx":410
 *                                         self.data_frequency
 *                                    )
 *                             for asset in assets             # <<<<<<<<<<<<<<
 *                             }, index=assets, name=field)
 *                         data[field] = series
 */
            if (likely(PyList_CheckExact(__pyx_v_assets)) || PyTuple_CheckExact(__pyx_v_assets)) {
              __pyx_t_8 = __pyx_v_assets; __Pyx_INCREF(__pyx_t_8); __pyx_t_16 = 0;
              __pyx_t_17 = NULL;
            } else {
              __pyx_t_16 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_assets); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 410, __pyx_L38_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_17 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 410, __pyx_L38_error)
            }
            for (;;) {
              if (likely(!__pyx_t_17)) {
                if (likely(PyList_CheckExact(__pyx_t_8))) {
                  if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_8)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_6 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_16); __Pyx_INCREF(__pyx_t_6); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 410, __pyx_L38_error)
                  #else
                  __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L38_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  #endif
                } else {
                  if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_16); __Pyx_INCREF(__pyx_t_6); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 410, __pyx_L38_error)
                  #else
                  __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L38_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  #endif
                }
              } else {
                __pyx_t_6 = __pyx_t_17(__pyx_t_8);
                if (unlikely(!__pyx_t_6)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 410, __pyx_L38_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_6);
              }
              __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_asset, __pyx_t_6);
              __pyx_t_6 = 0;

              /* "catalyst/_protocol.pyx":404
 *                     for field in fields:
 *                         series = pd.Series(data={
 *                             asset: self.data_portal.get_spot_value(             # <<<<<<<<<<<<<<
 *                                         asset,
 *                                         field,
 */
              __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 404, __pyx_L38_error)
              __Pyx_GOTREF(__pyx_t_13);

              /* "catalyst/_protocol.pyx":407
 *                                         asset,
 *                                         field,
 *                                         self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                                         self.data_frequency
 *                                    )
 */
              __pyx_t_15 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 407, __pyx_L38_error)
              __Pyx_GOTREF(__pyx_t_15);

              /* "catalyst/_protocol.pyx":408
 *                                         field,
 *                                         self._get_current_minute(),
 *                                         self.data_frequency             # <<<<<<<<<<<<<<
 *                                    )
 *                             for asset in assets
 */
              __pyx_t_14 = NULL;
              __pyx_t_7 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_13, function);
                  __pyx_t_7 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_13)) {
                PyObject *__pyx_temp[5] = {__pyx_t_14, __pyx_8genexpr4__pyx_v_asset, __pyx_v_field, __pyx_t_15, __pyx_v_self->data_frequency};
                __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 404, __pyx_L38_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
                PyObject *__pyx_temp[5] = {__pyx_t_14, __pyx_8genexpr4__pyx_v_asset, __pyx_v_field, __pyx_t_15, __pyx_v_self->data_frequency};
                __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 404, __pyx_L38_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              } else
              #endif
              {
                __pyx_t_12 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 404, __pyx_L38_error)
                __Pyx_GOTREF(__pyx_t_12);
                if (__pyx_t_14) {
                  __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_14); __pyx_t_14 = NULL;
                }
                __Pyx_INCREF(__pyx_8genexpr4__pyx_v_asset);
                __Pyx_GIVEREF(__pyx_8genexpr4__pyx_v_asset);
                PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_7, __pyx_8genexpr4__pyx_v_asset);
                __Pyx_INCREF(__pyx_v_field);
                __Pyx_GIVEREF(__pyx_v_field);
                PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_7, __pyx_v_field);
                __Pyx_GIVEREF(__pyx_t_15);
                PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_7, __pyx_t_15);
                __Pyx_INCREF(__pyx_v_self->data_frequency);
                __Pyx_GIVEREF(__pyx_v_self->data_frequency);
                PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_7, __pyx_v_self->data_frequency);
                __pyx_t_15 = 0;
                __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_12, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 404, __pyx_L38_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              }
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(PyDict_SetItem(__pyx_t_5, (PyObject*)__pyx_8genexpr4__pyx_v_asset, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 404, __pyx_L38_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

              /* "catalyst/_protocol.pyx":410
 *                                         self.data_frequency
 *                                    )
 *                             for asset in assets             # <<<<<<<<<<<<<<
 *                             }, index=assets, name=field)
 *                         data[field] = series
 */
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_asset);
            goto __pyx_L41_exit_scope;
            __pyx_L38_error:;
            __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_asset);
            goto __pyx_L1_error;
            __pyx_L41_exit_scope:;
          } /* exit inner scope */
          if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_data, __pyx_t_5) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "catalyst/_protocol.pyx":411
 *                                    )
 *                             for asset in assets
 *                             }, index=assets, name=field)             # <<<<<<<<<<<<<<
 *                         data[field] = series
 *                 else:
 */
          if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_index, __pyx_v_assets) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
          if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_name_2, __pyx_v_field) < 0) __PYX_ERR(0, 403, __pyx_L1_error)

          /* "catalyst/_protocol.pyx":403
 *                 if not self._adjust_minutes:
 *                     for field in fields:
 *                         series = pd.Series(data={             # <<<<<<<<<<<<<<
 *                             asset: self.data_portal.get_spot_value(
 *                                         asset,
 */
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 403, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF_SET(__pyx_v_series, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "catalyst/_protocol.pyx":412
 *                             for asset in assets
 *                             }, index=assets, name=field)
 *                         data[field] = series             # <<<<<<<<<<<<<<
 *                 else:
 *                     for field in fields:
 */
          if (unlikely(PyDict_SetItem(__pyx_v_data, __pyx_v_field, __pyx_v_series) < 0)) __PYX_ERR(0, 412, __pyx_L1_error)

          /* "catalyst/_protocol.pyx":402
 * 
 *                 if not self._adjust_minutes:
 *                     for field in fields:             # <<<<<<<<<<<<<<
 *                         series = pd.Series(data={
 *                             asset: self.data_portal.get_spot_value(
 */
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "catalyst/_protocol.pyx":401
 *                 data = {}
 * 
 *                 if not self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     for field in fields:
 *                         series = pd.Series(data={
 */
        goto __pyx_L33;
      }

      /* "catalyst/_protocol.pyx":414
 *                         data[field] = series
 *                 else:
 *                     for field in fields:             # <<<<<<<<<<<<<<
 *                         series = pd.Series(data={
 *                             asset: self.data_portal.get_adjusted_value(
 */
      /*else*/ {
        if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
          __pyx_t_9 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_9); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
        } else {
          __pyx_t_10 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_11 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 414, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_11)) {
            if (likely(PyList_CheckExact(__pyx_t_9))) {
              if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_9)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 414, __pyx_L1_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 414, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 414, __pyx_L1_error)
              #else
              __pyx_t_5 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 414, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_11(__pyx_t_9);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 414, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "catalyst/_protocol.pyx":415
 *                 else:
 *                     for field in fields:
 *                         series = pd.Series(data={             # <<<<<<<<<<<<<<
 *                             asset: self.data_portal.get_adjusted_value(
 *                                         asset,
 */
          __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Series); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          { /* enter inner scope */
            PyObject *__pyx_8genexpr5__pyx_v_asset = NULL;
            __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 415, __pyx_L46_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "catalyst/_protocol.pyx":423
 *                                         self.data_frequency
 *                                    )
 *                             for asset in assets             # <<<<<<<<<<<<<<
 *                             }, index=assets, name=field)
 *                         data[field] = series
 */
            if (likely(PyList_CheckExact(__pyx_v_assets)) || PyTuple_CheckExact(__pyx_v_assets)) {
              __pyx_t_8 = __pyx_v_assets; __Pyx_INCREF(__pyx_t_8); __pyx_t_16 = 0;
              __pyx_t_17 = NULL;
            } else {
              __pyx_t_16 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_assets); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 423, __pyx_L46_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_17 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 423, __pyx_L46_error)
            }
            for (;;) {
              if (likely(!__pyx_t_17)) {
                if (likely(PyList_CheckExact(__pyx_t_8))) {
                  if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_8)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_6 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_16); __Pyx_INCREF(__pyx_t_6); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 423, __pyx_L46_error)
                  #else
                  __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 423, __pyx_L46_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  #endif
                } else {
                  if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_16); __Pyx_INCREF(__pyx_t_6); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 423, __pyx_L46_error)
                  #else
                  __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 423, __pyx_L46_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  #endif
                }
              } else {
                __pyx_t_6 = __pyx_t_17(__pyx_t_8);
                if (unlikely(!__pyx_t_6)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 423, __pyx_L46_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_6);
              }
              __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_asset, __pyx_t_6);
              __pyx_t_6 = 0;

              /* "catalyst/_protocol.pyx":416
 *                     for field in fields:
 *                         series = pd.Series(data={
 *                             asset: self.data_portal.get_adjusted_value(             # <<<<<<<<<<<<<<
 *                                         asset,
 *                                         field,
 */
              __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_adjusted_value); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 416, __pyx_L46_error)
              __Pyx_GOTREF(__pyx_t_13);

              /* "catalyst/_protocol.pyx":419
 *                                         asset,
 *                                         field,
 *                                         self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                                         self.simulation_dt_func(),
 *                                         self.data_frequency
 */
              __pyx_t_12 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 419, __pyx_L46_error)
              __Pyx_GOTREF(__pyx_t_12);

              /* "catalyst/_protocol.pyx":420
 *                                         field,
 *                                         self._get_current_minute(),
 *                                         self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *                                         self.data_frequency
 *                                    )
 */
              __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
              __pyx_t_14 = __pyx_v_self->simulation_dt_func; __pyx_t_18 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_18)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_18);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              if (__pyx_t_18) {
                __pyx_t_15 = __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_18); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 420, __pyx_L46_error)
                __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              } else {
                __pyx_t_15 = __Pyx_PyObject_CallNoArg(__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 420, __pyx_L46_error)
              }
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "catalyst/_protocol.pyx":421
 *                                         self._get_current_minute(),
 *                                         self.simulation_dt_func(),
 *                                         self.data_frequency             # <<<<<<<<<<<<<<
 *                                    )
 *                             for asset in assets
 */
              __pyx_t_14 = NULL;
              __pyx_t_7 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_13, function);
                  __pyx_t_7 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_13)) {
                PyObject *__pyx_temp[6] = {__pyx_t_14, __pyx_8genexpr5__pyx_v_asset, __pyx_v_field, __pyx_t_12, __pyx_t_15, __pyx_v_self->data_frequency};
                __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L46_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
                PyObject *__pyx_temp[6] = {__pyx_t_14, __pyx_8genexpr5__pyx_v_asset, __pyx_v_field, __pyx_t_12, __pyx_t_15, __pyx_v_self->data_frequency};
                __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_7, 5+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L46_error)
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              } else
              #endif
              {
                __pyx_t_18 = PyTuple_New(5+__pyx_t_7); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 416, __pyx_L46_error)
                __Pyx_GOTREF(__pyx_t_18);
                if (__pyx_t_14) {
                  __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_14); __pyx_t_14 = NULL;
                }
                __Pyx_INCREF(__pyx_8genexpr5__pyx_v_asset);
                __Pyx_GIVEREF(__pyx_8genexpr5__pyx_v_asset);
                PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_7, __pyx_8genexpr5__pyx_v_asset);
                __Pyx_INCREF(__pyx_v_field);
                __Pyx_GIVEREF(__pyx_v_field);
                PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_7, __pyx_v_field);
                __Pyx_GIVEREF(__pyx_t_12);
                PyTuple_SET_ITEM(__pyx_t_18, 2+__pyx_t_7, __pyx_t_12);
                __Pyx_GIVEREF(__pyx_t_15);
                PyTuple_SET_ITEM(__pyx_t_18, 3+__pyx_t_7, __pyx_t_15);
                __Pyx_INCREF(__pyx_v_self->data_frequency);
                __Pyx_GIVEREF(__pyx_v_self->data_frequency);
                PyTuple_SET_ITEM(__pyx_t_18, 4+__pyx_t_7, __pyx_v_self->data_frequency);
                __pyx_t_12 = 0;
                __pyx_t_15 = 0;
                __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_18, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L46_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              }
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(PyDict_SetItem(__pyx_t_4, (PyObject*)__pyx_8genexpr5__pyx_v_asset, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 416, __pyx_L46_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

              /* "catalyst/_protocol.pyx":423
 *                                         self.data_frequency
 *                                    )
 *                             for asset in assets             # <<<<<<<<<<<<<<
 *                             }, index=assets, name=field)
 *                         data[field] = series
 */
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_asset);
            goto __pyx_L49_exit_scope;
            __pyx_L46_error:;
            __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_asset);
            goto __pyx_L1_error;
            __pyx_L49_exit_scope:;
          } /* exit inner scope */
          if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_data, __pyx_t_4) < 0) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "catalyst/_protocol.pyx":424
 *                                    )
 *                             for asset in assets
 *                             }, index=assets, name=field)             # <<<<<<<<<<<<<<
 *                         data[field] = series
 * 
 */
          if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_index, __pyx_v_assets) < 0) __PYX_ERR(0, 415, __pyx_L1_error)
          if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_name_2, __pyx_v_field) < 0) __PYX_ERR(0, 415, __pyx_L1_error)

          /* "catalyst/_protocol.pyx":415
 *                 else:
 *                     for field in fields:
 *                         series = pd.Series(data={             # <<<<<<<<<<<<<<
 *                             asset: self.data_portal.get_adjusted_value(
 *                                         asset,
 */
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 415, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF_SET(__pyx_v_series, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "catalyst/_protocol.pyx":425
 *                             for asset in assets
 *                             }, index=assets, name=field)
 *                         data[field] = series             # <<<<<<<<<<<<<<
 * 
 *                 return pd.DataFrame(data)
 */
          if (unlikely(PyDict_SetItem(__pyx_v_data, __pyx_v_field, __pyx_v_series) < 0)) __PYX_ERR(0, 425, __pyx_L1_error)

          /* "catalyst/_protocol.pyx":414
 *                         data[field] = series
 *                 else:
 *                     for field in fields:             # <<<<<<<<<<<<<<
 *                         series = pd.Series(data={
 *                             asset: self.data_portal.get_adjusted_value(
 */
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __pyx_L33:;

      /* "catalyst/_protocol.pyx":427
 *                         data[field] = series
 * 
 *                 return pd.DataFrame(data)             # <<<<<<<<<<<<<<
 * 
 *     @check_parameters(('continuous_future',),
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_data); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_data};
          __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_data};
          __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        {
          __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_data);
          __Pyx_GIVEREF(__pyx_v_data);
          PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_data);
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_9;
      __pyx_t_9 = 0;
      goto __pyx_L0;
    }
  }

  /* "catalyst/_protocol.pyx":254
 *     @check_parameters(('assets', 'fields'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types))
 *     def current(self, assets, fields):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the current value of the given assets for the given fields
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("catalyst._protocol.BarData.current", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_multiple_assets);
  __Pyx_XDECREF((PyObject *)__pyx_v_multiple_fields);
  __Pyx_XDECREF(__pyx_v_asset);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_series);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":431
 *     @check_parameters(('continuous_future',),
 *                       (ContinuousFuture,))
 *     def current_chain(self, continuous_future):             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_current_future_chain(
 *             continuous_future,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_5current_chain(PyObject *__pyx_v_self, PyObject *__pyx_v_continuous_future); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_5current_chain(PyObject *__pyx_v_self, PyObject *__pyx_v_continuous_future) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("current_chain (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_4current_chain(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), ((PyObject *)__pyx_v_continuous_future));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_4current_chain(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_continuous_future) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("current_chain", 0);

  /* "catalyst/_protocol.pyx":432
 *                       (ContinuousFuture,))
 *     def current_chain(self, continuous_future):
 *         return self.data_portal.get_current_future_chain(             # <<<<<<<<<<<<<<
 *             continuous_future,
 *             self.simulation_dt_func())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_current_future_chain); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":434
 *         return self.data_portal.get_current_future_chain(
 *             continuous_future,
 *             self.simulation_dt_func())             # <<<<<<<<<<<<<<
 * 
 *     @check_parameters(('assets',), (Asset,))
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_4 = __pyx_v_self->simulation_dt_func; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (__pyx_t_5) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_continuous_future, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_continuous_future, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_continuous_future);
    __Pyx_GIVEREF(__pyx_v_continuous_future);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_continuous_future);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":431
 *     @check_parameters(('continuous_future',),
 *                       (ContinuousFuture,))
 *     def current_chain(self, continuous_future):             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_current_future_chain(
 *             continuous_future,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.BarData.current_chain", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":437
 * 
 *     @check_parameters(('assets',), (Asset,))
 *     def can_trade(self, assets):             # <<<<<<<<<<<<<<
 *         """
 *         For the given asset or iterable of assets, returns true if all of the
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_7can_trade(PyObject *__pyx_v_self, PyObject *__pyx_v_assets); /*proto*/
static char __pyx_doc_8catalyst_9_protocol_7BarData_6can_trade[] = "\n        For the given asset or iterable of assets, returns true if all of the\n        following are true:\n        1) the asset is alive for the session of the current simulation time\n          (if current simulation time is not a market minute, we use the next\n          session)\n        2) (if we are in minute mode) the asset's exchange is open at the\n          current simulation time or at the simulation calendar's next market\n          minute\n        3) there is a known last price for the asset.\n\n        Notes\n        -----\n        The second condition above warrants some further explanation.\n        - If the asset's exchange calendar is identical to the simulation\n        calendar, then this condition always returns True.\n        - If there are market minutes in the simulation calendar outside of\n        this asset's exchange's trading hours (for example, if the simulation\n        is running on the CME calendar but the asset is MSFT, which trades on\n        the NYSE), during those minutes, this condition will return false\n        (for example, 3:15 am Eastern on a weekday, during which the CME is\n        open but the NYSE is closed).\n\n        Parameters\n        ----------\n        assets: Asset or iterable of assets\n\n        Returns\n        -------\n        can_trade : bool or pd.Series[bool] indexed by asset.\n        ";
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_7can_trade(PyObject *__pyx_v_self, PyObject *__pyx_v_assets) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_trade (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_6can_trade(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), ((PyObject *)__pyx_v_assets));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_6can_trade(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets) {
  PyObject *__pyx_v_dt = NULL;
  PyObject *__pyx_v_adjusted_dt = NULL;
  PyObject *__pyx_v_data_portal = NULL;
  PyObject *__pyx_v_tradeable = NULL;
  PyObject *__pyx_v_asset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  __Pyx_RefNannySetupContext("can_trade", 0);

  /* "catalyst/_protocol.pyx":469
 *         can_trade : bool or pd.Series[bool] indexed by asset.
 *         """
 *         dt = self.simulation_dt_func()             # <<<<<<<<<<<<<<
 * 
 *         if self._adjust_minutes:
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_2 = __pyx_v_self->simulation_dt_func; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_dt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":471
 *         dt = self.simulation_dt_func()
 * 
 *         if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *             adjusted_dt = self._get_current_minute()
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 471, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "catalyst/_protocol.pyx":472
 * 
 *         if self._adjust_minutes:
 *             adjusted_dt = self._get_current_minute()             # <<<<<<<<<<<<<<
 *         else:
 *             adjusted_dt = dt
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_adjusted_dt = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":471
 *         dt = self.simulation_dt_func()
 * 
 *         if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *             adjusted_dt = self._get_current_minute()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "catalyst/_protocol.pyx":474
 *             adjusted_dt = self._get_current_minute()
 *         else:
 *             adjusted_dt = dt             # <<<<<<<<<<<<<<
 * 
 *         data_portal = self.data_portal
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_dt);
    __pyx_v_adjusted_dt = __pyx_v_dt;
  }
  __pyx_L3:;

  /* "catalyst/_protocol.pyx":476
 *             adjusted_dt = dt
 * 
 *         data_portal = self.data_portal             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(assets, Asset):
 */
  __pyx_t_1 = __pyx_v_self->data_portal;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_data_portal = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":478
 *         data_portal = self.data_portal
 * 
 *         if isinstance(assets, Asset):             # <<<<<<<<<<<<<<
 *             return self._can_trade_for_asset(
 *                 assets, dt, adjusted_dt, data_portal
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_Asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_assets, __pyx_t_1); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "catalyst/_protocol.pyx":479
 * 
 *         if isinstance(assets, Asset):
 *             return self._can_trade_for_asset(             # <<<<<<<<<<<<<<
 *                 assets, dt, adjusted_dt, data_portal
 *             )
 */
    __Pyx_XDECREF(__pyx_r);

    /* "catalyst/_protocol.pyx":481
 *             return self._can_trade_for_asset(
 *                 assets, dt, adjusted_dt, data_portal
 *             )             # <<<<<<<<<<<<<<
 *         else:
 *             tradeable = [
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_can_trade_for_asset(__pyx_v_self, __pyx_v_assets, __pyx_v_dt, __pyx_v_adjusted_dt, __pyx_v_data_portal)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":478
 *         data_portal = self.data_portal
 * 
 *         if isinstance(assets, Asset):             # <<<<<<<<<<<<<<
 *             return self._can_trade_for_asset(
 *                 assets, dt, adjusted_dt, data_portal
 */
  }

  /* "catalyst/_protocol.pyx":483
 *             )
 *         else:
 *             tradeable = [             # <<<<<<<<<<<<<<
 *                 self._can_trade_for_asset(
 *                     asset, dt, adjusted_dt, data_portal
 */
  /*else*/ {
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "catalyst/_protocol.pyx":487
 *                     asset, dt, adjusted_dt, data_portal
 *                 )
 *                 for asset in assets             # <<<<<<<<<<<<<<
 *             ]
 *             return pd.Series(data=tradeable, index=assets, dtype=bool)
 */
    if (likely(PyList_CheckExact(__pyx_v_assets)) || PyTuple_CheckExact(__pyx_v_assets)) {
      __pyx_t_2 = __pyx_v_assets; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_assets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 487, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 487, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 487, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 487, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_asset, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "catalyst/_protocol.pyx":484
 *         else:
 *             tradeable = [
 *                 self._can_trade_for_asset(             # <<<<<<<<<<<<<<
 *                     asset, dt, adjusted_dt, data_portal
 *                 )
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_can_trade_for_asset(__pyx_v_self, __pyx_v_asset, __pyx_v_dt, __pyx_v_adjusted_dt, __pyx_v_data_portal)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "catalyst/_protocol.pyx":487
 *                     asset, dt, adjusted_dt, data_portal
 *                 )
 *                 for asset in assets             # <<<<<<<<<<<<<<
 *             ]
 *             return pd.Series(data=tradeable, index=assets, dtype=bool)
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_tradeable = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":489
 *                 for asset in assets
 *             ]
 *             return pd.Series(data=tradeable, index=assets, dtype=bool)             # <<<<<<<<<<<<<<
 * 
 *     cdef bool _can_trade_for_asset(self, asset, dt, adjusted_dt, data_portal):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Series); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_data, __pyx_v_tradeable) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_index, __pyx_v_assets) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, ((PyObject *)__pyx_ptype_7cpython_4bool_bool)) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "catalyst/_protocol.pyx":437
 * 
 *     @check_parameters(('assets',), (Asset,))
 *     def can_trade(self, assets):             # <<<<<<<<<<<<<<
 *         """
 *         For the given asset or iterable of assets, returns true if all of the
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("catalyst._protocol.BarData.can_trade", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dt);
  __Pyx_XDECREF(__pyx_v_adjusted_dt);
  __Pyx_XDECREF(__pyx_v_data_portal);
  __Pyx_XDECREF(__pyx_v_tradeable);
  __Pyx_XDECREF(__pyx_v_asset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":491
 *             return pd.Series(data=tradeable, index=assets, dtype=bool)
 * 
 *     cdef bool _can_trade_for_asset(self, asset, dt, adjusted_dt, data_portal):             # <<<<<<<<<<<<<<
 *         cdef object session_label
 *         cdef object dt_to_use_for_exchange_check,
 */

static PyBoolObject *__pyx_f_8catalyst_9_protocol_7BarData__can_trade_for_asset(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset, PyObject *__pyx_v_dt, PyObject *__pyx_v_adjusted_dt, PyObject *__pyx_v_data_portal) {
  PyObject *__pyx_v_session_label = 0;
  PyObject *__pyx_v_dt_to_use_for_exchange_check = 0;
  PyBoolObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("_can_trade_for_asset", 0);

  /* "catalyst/_protocol.pyx":495
 *         cdef object dt_to_use_for_exchange_check,
 * 
 *         if self._is_restricted(asset, adjusted_dt):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __Pyx_INCREF(__pyx_v_self->_is_restricted);
  __pyx_t_2 = __pyx_v_self->_is_restricted; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_asset, __pyx_v_adjusted_dt};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_asset, __pyx_v_adjusted_dt};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_asset);
    __Pyx_GIVEREF(__pyx_v_asset);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_asset);
    __Pyx_INCREF(__pyx_v_adjusted_dt);
    __Pyx_GIVEREF(__pyx_v_adjusted_dt);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_adjusted_dt);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "catalyst/_protocol.pyx":496
 * 
 *         if self._is_restricted(asset, adjusted_dt):
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         session_label = self._trading_calendar.minute_to_session_label(dt)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(Py_False);
    __pyx_r = ((PyBoolObject *)Py_False);
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":495
 *         cdef object dt_to_use_for_exchange_check,
 * 
 *         if self._is_restricted(asset, adjusted_dt):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "catalyst/_protocol.pyx":498
 *             return False
 * 
 *         session_label = self._trading_calendar.minute_to_session_label(dt)             # <<<<<<<<<<<<<<
 * 
 *         if not asset.is_alive_for_session(session_label):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_trading_calendar, __pyx_n_s_minute_to_session_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_dt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_dt};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_dt};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_dt);
      __Pyx_GIVEREF(__pyx_v_dt);
      PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_dt);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_session_label = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":500
 *         session_label = self._trading_calendar.minute_to_session_label(dt)
 * 
 *         if not asset.is_alive_for_session(session_label):             # <<<<<<<<<<<<<<
 *             # asset isn't alive
 *             return False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_asset, __pyx_n_s_is_alive_for_session); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_session_label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_session_label};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_session_label};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_session_label);
      __Pyx_GIVEREF(__pyx_v_session_label);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_session_label);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (__pyx_t_7) {

    /* "catalyst/_protocol.pyx":502
 *         if not asset.is_alive_for_session(session_label):
 *             # asset isn't alive
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         if asset.auto_close_date and session_label >= asset.auto_close_date:
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(Py_False);
    __pyx_r = ((PyBoolObject *)Py_False);
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":500
 *         session_label = self._trading_calendar.minute_to_session_label(dt)
 * 
 *         if not asset.is_alive_for_session(session_label):             # <<<<<<<<<<<<<<
 *             # asset isn't alive
 *             return False
 */
  }

  /* "catalyst/_protocol.pyx":504
 *             return False
 * 
 *         if asset.auto_close_date and session_label >= asset.auto_close_date:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_asset, __pyx_n_s_auto_close_date); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_7 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_asset, __pyx_n_s_auto_close_date); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_session_label, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_7) {

    /* "catalyst/_protocol.pyx":505
 * 
 *         if asset.auto_close_date and session_label >= asset.auto_close_date:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         if not self._daily_mode:
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(Py_False);
    __pyx_r = ((PyBoolObject *)Py_False);
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":504
 *             return False
 * 
 *         if asset.auto_close_date and session_label >= asset.auto_close_date:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "catalyst/_protocol.pyx":507
 *             return False
 * 
 *         if not self._daily_mode:             # <<<<<<<<<<<<<<
 *             # Find the next market minute for this calendar, and check if this
 *             # asset's exchange is open at that minute.
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_daily_mode)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 507, __pyx_L1_error)
  __pyx_t_6 = ((!__pyx_t_7) != 0);
  if (__pyx_t_6) {

    /* "catalyst/_protocol.pyx":510
 *             # Find the next market minute for this calendar, and check if this
 *             # asset's exchange is open at that minute.
 *             if self._trading_calendar.is_open_on_minute(dt):             # <<<<<<<<<<<<<<
 *                 dt_to_use_for_exchange_check = dt
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_trading_calendar, __pyx_n_s_is_open_on_minute); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_dt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_dt};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_dt};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_dt);
        __Pyx_GIVEREF(__pyx_v_dt);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_dt);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_6) {

      /* "catalyst/_protocol.pyx":511
 *             # asset's exchange is open at that minute.
 *             if self._trading_calendar.is_open_on_minute(dt):
 *                 dt_to_use_for_exchange_check = dt             # <<<<<<<<<<<<<<
 *             else:
 *                 dt_to_use_for_exchange_check = \
 */
      __Pyx_INCREF(__pyx_v_dt);
      __pyx_v_dt_to_use_for_exchange_check = __pyx_v_dt;

      /* "catalyst/_protocol.pyx":510
 *             # Find the next market minute for this calendar, and check if this
 *             # asset's exchange is open at that minute.
 *             if self._trading_calendar.is_open_on_minute(dt):             # <<<<<<<<<<<<<<
 *                 dt_to_use_for_exchange_check = dt
 *             else:
 */
      goto __pyx_L9;
    }

    /* "catalyst/_protocol.pyx":514
 *             else:
 *                 dt_to_use_for_exchange_check = \
 *                     self._trading_calendar.next_open(dt)             # <<<<<<<<<<<<<<
 * 
 *             if not asset.is_exchange_open(dt_to_use_for_exchange_check):
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_trading_calendar, __pyx_n_s_next_open); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      if (!__pyx_t_3) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_dt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_dt};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_dt};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          __Pyx_INCREF(__pyx_v_dt);
          __Pyx_GIVEREF(__pyx_v_dt);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_dt);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_dt_to_use_for_exchange_check = __pyx_t_2;
      __pyx_t_2 = 0;
    }
    __pyx_L9:;

    /* "catalyst/_protocol.pyx":516
 *                     self._trading_calendar.next_open(dt)
 * 
 *             if not asset.is_exchange_open(dt_to_use_for_exchange_check):             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_asset, __pyx_n_s_is_exchange_open); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_dt_to_use_for_exchange_check); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_dt_to_use_for_exchange_check};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_dt_to_use_for_exchange_check};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_dt_to_use_for_exchange_check);
        __Pyx_GIVEREF(__pyx_v_dt_to_use_for_exchange_check);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_dt_to_use_for_exchange_check);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = ((!__pyx_t_6) != 0);
    if (__pyx_t_7) {

      /* "catalyst/_protocol.pyx":517
 * 
 *             if not asset.is_exchange_open(dt_to_use_for_exchange_check):
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         # is there a last price?
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(Py_False);
      __pyx_r = ((PyBoolObject *)Py_False);
      goto __pyx_L0;

      /* "catalyst/_protocol.pyx":516
 *                     self._trading_calendar.next_open(dt)
 * 
 *             if not asset.is_exchange_open(dt_to_use_for_exchange_check):             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "catalyst/_protocol.pyx":507
 *             return False
 * 
 *         if not self._daily_mode:             # <<<<<<<<<<<<<<
 *             # Find the next market minute for this calendar, and check if this
 *             # asset's exchange is open at that minute.
 */
  }

  /* "catalyst/_protocol.pyx":520
 * 
 *         # is there a last price?
 *         return not np.isnan(             # <<<<<<<<<<<<<<
 *             data_portal.get_spot_value(
 *                 asset, "price", adjusted_dt, self.data_frequency
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_isnan); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":521
 *         # is there a last price?
 *         return not np.isnan(
 *             data_portal.get_spot_value(             # <<<<<<<<<<<<<<
 *                 asset, "price", adjusted_dt, self.data_frequency
 *             )
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "catalyst/_protocol.pyx":522
 *         return not np.isnan(
 *             data_portal.get_spot_value(
 *                 asset, "price", adjusted_dt, self.data_frequency             # <<<<<<<<<<<<<<
 *             )
 *         )
 */
  __pyx_t_8 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_asset, __pyx_n_s_price, __pyx_v_adjusted_dt, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_asset, __pyx_n_s_price, __pyx_v_adjusted_dt, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_v_asset);
    __Pyx_GIVEREF(__pyx_v_asset);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_4, __pyx_v_asset);
    __Pyx_INCREF(__pyx_n_s_price);
    __Pyx_GIVEREF(__pyx_n_s_price);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_4, __pyx_n_s_price);
    __Pyx_INCREF(__pyx_v_adjusted_dt);
    __Pyx_GIVEREF(__pyx_v_adjusted_dt);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_4, __pyx_v_adjusted_dt);
    __Pyx_INCREF(__pyx_v_self->data_frequency);
    __Pyx_GIVEREF(__pyx_v_self->data_frequency);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_4, __pyx_v_self->data_frequency);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "catalyst/_protocol.pyx":520
 * 
 *         # is there a last price?
 *         return not np.isnan(             # <<<<<<<<<<<<<<
 *             data_portal.get_spot_value(
 *                 asset, "price", adjusted_dt, self.data_frequency
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!__pyx_t_7)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7cpython_4bool_bool)))) __PYX_ERR(0, 520, __pyx_L1_error)
  __pyx_r = ((PyBoolObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":491
 *             return pd.Series(data=tradeable, index=assets, dtype=bool)
 * 
 *     cdef bool _can_trade_for_asset(self, asset, dt, adjusted_dt, data_portal):             # <<<<<<<<<<<<<<
 *         cdef object session_label
 *         cdef object dt_to_use_for_exchange_check,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("catalyst._protocol.BarData._can_trade_for_asset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_session_label);
  __Pyx_XDECREF(__pyx_v_dt_to_use_for_exchange_check);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":527
 * 
 *     @check_parameters(('assets',), (Asset,))
 *     def is_stale(self, assets):             # <<<<<<<<<<<<<<
 *         """
 *         For the given asset or iterable of assets, returns true if the asset
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_9is_stale(PyObject *__pyx_v_self, PyObject *__pyx_v_assets); /*proto*/
static char __pyx_doc_8catalyst_9_protocol_7BarData_8is_stale[] = "\n        For the given asset or iterable of assets, returns true if the asset\n        is alive and there is no trade data for the current simulation time.\n\n        If the asset has never traded, returns False.\n\n        If the current simulation time is not a valid market time, we use the\n        current time to check if the asset is alive, but we use the last\n        market minute/day for the trade data check.\n\n        Parameters\n        ----------\n        assets: Asset or iterable of assets\n\n        Returns\n        -------\n        boolean or Series of booleans, indexed by asset.\n        ";
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_9is_stale(PyObject *__pyx_v_self, PyObject *__pyx_v_assets) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_stale (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_8is_stale(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), ((PyObject *)__pyx_v_assets));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_8is_stale(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets) {
  PyObject *__pyx_v_dt = NULL;
  PyObject *__pyx_v_adjusted_dt = NULL;
  PyObject *__pyx_v_data_portal = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("is_stale", 0);

  /* "catalyst/_protocol.pyx":546
 *         boolean or Series of booleans, indexed by asset.
 *         """
 *         dt = self.simulation_dt_func()             # <<<<<<<<<<<<<<
 *         if self._adjust_minutes:
 *             adjusted_dt = self._get_current_minute()
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_2 = __pyx_v_self->simulation_dt_func; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_dt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":547
 *         """
 *         dt = self.simulation_dt_func()
 *         if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *             adjusted_dt = self._get_current_minute()
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 547, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "catalyst/_protocol.pyx":548
 *         dt = self.simulation_dt_func()
 *         if self._adjust_minutes:
 *             adjusted_dt = self._get_current_minute()             # <<<<<<<<<<<<<<
 *         else:
 *             adjusted_dt = dt
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_adjusted_dt = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":547
 *         """
 *         dt = self.simulation_dt_func()
 *         if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *             adjusted_dt = self._get_current_minute()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "catalyst/_protocol.pyx":550
 *             adjusted_dt = self._get_current_minute()
 *         else:
 *             adjusted_dt = dt             # <<<<<<<<<<<<<<
 * 
 *         data_portal = self.data_portal
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_dt);
    __pyx_v_adjusted_dt = __pyx_v_dt;
  }
  __pyx_L3:;

  /* "catalyst/_protocol.pyx":552
 *             adjusted_dt = dt
 * 
 *         data_portal = self.data_portal             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(assets, Asset):
 */
  __pyx_t_1 = __pyx_v_self->data_portal;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_data_portal = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":554
 *         data_portal = self.data_portal
 * 
 *         if isinstance(assets, Asset):             # <<<<<<<<<<<<<<
 *             return self._is_stale_for_asset(
 *                 assets, dt, adjusted_dt, data_portal
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_Asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_assets, __pyx_t_1); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "catalyst/_protocol.pyx":555
 * 
 *         if isinstance(assets, Asset):
 *             return self._is_stale_for_asset(             # <<<<<<<<<<<<<<
 *                 assets, dt, adjusted_dt, data_portal
 *             )
 */
    __Pyx_XDECREF(__pyx_r);

    /* "catalyst/_protocol.pyx":557
 *             return self._is_stale_for_asset(
 *                 assets, dt, adjusted_dt, data_portal
 *             )             # <<<<<<<<<<<<<<
 *         else:
 *             return pd.Series(data={
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_is_stale_for_asset(__pyx_v_self, __pyx_v_assets, __pyx_v_dt, __pyx_v_adjusted_dt, __pyx_v_data_portal)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":554
 *         data_portal = self.data_portal
 * 
 *         if isinstance(assets, Asset):             # <<<<<<<<<<<<<<
 *             return self._is_stale_for_asset(
 *                 assets, dt, adjusted_dt, data_portal
 */
  }

  /* "catalyst/_protocol.pyx":559
 *             )
 *         else:
 *             return pd.Series(data={             # <<<<<<<<<<<<<<
 *                 asset: self._is_stale_for_asset(
 *                     asset, dt, adjusted_dt, data_portal
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Series); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    { /* enter inner scope */
      PyObject *__pyx_8genexpr6__pyx_v_asset = NULL;
      __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "catalyst/_protocol.pyx":563
 *                     asset, dt, adjusted_dt, data_portal
 *                 )
 *                 for asset in assets             # <<<<<<<<<<<<<<
 *             })
 * 
 */
      if (likely(PyList_CheckExact(__pyx_v_assets)) || PyTuple_CheckExact(__pyx_v_assets)) {
        __pyx_t_6 = __pyx_v_assets; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_assets); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 563, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 563, __pyx_L7_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 563, __pyx_L7_error)
            #else
            __pyx_t_9 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 563, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 563, __pyx_L7_error)
            #else
            __pyx_t_9 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 563, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          }
        } else {
          __pyx_t_9 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_9)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 563, __pyx_L7_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_asset, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "catalyst/_protocol.pyx":560
 *         else:
 *             return pd.Series(data={
 *                 asset: self._is_stale_for_asset(             # <<<<<<<<<<<<<<
 *                     asset, dt, adjusted_dt, data_portal
 *                 )
 */
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_is_stale_for_asset(__pyx_v_self, __pyx_8genexpr6__pyx_v_asset, __pyx_v_dt, __pyx_v_adjusted_dt, __pyx_v_data_portal)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 560, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (unlikely(PyDict_SetItem(__pyx_t_3, (PyObject*)__pyx_8genexpr6__pyx_v_asset, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 560, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "catalyst/_protocol.pyx":563
 *                     asset, dt, adjusted_dt, data_portal
 *                 )
 *                 for asset in assets             # <<<<<<<<<<<<<<
 *             })
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_asset);
      goto __pyx_L10_exit_scope;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_asset);
      goto __pyx_L1_error;
      __pyx_L10_exit_scope:;
    } /* exit inner scope */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_data, __pyx_t_3) < 0) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "catalyst/_protocol.pyx":559
 *             )
 *         else:
 *             return pd.Series(data={             # <<<<<<<<<<<<<<
 *                 asset: self._is_stale_for_asset(
 *                     asset, dt, adjusted_dt, data_portal
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "catalyst/_protocol.pyx":527
 * 
 *     @check_parameters(('assets',), (Asset,))
 *     def is_stale(self, assets):             # <<<<<<<<<<<<<<
 *         """
 *         For the given asset or iterable of assets, returns true if the asset
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("catalyst._protocol.BarData.is_stale", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dt);
  __Pyx_XDECREF(__pyx_v_adjusted_dt);
  __Pyx_XDECREF(__pyx_v_data_portal);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":566
 *             })
 * 
 *     cdef bool _is_stale_for_asset(self, asset, dt, adjusted_dt, data_portal):             # <<<<<<<<<<<<<<
 *         session_label = normalize_date(dt) # FIXME
 * 
 */

static PyBoolObject *__pyx_f_8catalyst_9_protocol_7BarData__is_stale_for_asset(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset, PyObject *__pyx_v_dt, PyObject *__pyx_v_adjusted_dt, PyObject *__pyx_v_data_portal) {
  PyObject *__pyx_v_session_label = NULL;
  PyObject *__pyx_v_current_volume = NULL;
  PyObject *__pyx_v_last_traded_dt = NULL;
  PyBoolObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_is_stale_for_asset", 0);

  /* "catalyst/_protocol.pyx":567
 * 
 *     cdef bool _is_stale_for_asset(self, asset, dt, adjusted_dt, data_portal):
 *         session_label = normalize_date(dt) # FIXME             # <<<<<<<<<<<<<<
 * 
 *         if not asset.is_alive_for_session(session_label):
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_normalize_date); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_dt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_dt};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_dt};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_dt);
      __Pyx_GIVEREF(__pyx_v_dt);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_dt);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_session_label = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":569
 *         session_label = normalize_date(dt) # FIXME
 * 
 *         if not asset.is_alive_for_session(session_label):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_asset, __pyx_n_s_is_alive_for_session); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_session_label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_session_label};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_session_label};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_session_label);
      __Pyx_GIVEREF(__pyx_v_session_label);
      PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_session_label);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "catalyst/_protocol.pyx":570
 * 
 *         if not asset.is_alive_for_session(session_label):
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         current_volume = data_portal.get_spot_value(
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(Py_False);
    __pyx_r = ((PyBoolObject *)Py_False);
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":569
 *         session_label = normalize_date(dt) # FIXME
 * 
 *         if not asset.is_alive_for_session(session_label):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "catalyst/_protocol.pyx":572
 *             return False
 * 
 *         current_volume = data_portal.get_spot_value(             # <<<<<<<<<<<<<<
 *             asset, "volume",  adjusted_dt, self.data_frequency
 *         )
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":573
 * 
 *         current_volume = data_portal.get_spot_value(
 *             asset, "volume",  adjusted_dt, self.data_frequency             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_3 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_asset, __pyx_n_s_volume, __pyx_v_adjusted_dt, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_asset, __pyx_n_s_volume, __pyx_v_adjusted_dt, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_asset);
    __Pyx_GIVEREF(__pyx_v_asset);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_v_asset);
    __Pyx_INCREF(__pyx_n_s_volume);
    __Pyx_GIVEREF(__pyx_n_s_volume);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_n_s_volume);
    __Pyx_INCREF(__pyx_v_adjusted_dt);
    __Pyx_GIVEREF(__pyx_v_adjusted_dt);
    PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_7, __pyx_v_adjusted_dt);
    __Pyx_INCREF(__pyx_v_self->data_frequency);
    __Pyx_GIVEREF(__pyx_v_self->data_frequency);
    PyTuple_SET_ITEM(__pyx_t_4, 3+__pyx_t_7, __pyx_v_self->data_frequency);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_current_volume = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":576
 *         )
 * 
 *         if current_volume > 0:             # <<<<<<<<<<<<<<
 *             # found a current value, so we know this asset is not stale.
 *             return False
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_current_volume, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "catalyst/_protocol.pyx":578
 *         if current_volume > 0:
 *             # found a current value, so we know this asset is not stale.
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             # we need to distinguish between if this asset has ever traded
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(Py_False);
    __pyx_r = ((PyBoolObject *)Py_False);
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":576
 *         )
 * 
 *         if current_volume > 0:             # <<<<<<<<<<<<<<
 *             # found a current value, so we know this asset is not stale.
 *             return False
 */
  }

  /* "catalyst/_protocol.pyx":582
 *             # we need to distinguish between if this asset has ever traded
 *             # (stale = True) or has never traded (stale = False)
 *             last_traded_dt = \             # <<<<<<<<<<<<<<
 *                 data_portal.get_spot_value(asset, "last_traded", adjusted_dt,
 *                                            self.data_frequency)
 */
  /*else*/ {

    /* "catalyst/_protocol.pyx":583
 *             # (stale = True) or has never traded (stale = False)
 *             last_traded_dt = \
 *                 data_portal.get_spot_value(asset, "last_traded", adjusted_dt,             # <<<<<<<<<<<<<<
 *                                            self.data_frequency)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "catalyst/_protocol.pyx":584
 *             last_traded_dt = \
 *                 data_portal.get_spot_value(asset, "last_traded", adjusted_dt,
 *                                            self.data_frequency)             # <<<<<<<<<<<<<<
 * 
 *             return not (last_traded_dt is pd.NaT)
 */
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_asset, __pyx_n_s_last_traded, __pyx_v_adjusted_dt, __pyx_v_self->data_frequency};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_asset, __pyx_n_s_last_traded, __pyx_v_adjusted_dt, __pyx_v_self->data_frequency};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_asset);
      __Pyx_GIVEREF(__pyx_v_asset);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_v_asset);
      __Pyx_INCREF(__pyx_n_s_last_traded);
      __Pyx_GIVEREF(__pyx_n_s_last_traded);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_n_s_last_traded);
      __Pyx_INCREF(__pyx_v_adjusted_dt);
      __Pyx_GIVEREF(__pyx_v_adjusted_dt);
      PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_7, __pyx_v_adjusted_dt);
      __Pyx_INCREF(__pyx_v_self->data_frequency);
      __Pyx_GIVEREF(__pyx_v_self->data_frequency);
      PyTuple_SET_ITEM(__pyx_t_3, 3+__pyx_t_7, __pyx_v_self->data_frequency);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_last_traded_dt = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":586
 *                                            self.data_frequency)
 * 
 *             return not (last_traded_dt is pd.NaT)             # <<<<<<<<<<<<<<
 * 
 *     @check_parameters(('assets', 'fields', 'bar_count',
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_NaT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_v_last_traded_dt != __pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7cpython_4bool_bool)))) __PYX_ERR(0, 586, __pyx_L1_error)
    __pyx_r = ((PyBoolObject *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "catalyst/_protocol.pyx":566
 *             })
 * 
 *     cdef bool _is_stale_for_asset(self, asset, dt, adjusted_dt, data_portal):             # <<<<<<<<<<<<<<
 *         session_label = normalize_date(dt) # FIXME
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.BarData._is_stale_for_asset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_session_label);
  __Pyx_XDECREF(__pyx_v_current_volume);
  __Pyx_XDECREF(__pyx_v_last_traded_dt);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":593
 *                        int,
 *                        string_types))
 *     def history(self, assets, fields, bar_count, frequency):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a window of data for the given assets and fields.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_11history(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8catalyst_9_protocol_7BarData_10history[] = "\n        Returns a window of data for the given assets and fields.\n\n        This data is adjusted for splits, dividends, and mergers as of the\n        current algorithm time.\n\n        The semantics of missing data are identical to the ones described in\n        the notes for `get_spot_value`.\n\n        Parameters\n        ----------\n        assets: Asset or iterable of Asset\n\n        fields: string or iterable of string.  Valid values are \"open\", \"high\",\n            \"low\", \"close\", \"volume\", \"price\", and \"last_traded\".\n\n        bar_count: integer number of bars of trade data\n\n        frequency: string. \"1m\" for minutely data or \"1d\" for daily date\n\n        Returns\n        -------\n        history : Series or DataFrame or Panel\n            Return type depends on the dimensionality of the 'assets' and\n            'fields' parameters.\n\n            If single asset and field are passed in, the returned Series is\n            indexed by dt.\n\n            If multiple assets and single field are passed in, the returned\n            DataFrame is indexed by dt, and has assets as columns.\n\n            If a single asset and multiple fields are passed in, the returned\n            DataFrame is indexed by dt, and has fields as columns.\n\n            If multiple assets and multiple fields are passed in, the returned\n            Panel is indexed by field, has dt as the major axis, and assets\n            as the minor axis.\n\n        Notes\n        -----\n        If the current simulation time is not a valid market time, we use the\n        last market close instead.\n        ";
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_11history(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_assets = 0;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_bar_count = 0;
  PyObject *__pyx_v_frequency = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("history (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_assets,&__pyx_n_s_fields,&__pyx_n_s_bar_count,&__pyx_n_s_frequency,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_assets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_fields)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("history", 1, 4, 4, 1); __PYX_ERR(0, 593, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bar_count)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("history", 1, 4, 4, 2); __PYX_ERR(0, 593, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_frequency)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("history", 1, 4, 4, 3); __PYX_ERR(0, 593, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "history") < 0)) __PYX_ERR(0, 593, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_assets = values[0];
    __pyx_v_fields = values[1];
    __pyx_v_bar_count = values[2];
    __pyx_v_frequency = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("history", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 593, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("catalyst._protocol.BarData.history", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_10history(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), __pyx_v_assets, __pyx_v_fields, __pyx_v_bar_count, __pyx_v_frequency);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_10history(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_assets, PyObject *__pyx_v_fields, PyObject *__pyx_v_bar_count, PyObject *__pyx_v_frequency) {
  int __pyx_v_single_asset;
  PyObject *__pyx_v_asset_list = NULL;
  PyObject *__pyx_v_df = NULL;
  PyObject *__pyx_v_adjs = NULL;
  PyObject *__pyx_v_df_dict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("history", 0);

  /* "catalyst/_protocol.pyx":638
 *         last market close instead.
 *         """
 *         if isinstance(fields, string_types):             # <<<<<<<<<<<<<<
 *             single_asset = isinstance(assets, PricingDataAssociable)
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_string_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_fields, __pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "catalyst/_protocol.pyx":639
 *         """
 *         if isinstance(fields, string_types):
 *             single_asset = isinstance(assets, PricingDataAssociable)             # <<<<<<<<<<<<<<
 * 
 *             if single_asset:
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_PricingDataAssociable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_assets, __pyx_t_1); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_single_asset = __pyx_t_3;

    /* "catalyst/_protocol.pyx":641
 *             single_asset = isinstance(assets, PricingDataAssociable)
 * 
 *             if single_asset:             # <<<<<<<<<<<<<<
 *                 asset_list = [assets]
 *             else:
 */
    __pyx_t_3 = (__pyx_v_single_asset != 0);
    if (__pyx_t_3) {

      /* "catalyst/_protocol.pyx":642
 * 
 *             if single_asset:
 *                 asset_list = [assets]             # <<<<<<<<<<<<<<
 *             else:
 *                 asset_list = assets
 */
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_assets);
      __Pyx_GIVEREF(__pyx_v_assets);
      PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_assets);
      __pyx_v_asset_list = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "catalyst/_protocol.pyx":641
 *             single_asset = isinstance(assets, PricingDataAssociable)
 * 
 *             if single_asset:             # <<<<<<<<<<<<<<
 *                 asset_list = [assets]
 *             else:
 */
      goto __pyx_L4;
    }

    /* "catalyst/_protocol.pyx":644
 *                 asset_list = [assets]
 *             else:
 *                 asset_list = assets             # <<<<<<<<<<<<<<
 * 
 *             df = self.data_portal.get_history_window(
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v_assets);
      __pyx_v_asset_list = __pyx_v_assets;
    }
    __pyx_L4:;

    /* "catalyst/_protocol.pyx":646
 *                 asset_list = assets
 * 
 *             df = self.data_portal.get_history_window(             # <<<<<<<<<<<<<<
 *                 asset_list,
 *                 self._get_current_minute(),
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_history_window); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "catalyst/_protocol.pyx":648
 *             df = self.data_portal.get_history_window(
 *                 asset_list,
 *                 self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                 bar_count,
 *                 frequency,
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "catalyst/_protocol.pyx":652
 *                 frequency,
 *                 fields,
 *                 self.data_frequency,             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[7] = {__pyx_t_6, __pyx_v_asset_list, __pyx_t_5, __pyx_v_bar_count, __pyx_v_frequency, __pyx_v_fields, __pyx_v_self->data_frequency};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[7] = {__pyx_t_6, __pyx_v_asset_list, __pyx_t_5, __pyx_v_bar_count, __pyx_v_frequency, __pyx_v_fields, __pyx_v_self->data_frequency};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(6+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_asset_list);
      __Pyx_GIVEREF(__pyx_v_asset_list);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_asset_list);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_bar_count);
      __Pyx_GIVEREF(__pyx_v_bar_count);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_v_bar_count);
      __Pyx_INCREF(__pyx_v_frequency);
      __Pyx_GIVEREF(__pyx_v_frequency);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_frequency);
      __Pyx_INCREF(__pyx_v_fields);
      __Pyx_GIVEREF(__pyx_v_fields);
      PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_7, __pyx_v_fields);
      __Pyx_INCREF(__pyx_v_self->data_frequency);
      __Pyx_GIVEREF(__pyx_v_self->data_frequency);
      PyTuple_SET_ITEM(__pyx_t_8, 5+__pyx_t_7, __pyx_v_self->data_frequency);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_df = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":655
 *             )
 * 
 *             if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                 adjs = self.data_portal.get_adjustments(
 *                     assets,
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 655, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "catalyst/_protocol.pyx":656
 * 
 *             if self._adjust_minutes:
 *                 adjs = self.data_portal.get_adjustments(             # <<<<<<<<<<<<<<
 *                     assets,
 *                     fields,
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_adjustments); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 656, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "catalyst/_protocol.pyx":659
 *                     assets,
 *                     fields,
 *                     self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                     self.simulation_dt_func()
 *                 )
 */
      __pyx_t_8 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 659, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "catalyst/_protocol.pyx":660
 *                     fields,
 *                     self._get_current_minute(),
 *                     self.simulation_dt_func()             # <<<<<<<<<<<<<<
 *                 )
 * 
 */
      __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
      __pyx_t_6 = __pyx_v_self->simulation_dt_func; __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      if (__pyx_t_9) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 660, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 660, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_assets, __pyx_v_fields, __pyx_t_8, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_assets, __pyx_v_fields, __pyx_t_8, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_assets);
        __Pyx_GIVEREF(__pyx_v_assets);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_7, __pyx_v_assets);
        __Pyx_INCREF(__pyx_v_fields);
        __Pyx_GIVEREF(__pyx_v_fields);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_7, __pyx_v_fields);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_7, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_7, __pyx_t_5);
        __pyx_t_8 = 0;
        __pyx_t_5 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_adjs = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "catalyst/_protocol.pyx":663
 *                 )
 * 
 *                 df = df * adjs             # <<<<<<<<<<<<<<
 * 
 *             if single_asset:
 */
      __pyx_t_1 = PyNumber_Multiply(__pyx_v_df, __pyx_v_adjs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_df, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "catalyst/_protocol.pyx":655
 *             )
 * 
 *             if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                 adjs = self.data_portal.get_adjustments(
 *                     assets,
 */
    }

    /* "catalyst/_protocol.pyx":665
 *                 df = df * adjs
 * 
 *             if single_asset:             # <<<<<<<<<<<<<<
 *                 # single asset, single field, return a series.
 *                 return df[assets]
 */
    __pyx_t_3 = (__pyx_v_single_asset != 0);
    if (__pyx_t_3) {

      /* "catalyst/_protocol.pyx":667
 *             if single_asset:
 *                 # single asset, single field, return a series.
 *                 return df[assets]             # <<<<<<<<<<<<<<
 *             else:
 *                 # multiple assets, single field, return a dataframe whose
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = PyObject_GetItem(__pyx_v_df, __pyx_v_assets); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "catalyst/_protocol.pyx":665
 *                 df = df * adjs
 * 
 *             if single_asset:             # <<<<<<<<<<<<<<
 *                 # single asset, single field, return a series.
 *                 return df[assets]
 */
    }

    /* "catalyst/_protocol.pyx":671
 *                 # multiple assets, single field, return a dataframe whose
 *                 # columns are the assets, indexed by dt.
 *                 return df             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(assets, PricingDataAssociable):
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_df);
      __pyx_r = __pyx_v_df;
      goto __pyx_L0;
    }

    /* "catalyst/_protocol.pyx":638
 *         last market close instead.
 *         """
 *         if isinstance(fields, string_types):             # <<<<<<<<<<<<<<
 *             single_asset = isinstance(assets, PricingDataAssociable)
 * 
 */
  }

  /* "catalyst/_protocol.pyx":673
 *                 return df
 *         else:
 *             if isinstance(assets, PricingDataAssociable):             # <<<<<<<<<<<<<<
 *                 # one asset, multiple fields. for now, just make multiple
 *                 # history calls, one per field, then stitch together the
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_PricingDataAssociable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_assets, __pyx_t_1); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {

      /* "catalyst/_protocol.pyx":678
 *                 # results. this can definitely be optimized!
 * 
 *                 df_dict = {             # <<<<<<<<<<<<<<
 *                     field: self.data_portal.get_history_window(
 *                         [assets],
 */
      { /* enter inner scope */
        PyObject *__pyx_8genexpr7__pyx_v_field = NULL;
        __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "catalyst/_protocol.pyx":686
 *                         field,
 *                         self.data_frequency,
 *                     )[assets] for field in fields             # <<<<<<<<<<<<<<
 *                 }
 * 
 */
        if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
          __pyx_t_4 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_4); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
        } else {
          __pyx_t_10 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 686, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 686, __pyx_L10_error)
        }
        for (;;) {
          if (likely(!__pyx_t_11)) {
            if (likely(PyList_CheckExact(__pyx_t_4))) {
              if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_9 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_9); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 686, __pyx_L10_error)
              #else
              __pyx_t_9 = PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 686, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_9);
              #endif
            } else {
              if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_9); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 686, __pyx_L10_error)
              #else
              __pyx_t_9 = PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 686, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_9);
              #endif
            }
          } else {
            __pyx_t_9 = __pyx_t_11(__pyx_t_4);
            if (unlikely(!__pyx_t_9)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 686, __pyx_L10_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_9);
          }
          __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_field, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "catalyst/_protocol.pyx":679
 * 
 *                 df_dict = {
 *                     field: self.data_portal.get_history_window(             # <<<<<<<<<<<<<<
 *                         [assets],
 *                         self._get_current_minute(),
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_history_window); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 679, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "catalyst/_protocol.pyx":680
 *                 df_dict = {
 *                     field: self.data_portal.get_history_window(
 *                         [assets],             # <<<<<<<<<<<<<<
 *                         self._get_current_minute(),
 *                         bar_count,
 */
          __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 680, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_v_assets);
          __Pyx_GIVEREF(__pyx_v_assets);
          PyList_SET_ITEM(__pyx_t_8, 0, __pyx_v_assets);

          /* "catalyst/_protocol.pyx":681
 *                     field: self.data_portal.get_history_window(
 *                         [assets],
 *                         self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                         bar_count,
 *                         frequency,
 */
          __pyx_t_6 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 681, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_6);

          /* "catalyst/_protocol.pyx":685
 *                         frequency,
 *                         field,
 *                         self.data_frequency,             # <<<<<<<<<<<<<<
 *                     )[assets] for field in fields
 *                 }
 */
          __pyx_t_12 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[7] = {__pyx_t_12, __pyx_t_8, __pyx_t_6, __pyx_v_bar_count, __pyx_v_frequency, __pyx_8genexpr7__pyx_v_field, __pyx_v_self->data_frequency};
            __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 679, __pyx_L10_error)
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[7] = {__pyx_t_12, __pyx_t_8, __pyx_t_6, __pyx_v_bar_count, __pyx_v_frequency, __pyx_8genexpr7__pyx_v_field, __pyx_v_self->data_frequency};
            __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 679, __pyx_L10_error)
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(6+__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 679, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_12) {
              __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_12); __pyx_t_12 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_7, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_7, __pyx_t_6);
            __Pyx_INCREF(__pyx_v_bar_count);
            __Pyx_GIVEREF(__pyx_v_bar_count);
            PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_7, __pyx_v_bar_count);
            __Pyx_INCREF(__pyx_v_frequency);
            __Pyx_GIVEREF(__pyx_v_frequency);
            PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_7, __pyx_v_frequency);
            __Pyx_INCREF(__pyx_8genexpr7__pyx_v_field);
            __Pyx_GIVEREF(__pyx_8genexpr7__pyx_v_field);
            PyTuple_SET_ITEM(__pyx_t_13, 4+__pyx_t_7, __pyx_8genexpr7__pyx_v_field);
            __Pyx_INCREF(__pyx_v_self->data_frequency);
            __Pyx_GIVEREF(__pyx_v_self->data_frequency);
            PyTuple_SET_ITEM(__pyx_t_13, 5+__pyx_t_7, __pyx_v_self->data_frequency);
            __pyx_t_8 = 0;
            __pyx_t_6 = 0;
            __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 679, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "catalyst/_protocol.pyx":686
 *                         field,
 *                         self.data_frequency,
 *                     )[assets] for field in fields             # <<<<<<<<<<<<<<
 *                 }
 * 
 */
          __pyx_t_5 = PyObject_GetItem(__pyx_t_9, __pyx_v_assets); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 686, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr7__pyx_v_field, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 679, __pyx_L10_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_field);
        goto __pyx_L13_exit_scope;
        __pyx_L10_error:;
        __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_field);
        goto __pyx_L1_error;
        __pyx_L13_exit_scope:;
      } /* exit inner scope */
      __pyx_v_df_dict = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "catalyst/_protocol.pyx":689
 *                 }
 * 
 *                 if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     adjs = {
 *                         field: self.data_portal.get_adjustments(
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 689, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "catalyst/_protocol.pyx":690
 * 
 *                 if self._adjust_minutes:
 *                     adjs = {             # <<<<<<<<<<<<<<
 *                         field: self.data_portal.get_adjustments(
 *                             assets,
 */
        { /* enter inner scope */
          PyObject *__pyx_8genexpr8__pyx_v_field = NULL;
          __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_1);

          /* "catalyst/_protocol.pyx":696
 *                             self._get_current_minute(),
 *                             self.simulation_dt_func()
 *                         )[0] for field in fields             # <<<<<<<<<<<<<<
 *                     }
 * 
 */
          if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
            __pyx_t_4 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_4); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 696, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_11 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 696, __pyx_L17_error)
          }
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 696, __pyx_L17_error)
                #else
                __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L17_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 696, __pyx_L17_error)
                #else
                __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L17_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
              }
            } else {
              __pyx_t_5 = __pyx_t_11(__pyx_t_4);
              if (unlikely(!__pyx_t_5)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 696, __pyx_L17_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_5);
            }
            __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_field, __pyx_t_5);
            __pyx_t_5 = 0;

            /* "catalyst/_protocol.pyx":691
 *                 if self._adjust_minutes:
 *                     adjs = {
 *                         field: self.data_portal.get_adjustments(             # <<<<<<<<<<<<<<
 *                             assets,
 *                             field,
 */
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_adjustments); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 691, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "catalyst/_protocol.pyx":694
 *                             assets,
 *                             field,
 *                             self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                             self.simulation_dt_func()
 *                         )[0] for field in fields
 */
            __pyx_t_13 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 694, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "catalyst/_protocol.pyx":695
 *                             field,
 *                             self._get_current_minute(),
 *                             self.simulation_dt_func()             # <<<<<<<<<<<<<<
 *                         )[0] for field in fields
 *                     }
 */
            __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
            __pyx_t_8 = __pyx_v_self->simulation_dt_func; __pyx_t_12 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_8);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_8, function);
              }
            }
            if (__pyx_t_12) {
              __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 695, __pyx_L17_error)
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else {
              __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 695, __pyx_L17_error)
            }
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_8 = NULL;
            __pyx_t_7 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_9, function);
                __pyx_t_7 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_assets, __pyx_8genexpr8__pyx_v_field, __pyx_t_13, __pyx_t_6};
              __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 691, __pyx_L17_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_assets, __pyx_8genexpr8__pyx_v_field, __pyx_t_13, __pyx_t_6};
              __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 691, __pyx_L17_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            } else
            #endif
            {
              __pyx_t_12 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 691, __pyx_L17_error)
              __Pyx_GOTREF(__pyx_t_12);
              if (__pyx_t_8) {
                __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_8); __pyx_t_8 = NULL;
              }
              __Pyx_INCREF(__pyx_v_assets);
              __Pyx_GIVEREF(__pyx_v_assets);
              PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_7, __pyx_v_assets);
              __Pyx_INCREF(__pyx_8genexpr8__pyx_v_field);
              __Pyx_GIVEREF(__pyx_8genexpr8__pyx_v_field);
              PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_7, __pyx_8genexpr8__pyx_v_field);
              __Pyx_GIVEREF(__pyx_t_13);
              PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_7, __pyx_t_13);
              __Pyx_GIVEREF(__pyx_t_6);
              PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_7, __pyx_t_6);
              __pyx_t_13 = 0;
              __pyx_t_6 = 0;
              __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 691, __pyx_L17_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

            /* "catalyst/_protocol.pyx":696
 *                             self._get_current_minute(),
 *                             self.simulation_dt_func()
 *                         )[0] for field in fields             # <<<<<<<<<<<<<<
 *                     }
 * 
 */
            __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 696, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr8__pyx_v_field, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 691, __pyx_L17_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_field);
          goto __pyx_L20_exit_scope;
          __pyx_L17_error:;
          __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_field);
          goto __pyx_L1_error;
          __pyx_L20_exit_scope:;
        } /* exit inner scope */
        __pyx_v_adjs = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "catalyst/_protocol.pyx":699
 *                     }
 * 
 *                     df_dict = {field: df * adjs[field]             # <<<<<<<<<<<<<<
 *                                for field, df in iteritems(df_dict)}
 * 
 */
        { /* enter inner scope */
          PyObject *__pyx_8genexpr9__pyx_v_field = NULL;
          PyObject *__pyx_8genexpr9__pyx_v_df = NULL;
          __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_1);

          /* "catalyst/_protocol.pyx":700
 * 
 *                     df_dict = {field: df * adjs[field]
 *                                for field, df in iteritems(df_dict)}             # <<<<<<<<<<<<<<
 * 
 *                 # returned dataframe whose columns are the fields, indexed by
 */
          __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_iteritems); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 700, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
            }
          }
          if (!__pyx_t_5) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_df_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_df_dict};
              __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L23_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_4);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_df_dict};
              __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L23_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_4);
            } else
            #endif
            {
              __pyx_t_12 = PyTuple_New(1+1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 700, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_INCREF(__pyx_v_df_dict);
              __Pyx_GIVEREF(__pyx_v_df_dict);
              PyTuple_SET_ITEM(__pyx_t_12, 0+1, __pyx_v_df_dict);
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
            __pyx_t_9 = __pyx_t_4; __Pyx_INCREF(__pyx_t_9); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 700, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_11 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 700, __pyx_L23_error)
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_9))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_9)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_4 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 700, __pyx_L23_error)
                #else
                __pyx_t_4 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_4);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 700, __pyx_L23_error)
                #else
                __pyx_t_4 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_4);
                #endif
              }
            } else {
              __pyx_t_4 = __pyx_t_11(__pyx_t_9);
              if (unlikely(!__pyx_t_4)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 700, __pyx_L23_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_4);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
              PyObject* sequence = __pyx_t_4;
              #if !CYTHON_COMPILING_IN_PYPY
              Py_ssize_t size = Py_SIZE(sequence);
              #else
              Py_ssize_t size = PySequence_Size(sequence);
              #endif
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 700, __pyx_L23_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_12 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_12 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_5);
              #else
              __pyx_t_12 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 700, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 700, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 700, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_14 = Py_TYPE(__pyx_t_6)->tp_iternext;
              index = 0; __pyx_t_12 = __pyx_t_14(__pyx_t_6); if (unlikely(!__pyx_t_12)) goto __pyx_L26_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_12);
              index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L26_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_5);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_6), 2) < 0) __PYX_ERR(0, 700, __pyx_L23_error)
              __pyx_t_14 = NULL;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              goto __pyx_L27_unpacking_done;
              __pyx_L26_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_14 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 700, __pyx_L23_error)
              __pyx_L27_unpacking_done:;
            }
            __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_field, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_df, __pyx_t_5);
            __pyx_t_5 = 0;

            /* "catalyst/_protocol.pyx":699
 *                     }
 * 
 *                     df_dict = {field: df * adjs[field]             # <<<<<<<<<<<<<<
 *                                for field, df in iteritems(df_dict)}
 * 
 */
            __pyx_t_4 = PyObject_GetItem(__pyx_v_adjs, __pyx_8genexpr9__pyx_v_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 699, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = PyNumber_Multiply(__pyx_8genexpr9__pyx_v_df, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 699, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr9__pyx_v_field, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 699, __pyx_L23_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "catalyst/_protocol.pyx":700
 * 
 *                     df_dict = {field: df * adjs[field]
 *                                for field, df in iteritems(df_dict)}             # <<<<<<<<<<<<<<
 * 
 *                 # returned dataframe whose columns are the fields, indexed by
 */
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_field);
          __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_df);
          goto __pyx_L28_exit_scope;
          __pyx_L23_error:;
          __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_field);
          __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_df);
          goto __pyx_L1_error;
          __pyx_L28_exit_scope:;
        } /* exit inner scope */
        __Pyx_DECREF_SET(__pyx_v_df_dict, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "catalyst/_protocol.pyx":689
 *                 }
 * 
 *                 if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     adjs = {
 *                         field: self.data_portal.get_adjustments(
 */
      }

      /* "catalyst/_protocol.pyx":704
 *                 # returned dataframe whose columns are the fields, indexed by
 *                 # dt.
 *                 return pd.DataFrame(df_dict)             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_9) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_df_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_v_df_dict};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_v_df_dict};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 704, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_9); __pyx_t_9 = NULL;
          __Pyx_INCREF(__pyx_v_df_dict);
          __Pyx_GIVEREF(__pyx_v_df_dict);
          PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_df_dict);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "catalyst/_protocol.pyx":673
 *                 return df
 *         else:
 *             if isinstance(assets, PricingDataAssociable):             # <<<<<<<<<<<<<<
 *                 # one asset, multiple fields. for now, just make multiple
 *                 # history calls, one per field, then stitch together the
 */
    }

    /* "catalyst/_protocol.pyx":707
 * 
 *             else:
 *                 df_dict = {             # <<<<<<<<<<<<<<
 *                     field: self.data_portal.get_history_window(
 *                         assets,
 */
    /*else*/ {
      { /* enter inner scope */
        PyObject *__pyx_9genexpr10__pyx_v_field = NULL;
        __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "catalyst/_protocol.pyx":715
 *                         field,
 *                         self.data_frequency,
 *                     ) for field in fields             # <<<<<<<<<<<<<<
 *                 }
 * 
 */
        if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
          __pyx_t_5 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
        } else {
          __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 715, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 715, __pyx_L31_error)
        }
        for (;;) {
          if (likely(!__pyx_t_11)) {
            if (likely(PyList_CheckExact(__pyx_t_5))) {
              if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 715, __pyx_L31_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L31_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            } else {
              if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 715, __pyx_L31_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L31_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            }
          } else {
            __pyx_t_4 = __pyx_t_11(__pyx_t_5);
            if (unlikely(!__pyx_t_4)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 715, __pyx_L31_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_field, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "catalyst/_protocol.pyx":708
 *             else:
 *                 df_dict = {
 *                     field: self.data_portal.get_history_window(             # <<<<<<<<<<<<<<
 *                         assets,
 *                         self._get_current_minute(),
 */
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_history_window); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 708, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_9);

          /* "catalyst/_protocol.pyx":710
 *                     field: self.data_portal.get_history_window(
 *                         assets,
 *                         self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                         bar_count,
 *                         frequency,
 */
          __pyx_t_12 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 710, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_12);

          /* "catalyst/_protocol.pyx":714
 *                         frequency,
 *                         field,
 *                         self.data_frequency,             # <<<<<<<<<<<<<<
 *                     ) for field in fields
 *                 }
 */
          __pyx_t_6 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_9)) {
            PyObject *__pyx_temp[7] = {__pyx_t_6, __pyx_v_assets, __pyx_t_12, __pyx_v_bar_count, __pyx_v_frequency, __pyx_9genexpr10__pyx_v_field, __pyx_v_self->data_frequency};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
            PyObject *__pyx_temp[7] = {__pyx_t_6, __pyx_v_assets, __pyx_t_12, __pyx_v_bar_count, __pyx_v_frequency, __pyx_9genexpr10__pyx_v_field, __pyx_v_self->data_frequency};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L31_error)
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(6+__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 708, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_6) {
              __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_6); __pyx_t_6 = NULL;
            }
            __Pyx_INCREF(__pyx_v_assets);
            __Pyx_GIVEREF(__pyx_v_assets);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_7, __pyx_v_assets);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_7, __pyx_t_12);
            __Pyx_INCREF(__pyx_v_bar_count);
            __Pyx_GIVEREF(__pyx_v_bar_count);
            PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_7, __pyx_v_bar_count);
            __Pyx_INCREF(__pyx_v_frequency);
            __Pyx_GIVEREF(__pyx_v_frequency);
            PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_7, __pyx_v_frequency);
            __Pyx_INCREF(__pyx_9genexpr10__pyx_v_field);
            __Pyx_GIVEREF(__pyx_9genexpr10__pyx_v_field);
            PyTuple_SET_ITEM(__pyx_t_13, 4+__pyx_t_7, __pyx_9genexpr10__pyx_v_field);
            __Pyx_INCREF(__pyx_v_self->data_frequency);
            __Pyx_GIVEREF(__pyx_v_self->data_frequency);
            PyTuple_SET_ITEM(__pyx_t_13, 5+__pyx_t_7, __pyx_v_self->data_frequency);
            __pyx_t_12 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr10__pyx_v_field, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 708, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "catalyst/_protocol.pyx":715
 *                         field,
 *                         self.data_frequency,
 *                     ) for field in fields             # <<<<<<<<<<<<<<
 *                 }
 * 
 */
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_field);
        goto __pyx_L34_exit_scope;
        __pyx_L31_error:;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_field);
        goto __pyx_L1_error;
        __pyx_L34_exit_scope:;
      } /* exit inner scope */
      __pyx_v_df_dict = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "catalyst/_protocol.pyx":718
 *                 }
 * 
 *                 if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     adjs = {
 *                         field: self.data_portal.get_adjustments(
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_adjust_minutes)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 718, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "catalyst/_protocol.pyx":719
 * 
 *                 if self._adjust_minutes:
 *                     adjs = {             # <<<<<<<<<<<<<<
 *                         field: self.data_portal.get_adjustments(
 *                             assets,
 */
        { /* enter inner scope */
          PyObject *__pyx_9genexpr11__pyx_v_field = NULL;
          __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L38_error)
          __Pyx_GOTREF(__pyx_t_1);

          /* "catalyst/_protocol.pyx":725
 *                             self._get_current_minute(),
 *                             self.simulation_dt_func()
 *                         ) for field in fields             # <<<<<<<<<<<<<<
 *                     }
 * 
 */
          if (likely(PyList_CheckExact(__pyx_v_fields)) || PyTuple_CheckExact(__pyx_v_fields)) {
            __pyx_t_5 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 725, __pyx_L38_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 725, __pyx_L38_error)
          }
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_5))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 725, __pyx_L38_error)
                #else
                __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 725, __pyx_L38_error)
                __Pyx_GOTREF(__pyx_t_4);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 725, __pyx_L38_error)
                #else
                __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 725, __pyx_L38_error)
                __Pyx_GOTREF(__pyx_t_4);
                #endif
              }
            } else {
              __pyx_t_4 = __pyx_t_11(__pyx_t_5);
              if (unlikely(!__pyx_t_4)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 725, __pyx_L38_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_field, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "catalyst/_protocol.pyx":720
 *                 if self._adjust_minutes:
 *                     adjs = {
 *                         field: self.data_portal.get_adjustments(             # <<<<<<<<<<<<<<
 *                             assets,
 *                             field,
 */
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_adjustments); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 720, __pyx_L38_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "catalyst/_protocol.pyx":723
 *                             assets,
 *                             field,
 *                             self._get_current_minute(),             # <<<<<<<<<<<<<<
 *                             self.simulation_dt_func()
 *                         ) for field in fields
 */
            __pyx_t_13 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_current_minute(__pyx_v_self); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 723, __pyx_L38_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "catalyst/_protocol.pyx":724
 *                             field,
 *                             self._get_current_minute(),
 *                             self.simulation_dt_func()             # <<<<<<<<<<<<<<
 *                         ) for field in fields
 *                     }
 */
            __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
            __pyx_t_6 = __pyx_v_self->simulation_dt_func; __pyx_t_8 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
              }
            }
            if (__pyx_t_8) {
              __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 724, __pyx_L38_error)
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            } else {
              __pyx_t_12 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 724, __pyx_L38_error)
            }
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = NULL;
            __pyx_t_7 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_9);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_9, function);
                __pyx_t_7 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_assets, __pyx_9genexpr11__pyx_v_field, __pyx_t_13, __pyx_t_12};
              __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 720, __pyx_L38_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_assets, __pyx_9genexpr11__pyx_v_field, __pyx_t_13, __pyx_t_12};
              __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 720, __pyx_L38_error)
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            {
              __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 720, __pyx_L38_error)
              __Pyx_GOTREF(__pyx_t_8);
              if (__pyx_t_6) {
                __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
              }
              __Pyx_INCREF(__pyx_v_assets);
              __Pyx_GIVEREF(__pyx_v_assets);
              PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_assets);
              __Pyx_INCREF(__pyx_9genexpr11__pyx_v_field);
              __Pyx_GIVEREF(__pyx_9genexpr11__pyx_v_field);
              PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_9genexpr11__pyx_v_field);
              __Pyx_GIVEREF(__pyx_t_13);
              PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_13);
              __Pyx_GIVEREF(__pyx_t_12);
              PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_t_12);
              __pyx_t_13 = 0;
              __pyx_t_12 = 0;
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 720, __pyx_L38_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            }
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr11__pyx_v_field, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 720, __pyx_L38_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "catalyst/_protocol.pyx":725
 *                             self._get_current_minute(),
 *                             self.simulation_dt_func()
 *                         ) for field in fields             # <<<<<<<<<<<<<<
 *                     }
 * 
 */
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_field);
          goto __pyx_L41_exit_scope;
          __pyx_L38_error:;
          __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_field);
          goto __pyx_L1_error;
          __pyx_L41_exit_scope:;
        } /* exit inner scope */
        __pyx_v_adjs = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "catalyst/_protocol.pyx":728
 *                     }
 * 
 *                     df_dict = {field: df * adjs[field]             # <<<<<<<<<<<<<<
 *                                for field, df in iteritems(df_dict)}
 * 
 */
        { /* enter inner scope */
          PyObject *__pyx_9genexpr12__pyx_v_field = NULL;
          PyObject *__pyx_9genexpr12__pyx_v_df = NULL;
          __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_1);

          /* "catalyst/_protocol.pyx":729
 * 
 *                     df_dict = {field: df * adjs[field]
 *                                for field, df in iteritems(df_dict)}             # <<<<<<<<<<<<<<
 * 
 *                 # returned panel has:
 */
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_iteritems); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 729, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_9 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_9) {
            __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_df_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 729, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_5);
          } else {
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_v_df_dict};
              __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 729, __pyx_L44_error)
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_GOTREF(__pyx_t_5);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_v_df_dict};
              __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 729, __pyx_L44_error)
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_GOTREF(__pyx_t_5);
            } else
            #endif
            {
              __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 729, __pyx_L44_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9); __pyx_t_9 = NULL;
              __Pyx_INCREF(__pyx_v_df_dict);
              __Pyx_GIVEREF(__pyx_v_df_dict);
              PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_v_df_dict);
              __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 729, __pyx_L44_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            }
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
            __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 729, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_11 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 729, __pyx_L44_error)
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 729, __pyx_L44_error)
                #else
                __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 729, __pyx_L44_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 729, __pyx_L44_error)
                #else
                __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 729, __pyx_L44_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
              }
            } else {
              __pyx_t_5 = __pyx_t_11(__pyx_t_4);
              if (unlikely(!__pyx_t_5)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 729, __pyx_L44_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_5);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
              PyObject* sequence = __pyx_t_5;
              #if !CYTHON_COMPILING_IN_PYPY
              Py_ssize_t size = Py_SIZE(sequence);
              #else
              Py_ssize_t size = PySequence_Size(sequence);
              #endif
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 729, __pyx_L44_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_9);
              #else
              __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 729, __pyx_L44_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 729, __pyx_L44_error)
              __Pyx_GOTREF(__pyx_t_9);
              #endif
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_12 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 729, __pyx_L44_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_14 = Py_TYPE(__pyx_t_12)->tp_iternext;
              index = 0; __pyx_t_8 = __pyx_t_14(__pyx_t_12); if (unlikely(!__pyx_t_8)) goto __pyx_L47_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_8);
              index = 1; __pyx_t_9 = __pyx_t_14(__pyx_t_12); if (unlikely(!__pyx_t_9)) goto __pyx_L47_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_9);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_12), 2) < 0) __PYX_ERR(0, 729, __pyx_L44_error)
              __pyx_t_14 = NULL;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              goto __pyx_L48_unpacking_done;
              __pyx_L47_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_14 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 729, __pyx_L44_error)
              __pyx_L48_unpacking_done:;
            }
            __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_field, __pyx_t_8);
            __pyx_t_8 = 0;
            __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_df, __pyx_t_9);
            __pyx_t_9 = 0;

            /* "catalyst/_protocol.pyx":728
 *                     }
 * 
 *                     df_dict = {field: df * adjs[field]             # <<<<<<<<<<<<<<
 *                                for field, df in iteritems(df_dict)}
 * 
 */
            __pyx_t_5 = PyObject_GetItem(__pyx_v_adjs, __pyx_9genexpr12__pyx_v_field); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 728, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_9 = PyNumber_Multiply(__pyx_9genexpr12__pyx_v_df, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 728, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr12__pyx_v_field, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 728, __pyx_L44_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

            /* "catalyst/_protocol.pyx":729
 * 
 *                     df_dict = {field: df * adjs[field]
 *                                for field, df in iteritems(df_dict)}             # <<<<<<<<<<<<<<
 * 
 *                 # returned panel has:
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_field);
          __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_df);
          goto __pyx_L49_exit_scope;
          __pyx_L44_error:;
          __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_field);
          __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_df);
          goto __pyx_L1_error;
          __pyx_L49_exit_scope:;
        } /* exit inner scope */
        __Pyx_DECREF_SET(__pyx_v_df_dict, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "catalyst/_protocol.pyx":718
 *                 }
 * 
 *                 if self._adjust_minutes:             # <<<<<<<<<<<<<<
 *                     adjs = {
 *                         field: self.data_portal.get_adjustments(
 */
      }

      /* "catalyst/_protocol.pyx":735
 *                 # major axis: dt
 *                 # minor axis: assets
 *                 return pd.Panel(df_dict)             # <<<<<<<<<<<<<<
 * 
 *     property current_dt:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_pd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 735, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Panel); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 735, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_df_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_df_dict};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_df_dict};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 735, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_df_dict);
          __Pyx_GIVEREF(__pyx_v_df_dict);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_df_dict);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }

  /* "catalyst/_protocol.pyx":593
 *                        int,
 *                        string_types))
 *     def history(self, assets, fields, bar_count, frequency):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a window of data for the given assets and fields.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("catalyst._protocol.BarData.history", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_asset_list);
  __Pyx_XDECREF(__pyx_v_df);
  __Pyx_XDECREF(__pyx_v_adjs);
  __Pyx_XDECREF(__pyx_v_df_dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":738
 * 
 *     property current_dt:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.simulation_dt_func()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_10current_dt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_10current_dt_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_10current_dt___get__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_10current_dt___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":739
 *     property current_dt:
 *         def __get__(self):
 *             return self.simulation_dt_func()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_2 = __pyx_v_self->simulation_dt_func; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":738
 * 
 *     property current_dt:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.simulation_dt_func()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("catalyst._protocol.BarData.current_dt.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":742
 * 
 *     @property
 *     def fetcher_assets(self):             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_fetcher_assets(self.simulation_dt_func())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_14fetcher_assets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_14fetcher_assets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_14fetcher_assets___get__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_14fetcher_assets___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":743
 *     @property
 *     def fetcher_assets(self):
 *         return self.data_portal.get_fetcher_assets(self.simulation_dt_func())             # <<<<<<<<<<<<<<
 * 
 *     property _handle_non_market_minutes:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_fetcher_assets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_4 = __pyx_v_self->simulation_dt_func; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (__pyx_t_5) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 743, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 743, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":742
 * 
 *     @property
 *     def fetcher_assets(self):             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_fetcher_assets(self.simulation_dt_func())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.BarData.fetcher_assets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":746
 * 
 *     property _handle_non_market_minutes:
 *         def __set__(self, val):             # <<<<<<<<<<<<<<
 *             self._adjust_minutes = val
 * 
 */

/* Python wrapper */
static int __pyx_pw_8catalyst_9_protocol_7BarData_26_handle_non_market_minutes_1__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_8catalyst_9_protocol_7BarData_26_handle_non_market_minutes_1__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_26_handle_non_market_minutes___set__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8catalyst_9_protocol_7BarData_26_handle_non_market_minutes___set__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "catalyst/_protocol.pyx":747
 *     property _handle_non_market_minutes:
 *         def __set__(self, val):
 *             self._adjust_minutes = val             # <<<<<<<<<<<<<<
 * 
 *     property current_session:
 */
  if (!(likely(((__pyx_v_val) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_val, __pyx_ptype_7cpython_4bool_bool))))) __PYX_ERR(0, 747, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_val;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_adjust_minutes);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_adjust_minutes));
  __pyx_v_self->_adjust_minutes = ((PyBoolObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":746
 * 
 *     property _handle_non_market_minutes:
 *         def __set__(self, val):             # <<<<<<<<<<<<<<
 *             self._adjust_minutes = val
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("catalyst._protocol.BarData._handle_non_market_minutes.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":750
 * 
 *     property current_session:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._trading_calendar.minute_to_session_label(
 *                 self.simulation_dt_func(),
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_15current_session_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_15current_session_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_15current_session___get__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_15current_session___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":751
 *     property current_session:
 *         def __get__(self):
 *             return self._trading_calendar.minute_to_session_label(             # <<<<<<<<<<<<<<
 *                 self.simulation_dt_func(),
 *                 direction="next"
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_trading_calendar, __pyx_n_s_minute_to_session_label); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "catalyst/_protocol.pyx":752
 *         def __get__(self):
 *             return self._trading_calendar.minute_to_session_label(
 *                 self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *                 direction="next"
 *             )
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_3 = __pyx_v_self->simulation_dt_func; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 752, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 752, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "catalyst/_protocol.pyx":751
 *     property current_session:
 *         def __get__(self):
 *             return self._trading_calendar.minute_to_session_label(             # <<<<<<<<<<<<<<
 *                 self.simulation_dt_func(),
 *                 direction="next"
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":753
 *             return self._trading_calendar.minute_to_session_label(
 *                 self.simulation_dt_func(),
 *                 direction="next"             # <<<<<<<<<<<<<<
 *             )
 * 
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_direction, __pyx_n_s_next) < 0) __PYX_ERR(0, 753, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":751
 *     property current_session:
 *         def __get__(self):
 *             return self._trading_calendar.minute_to_session_label(             # <<<<<<<<<<<<<<
 *                 self.simulation_dt_func(),
 *                 direction="next"
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":750
 * 
 *     property current_session:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._trading_calendar.minute_to_session_label(
 *                 self.simulation_dt_func(),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.BarData.current_session.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":757
 * 
 *     property current_session_minutes:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._trading_calendar.minutes_for_session(
 *                 self.current_session
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_23current_session_minutes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_23current_session_minutes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_23current_session_minutes___get__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_23current_session_minutes___get__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":758
 *     property current_session_minutes:
 *         def __get__(self):
 *             return self._trading_calendar.minutes_for_session(             # <<<<<<<<<<<<<<
 *                 self.current_session
 *             )
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_trading_calendar, __pyx_n_s_minutes_for_session); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":759
 *         def __get__(self):
 *             return self._trading_calendar.minutes_for_session(
 *                 self.current_session             # <<<<<<<<<<<<<<
 *             )
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_current_session); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":757
 * 
 *     property current_session_minutes:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._trading_calendar.minutes_for_session(
 *                 self.current_session
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.BarData.current_session_minutes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":765
 *     # OLD API SUPPORT
 *     #################
 *     cdef _calculate_universe(self):             # <<<<<<<<<<<<<<
 *         if self._universe_func is None:
 *             return []
 */

static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__calculate_universe(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_v_simulation_dt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_calculate_universe", 0);

  /* "catalyst/_protocol.pyx":766
 *     #################
 *     cdef _calculate_universe(self):
 *         if self._universe_func is None:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_universe_func == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "catalyst/_protocol.pyx":767
 *     cdef _calculate_universe(self):
 *         if self._universe_func is None:
 *             return []             # <<<<<<<<<<<<<<
 * 
 *         simulation_dt = self.simulation_dt_func()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":766
 *     #################
 *     cdef _calculate_universe(self):
 *         if self._universe_func is None:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */
  }

  /* "catalyst/_protocol.pyx":769
 *             return []
 * 
 *         simulation_dt = self.simulation_dt_func()             # <<<<<<<<<<<<<<
 *         if self._last_calculated_universe is None or \
 *                 self._universe_last_updated_at != simulation_dt:
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_4 = __pyx_v_self->simulation_dt_func; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (__pyx_t_5) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_simulation_dt = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "catalyst/_protocol.pyx":770
 * 
 *         simulation_dt = self.simulation_dt_func()
 *         if self._last_calculated_universe is None or \             # <<<<<<<<<<<<<<
 *                 self._universe_last_updated_at != simulation_dt:
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_last_calculated_universe == Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }

  /* "catalyst/_protocol.pyx":771
 *         simulation_dt = self.simulation_dt_func()
 *         if self._last_calculated_universe is None or \
 *                 self._universe_last_updated_at != simulation_dt:             # <<<<<<<<<<<<<<
 * 
 *             self._last_calculated_universe = self._universe_func()
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_self->_universe_last_updated_at, __pyx_v_simulation_dt, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;

  /* "catalyst/_protocol.pyx":770
 * 
 *         simulation_dt = self.simulation_dt_func()
 *         if self._last_calculated_universe is None or \             # <<<<<<<<<<<<<<
 *                 self._universe_last_updated_at != simulation_dt:
 * 
 */
  if (__pyx_t_2) {

    /* "catalyst/_protocol.pyx":773
 *                 self._universe_last_updated_at != simulation_dt:
 * 
 *             self._last_calculated_universe = self._universe_func()             # <<<<<<<<<<<<<<
 *             self._universe_last_updated_at = simulation_dt
 * 
 */
    __Pyx_INCREF(__pyx_v_self->_universe_func);
    __pyx_t_4 = __pyx_v_self->_universe_func; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_last_calculated_universe);
    __Pyx_DECREF(__pyx_v_self->_last_calculated_universe);
    __pyx_v_self->_last_calculated_universe = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "catalyst/_protocol.pyx":774
 * 
 *             self._last_calculated_universe = self._universe_func()
 *             self._universe_last_updated_at = simulation_dt             # <<<<<<<<<<<<<<
 * 
 *         return self._last_calculated_universe
 */
    __Pyx_INCREF(__pyx_v_simulation_dt);
    __Pyx_GIVEREF(__pyx_v_simulation_dt);
    __Pyx_GOTREF(__pyx_v_self->_universe_last_updated_at);
    __Pyx_DECREF(__pyx_v_self->_universe_last_updated_at);
    __pyx_v_self->_universe_last_updated_at = __pyx_v_simulation_dt;

    /* "catalyst/_protocol.pyx":770
 * 
 *         simulation_dt = self.simulation_dt_func()
 *         if self._last_calculated_universe is None or \             # <<<<<<<<<<<<<<
 *                 self._universe_last_updated_at != simulation_dt:
 * 
 */
  }

  /* "catalyst/_protocol.pyx":776
 *             self._universe_last_updated_at = simulation_dt
 * 
 *         return self._last_calculated_universe             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_last_calculated_universe);
  __pyx_r = __pyx_v_self->_last_calculated_universe;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":765
 *     # OLD API SUPPORT
 *     #################
 *     cdef _calculate_universe(self):             # <<<<<<<<<<<<<<
 *         if self._universe_func is None:
 *             return []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.BarData._calculate_universe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_simulation_dt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8catalyst_9_protocol_7BarData_14generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "catalyst/_protocol.pyx":778
 *         return self._last_calculated_universe
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_13__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_13__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_12__iter__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_12__iter__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *)__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_2___iter__(__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_2___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 778, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8catalyst_9_protocol_7BarData_14generator1, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_BarData___iter, __pyx_n_s_catalyst__protocol); if (unlikely(!gen)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("catalyst._protocol.BarData.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8catalyst_9_protocol_7BarData_14generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *__pyx_cur_scope = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 778, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":779
 * 
 *     def __iter__(self):
 *         self._warn_deprecated("Iterating over the assets in `data` is "             # <<<<<<<<<<<<<<
 *                         "deprecated.")
 *         for asset in self._calculate_universe():
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_cur_scope->__pyx_v_self, __pyx_kp_s_Iterating_over_the_assets_in_dat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":781
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 *         for asset in self._calculate_universe():             # <<<<<<<<<<<<<<
 *             yield asset
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_calculate_universe(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 781, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 781, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 781, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 781, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_asset);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_asset, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":782
 *                         "deprecated.")
 *         for asset in self._calculate_universe():
 *             yield asset             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, asset):
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_asset);
    __pyx_r = __pyx_cur_scope->__pyx_v_asset;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 782, __pyx_L1_error)

    /* "catalyst/_protocol.pyx":781
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 *         for asset in self._calculate_universe():             # <<<<<<<<<<<<<<
 *             yield asset
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "catalyst/_protocol.pyx":778
 *         return self._last_calculated_universe
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":784
 *             yield asset
 * 
 *     def __contains__(self, asset):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Checking whether an asset is in data is "
 *                         "deprecated.")
 */

/* Python wrapper */
static int __pyx_pw_8catalyst_9_protocol_7BarData_16__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_asset); /*proto*/
static int __pyx_pw_8catalyst_9_protocol_7BarData_16__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_asset) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_15__contains__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), ((PyObject *)__pyx_v_asset));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8catalyst_9_protocol_7BarData_15__contains__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_asset) {
  PyObject *__pyx_v_universe = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "catalyst/_protocol.pyx":785
 * 
 *     def __contains__(self, asset):
 *         self._warn_deprecated("Checking whether an asset is in data is "             # <<<<<<<<<<<<<<
 *                         "deprecated.")
 *         universe = self._calculate_universe()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_Checking_whether_an_asset_is_in); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":787
 *         self._warn_deprecated("Checking whether an asset is in data is "
 *                         "deprecated.")
 *         universe = self._calculate_universe()             # <<<<<<<<<<<<<<
 *         return asset in universe
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_calculate_universe(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_universe = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":788
 *                         "deprecated.")
 *         universe = self._calculate_universe()
 *         return asset in universe             # <<<<<<<<<<<<<<
 * 
 *     def items(self):
 */
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_asset, __pyx_v_universe, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 788, __pyx_L1_error)
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":784
 *             yield asset
 * 
 *     def __contains__(self, asset):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Checking whether an asset is in data is "
 *                         "deprecated.")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("catalyst._protocol.BarData.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_universe);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":790
 *         return asset in universe
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_18items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_18items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_17items(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_17items(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_v_asset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("items", 0);

  /* "catalyst/_protocol.pyx":791
 * 
 *     def items(self):
 *         self._warn_deprecated("Iterating over the assets in `data` is "             # <<<<<<<<<<<<<<
 *                         "deprecated.")
 *         return [(asset, self[asset]) for asset in self._calculate_universe()]
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_Iterating_over_the_assets_in_dat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":793
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 *         return [(asset, self[asset]) for asset in self._calculate_universe()]             # <<<<<<<<<<<<<<
 * 
 *     def iteritems(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_calculate_universe(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 793, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 793, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 793, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 793, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_asset, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_asset);
    __Pyx_GIVEREF(__pyx_v_asset);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_asset);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":790
 *         return asset in universe
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("catalyst._protocol.BarData.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_asset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8catalyst_9_protocol_7BarData_21generator2(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "catalyst/_protocol.pyx":795
 *         return [(asset, self[asset]) for asset in self._calculate_universe()]
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_20iteritems(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_20iteritems(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iteritems (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_19iteritems(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_19iteritems(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iteritems", 0);
  __pyx_cur_scope = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *)__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_3_iteritems(__pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_3_iteritems, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 795, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8catalyst_9_protocol_7BarData_21generator2, (PyObject *) __pyx_cur_scope, __pyx_n_s_iteritems, __pyx_n_s_BarData_iteritems, __pyx_n_s_catalyst__protocol); if (unlikely(!gen)) __PYX_ERR(0, 795, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("catalyst._protocol.BarData.iteritems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8catalyst_9_protocol_7BarData_21generator2(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *__pyx_cur_scope = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 795, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":796
 * 
 *     def iteritems(self):
 *         self._warn_deprecated("Iterating over the assets in `data` is "             # <<<<<<<<<<<<<<
 *                         "deprecated.")
 *         for asset in self._calculate_universe():
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_cur_scope->__pyx_v_self, __pyx_kp_s_Iterating_over_the_assets_in_dat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":798
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 *         for asset in self._calculate_universe():             # <<<<<<<<<<<<<<
 *             yield asset, self[asset]
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_calculate_universe(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 798, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_asset);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_asset, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":799
 *                         "deprecated.")
 *         for asset in self._calculate_universe():
 *             yield asset, self[asset]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_asset);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_asset);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_cur_scope->__pyx_v_asset);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 799, __pyx_L1_error)

    /* "catalyst/_protocol.pyx":798
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 *         for asset in self._calculate_universe():             # <<<<<<<<<<<<<<
 *             yield asset, self[asset]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "catalyst/_protocol.pyx":795
 *         return [(asset, self[asset]) for asset in self._calculate_universe()]
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("iteritems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":801
 *             yield asset, self[asset]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_8catalyst_9_protocol_7BarData_23__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_8catalyst_9_protocol_7BarData_23__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_22__len__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_8catalyst_9_protocol_7BarData_22__len__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "catalyst/_protocol.pyx":802
 * 
 *     def __len__(self):
 *         self._warn_deprecated("Iterating over the assets in `data` is "             # <<<<<<<<<<<<<<
 *                         "deprecated.")
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_Iterating_over_the_assets_in_dat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":805
 *                         "deprecated.")
 * 
 *         return len(self._calculate_universe())             # <<<<<<<<<<<<<<
 * 
 *     def keys(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_calculate_universe(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":801
 *             yield asset, self[asset]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("catalyst._protocol.BarData.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":807
 *         return len(self._calculate_universe())
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_25keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_25keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_24keys(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_24keys(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("keys", 0);

  /* "catalyst/_protocol.pyx":808
 * 
 *     def keys(self):
 *         self._warn_deprecated("Iterating over the assets in `data` is "             # <<<<<<<<<<<<<<
 *                         "deprecated.")
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_Iterating_over_the_assets_in_dat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":811
 *                         "deprecated.")
 * 
 *         return list(self._calculate_universe())             # <<<<<<<<<<<<<<
 * 
 *     def iterkeys(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_calculate_universe(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":807
 *         return len(self._calculate_universe())
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("Iterating over the assets in `data` is "
 *                         "deprecated.")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("catalyst._protocol.BarData.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":813
 *         return list(self._calculate_universe())
 * 
 *     def iterkeys(self):             # <<<<<<<<<<<<<<
 *         return iter(self.keys())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_27iterkeys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_27iterkeys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iterkeys (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_26iterkeys(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_26iterkeys(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("iterkeys", 0);

  /* "catalyst/_protocol.pyx":814
 * 
 *     def iterkeys(self):
 *         return iter(self.keys())             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, name):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":813
 *         return list(self._calculate_universe())
 * 
 *     def iterkeys(self):             # <<<<<<<<<<<<<<
 *         return iter(self.keys())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("catalyst._protocol.BarData.iterkeys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":816
 *         return iter(self.keys())
 * 
 *     def __getitem__(self, name):             # <<<<<<<<<<<<<<
 *         return self._get_equity_price_view(name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_29__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_name); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7BarData_29__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7BarData_28__getitem__(((struct __pyx_obj_8catalyst_9_protocol_BarData *)__pyx_v_self), ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7BarData_28__getitem__(struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "catalyst/_protocol.pyx":817
 * 
 *     def __getitem__(self, name):
 *         return self._get_equity_price_view(name)             # <<<<<<<<<<<<<<
 * 
 *     cdef _warn_deprecated(self, msg):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_BarData *)__pyx_v_self->__pyx_vtab)->_get_equity_price_view(__pyx_v_self, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":816
 *         return iter(self.keys())
 * 
 *     def __getitem__(self, name):             # <<<<<<<<<<<<<<
 *         return self._get_equity_price_view(name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("catalyst._protocol.BarData.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":819
 *         return self._get_equity_price_view(name)
 * 
 *     cdef _warn_deprecated(self, msg):             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             msg,
 */

static PyObject *__pyx_f_8catalyst_9_protocol_7BarData__warn_deprecated(CYTHON_UNUSED struct __pyx_obj_8catalyst_9_protocol_BarData *__pyx_v_self, PyObject *__pyx_v_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_warn_deprecated", 0);

  /* "catalyst/_protocol.pyx":820
 * 
 *     cdef _warn_deprecated(self, msg):
 *         warnings.warn(             # <<<<<<<<<<<<<<
 *             msg,
 *             category=ZiplineDeprecationWarning,
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":821
 *     cdef _warn_deprecated(self, msg):
 *         warnings.warn(
 *             msg,             # <<<<<<<<<<<<<<
 *             category=ZiplineDeprecationWarning,
 *             stacklevel=1
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_msg);
  __Pyx_GIVEREF(__pyx_v_msg);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_msg);

  /* "catalyst/_protocol.pyx":822
 *         warnings.warn(
 *             msg,
 *             category=ZiplineDeprecationWarning,             # <<<<<<<<<<<<<<
 *             stacklevel=1
 *         )
 */
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ZiplineDeprecationWarning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_category, __pyx_t_4) < 0) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_1) < 0) __PYX_ERR(0, 822, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":820
 * 
 *     cdef _warn_deprecated(self, msg):
 *         warnings.warn(             # <<<<<<<<<<<<<<
 *             msg,
 *             category=ZiplineDeprecationWarning,
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "catalyst/_protocol.pyx":819
 *         return self._get_equity_price_view(name)
 * 
 *     cdef _warn_deprecated(self, msg):             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             msg,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.BarData._warn_deprecated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":835
 *     This class exists to temporarily support the deprecated data[sid(N)] API.
 *     """
 *     def __init__(self, asset, data_portal, simulation_dt_func, data_frequency):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

/* Python wrapper */
static int __pyx_pw_8catalyst_9_protocol_7SidView_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8catalyst_9_protocol_7SidView___init__[] = "\n        Parameters\n        ---------\n        asset : Asset\n            The asset for which the instance retrieves data.\n\n        data_portal : DataPortal\n            Provider for bar pricing data.\n\n        simulation_dt_func: function\n            Function which returns the current simulation time.\n            This is usually bound to a method of TradingSimulation.\n\n        data_frequency: string\n            The frequency of the bar data; i.e. whether the data is\n            'daily' or 'minute' bars\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_8catalyst_9_protocol_7SidView___init__;
#endif
static int __pyx_pw_8catalyst_9_protocol_7SidView_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_asset = 0;
  PyObject *__pyx_v_data_portal = 0;
  PyObject *__pyx_v_simulation_dt_func = 0;
  PyObject *__pyx_v_data_frequency = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_asset,&__pyx_n_s_data_portal,&__pyx_n_s_simulation_dt_func,&__pyx_n_s_data_frequency,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_asset)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data_portal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(0, 835, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_simulation_dt_func)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(0, 835, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data_frequency)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(0, 835, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 835, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_asset = values[0];
    __pyx_v_data_portal = values[1];
    __pyx_v_simulation_dt_func = values[2];
    __pyx_v_data_frequency = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 835, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("catalyst._protocol.SidView.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView___init__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self), __pyx_v_asset, __pyx_v_data_portal, __pyx_v_simulation_dt_func, __pyx_v_data_frequency);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8catalyst_9_protocol_7SidView___init__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_asset, PyObject *__pyx_v_data_portal, PyObject *__pyx_v_simulation_dt_func, PyObject *__pyx_v_data_frequency) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "catalyst/_protocol.pyx":853
 *             'daily' or 'minute' bars
 *         """
 *         self.asset = asset             # <<<<<<<<<<<<<<
 *         self.data_portal = data_portal
 *         self.simulation_dt_func = simulation_dt_func
 */
  __Pyx_INCREF(__pyx_v_asset);
  __Pyx_GIVEREF(__pyx_v_asset);
  __Pyx_GOTREF(__pyx_v_self->asset);
  __Pyx_DECREF(__pyx_v_self->asset);
  __pyx_v_self->asset = __pyx_v_asset;

  /* "catalyst/_protocol.pyx":854
 *         """
 *         self.asset = asset
 *         self.data_portal = data_portal             # <<<<<<<<<<<<<<
 *         self.simulation_dt_func = simulation_dt_func
 *         self.data_frequency = data_frequency
 */
  __Pyx_INCREF(__pyx_v_data_portal);
  __Pyx_GIVEREF(__pyx_v_data_portal);
  __Pyx_GOTREF(__pyx_v_self->data_portal);
  __Pyx_DECREF(__pyx_v_self->data_portal);
  __pyx_v_self->data_portal = __pyx_v_data_portal;

  /* "catalyst/_protocol.pyx":855
 *         self.asset = asset
 *         self.data_portal = data_portal
 *         self.simulation_dt_func = simulation_dt_func             # <<<<<<<<<<<<<<
 *         self.data_frequency = data_frequency
 * 
 */
  __Pyx_INCREF(__pyx_v_simulation_dt_func);
  __Pyx_GIVEREF(__pyx_v_simulation_dt_func);
  __Pyx_GOTREF(__pyx_v_self->simulation_dt_func);
  __Pyx_DECREF(__pyx_v_self->simulation_dt_func);
  __pyx_v_self->simulation_dt_func = __pyx_v_simulation_dt_func;

  /* "catalyst/_protocol.pyx":856
 *         self.data_portal = data_portal
 *         self.simulation_dt_func = simulation_dt_func
 *         self.data_frequency = data_frequency             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, column):
 */
  __Pyx_INCREF(__pyx_v_data_frequency);
  __Pyx_GIVEREF(__pyx_v_data_frequency);
  __Pyx_GOTREF(__pyx_v_self->data_frequency);
  __Pyx_DECREF(__pyx_v_self->data_frequency);
  __pyx_v_self->data_frequency = __pyx_v_data_frequency;

  /* "catalyst/_protocol.pyx":835
 *     This class exists to temporarily support the deprecated data[sid(N)] API.
 *     """
 *     def __init__(self, asset, data_portal, simulation_dt_func, data_frequency):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":858
 *         self.data_frequency = data_frequency
 * 
 *     def __getattr__(self, column):             # <<<<<<<<<<<<<<
 *         # backwards compatibility code for Q1 API
 *         if column == "close_price":
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_3__getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_column); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_3__getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_column) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_2__getattr__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self), ((PyObject *)__pyx_v_column));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_2__getattr__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_column) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__getattr__", 0);
  __Pyx_INCREF(__pyx_v_column);

  /* "catalyst/_protocol.pyx":860
 *     def __getattr__(self, column):
 *         # backwards compatibility code for Q1 API
 *         if column == "close_price":             # <<<<<<<<<<<<<<
 *             column = "close"
 *         elif column == "open_price":
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_column, __pyx_n_s_close_price, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 860, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "catalyst/_protocol.pyx":861
 *         # backwards compatibility code for Q1 API
 *         if column == "close_price":
 *             column = "close"             # <<<<<<<<<<<<<<
 *         elif column == "open_price":
 *             column = "open"
 */
    __Pyx_INCREF(__pyx_n_s_close);
    __Pyx_DECREF_SET(__pyx_v_column, __pyx_n_s_close);

    /* "catalyst/_protocol.pyx":860
 *     def __getattr__(self, column):
 *         # backwards compatibility code for Q1 API
 *         if column == "close_price":             # <<<<<<<<<<<<<<
 *             column = "close"
 *         elif column == "open_price":
 */
    goto __pyx_L3;
  }

  /* "catalyst/_protocol.pyx":862
 *         if column == "close_price":
 *             column = "close"
 *         elif column == "open_price":             # <<<<<<<<<<<<<<
 *             column = "open"
 *         elif column == "dt":
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_column, __pyx_n_s_open_price, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 862, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "catalyst/_protocol.pyx":863
 *             column = "close"
 *         elif column == "open_price":
 *             column = "open"             # <<<<<<<<<<<<<<
 *         elif column == "dt":
 *             return self.dt
 */
    __Pyx_INCREF(__pyx_n_s_open);
    __Pyx_DECREF_SET(__pyx_v_column, __pyx_n_s_open);

    /* "catalyst/_protocol.pyx":862
 *         if column == "close_price":
 *             column = "close"
 *         elif column == "open_price":             # <<<<<<<<<<<<<<
 *             column = "open"
 *         elif column == "dt":
 */
    goto __pyx_L3;
  }

  /* "catalyst/_protocol.pyx":864
 *         elif column == "open_price":
 *             column = "open"
 *         elif column == "dt":             # <<<<<<<<<<<<<<
 *             return self.dt
 *         elif column == "datetime":
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_column, __pyx_n_s_dt, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 864, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "catalyst/_protocol.pyx":865
 *             column = "open"
 *         elif column == "dt":
 *             return self.dt             # <<<<<<<<<<<<<<
 *         elif column == "datetime":
 *             return self.datetime
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":864
 *         elif column == "open_price":
 *             column = "open"
 *         elif column == "dt":             # <<<<<<<<<<<<<<
 *             return self.dt
 *         elif column == "datetime":
 */
  }

  /* "catalyst/_protocol.pyx":866
 *         elif column == "dt":
 *             return self.dt
 *         elif column == "datetime":             # <<<<<<<<<<<<<<
 *             return self.datetime
 *         elif column == "sid":
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_column, __pyx_n_s_datetime, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 866, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "catalyst/_protocol.pyx":867
 *             return self.dt
 *         elif column == "datetime":
 *             return self.datetime             # <<<<<<<<<<<<<<
 *         elif column == "sid":
 *             return self.sid
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_datetime); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":866
 *         elif column == "dt":
 *             return self.dt
 *         elif column == "datetime":             # <<<<<<<<<<<<<<
 *             return self.datetime
 *         elif column == "sid":
 */
  }

  /* "catalyst/_protocol.pyx":868
 *         elif column == "datetime":
 *             return self.datetime
 *         elif column == "sid":             # <<<<<<<<<<<<<<
 *             return self.sid
 * 
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_column, __pyx_n_s_sid, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 868, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "catalyst/_protocol.pyx":869
 *             return self.datetime
 *         elif column == "sid":
 *             return self.sid             # <<<<<<<<<<<<<<
 * 
 *         return self.data_portal.get_spot_value(
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "catalyst/_protocol.pyx":868
 *         elif column == "datetime":
 *             return self.datetime
 *         elif column == "sid":             # <<<<<<<<<<<<<<
 *             return self.sid
 * 
 */
  }
  __pyx_L3:;

  /* "catalyst/_protocol.pyx":871
 *             return self.sid
 * 
 *         return self.data_portal.get_spot_value(             # <<<<<<<<<<<<<<
 *             self.asset,
 *             column,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_spot_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "catalyst/_protocol.pyx":874
 *             self.asset,
 *             column,
 *             self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *             self.data_frequency
 *         )
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_5 = __pyx_v_self->simulation_dt_func; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (__pyx_t_6) {
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 874, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "catalyst/_protocol.pyx":875
 *             column,
 *             self.simulation_dt_func(),
 *             self.data_frequency             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_5 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_self->asset, __pyx_v_column, __pyx_t_4, __pyx_v_self->data_frequency};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_self->asset, __pyx_v_column, __pyx_t_4, __pyx_v_self->data_frequency};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->asset);
    __Pyx_GIVEREF(__pyx_v_self->asset);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_self->asset);
    __Pyx_INCREF(__pyx_v_column);
    __Pyx_GIVEREF(__pyx_v_column);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_column);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_self->data_frequency);
    __Pyx_GIVEREF(__pyx_v_self->data_frequency);
    PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_7, __pyx_v_self->data_frequency);
    __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":858
 *         self.data_frequency = data_frequency
 * 
 *     def __getattr__(self, column):             # <<<<<<<<<<<<<<
 *         # backwards compatibility code for Q1 API
 *         if column == "close_price":
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("catalyst._protocol.SidView.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_column);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":878
 *         )
 * 
 *     def __contains__(self, column):             # <<<<<<<<<<<<<<
 *         return self.data_portal.contains(self.asset, column)
 * 
 */

/* Python wrapper */
static int __pyx_pw_8catalyst_9_protocol_7SidView_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_column); /*proto*/
static int __pyx_pw_8catalyst_9_protocol_7SidView_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_column) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_4__contains__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self), ((PyObject *)__pyx_v_column));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8catalyst_9_protocol_7SidView_4__contains__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_column) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "catalyst/_protocol.pyx":879
 * 
 *     def __contains__(self, column):
 *         return self.data_portal.contains(self.asset, column)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, column):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self->asset, __pyx_v_column};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self->asset, __pyx_v_column};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->asset);
    __Pyx_GIVEREF(__pyx_v_self->asset);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_self->asset);
    __Pyx_INCREF(__pyx_v_column);
    __Pyx_GIVEREF(__pyx_v_column);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_column);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":878
 *         )
 * 
 *     def __contains__(self, column):             # <<<<<<<<<<<<<<
 *         return self.data_portal.contains(self.asset, column)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.SidView.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":881
 *         return self.data_portal.contains(self.asset, column)
 * 
 *     def __getitem__(self, column):             # <<<<<<<<<<<<<<
 *         return self.__getattr__(column)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_column); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_column) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_6__getitem__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self), ((PyObject *)__pyx_v_column));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_6__getitem__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_column) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "catalyst/_protocol.pyx":882
 * 
 *     def __getitem__(self, column):
 *         return self.__getattr__(column)             # <<<<<<<<<<<<<<
 * 
 *     property sid:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getattr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_column); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_column};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_column};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_column);
      __Pyx_GIVEREF(__pyx_v_column);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_column);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":881
 *         return self.data_portal.contains(self.asset, column)
 * 
 *     def __getitem__(self, column):             # <<<<<<<<<<<<<<
 *         return self.__getattr__(column)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.SidView.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":885
 * 
 *     property sid:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.asset
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_3sid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_3sid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_3sid___get__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_3sid___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":886
 *     property sid:
 *         def __get__(self):
 *             return self.asset             # <<<<<<<<<<<<<<
 * 
 *     property dt:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->asset);
  __pyx_r = __pyx_v_self->asset;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":885
 * 
 *     property sid:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.asset
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":889
 * 
 *     property dt:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.datetime
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_2dt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_2dt_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_2dt___get__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_2dt___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":890
 *     property dt:
 *         def __get__(self):
 *             return self.datetime             # <<<<<<<<<<<<<<
 * 
 *     property datetime:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_datetime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":889
 * 
 *     property dt:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.datetime
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("catalyst._protocol.SidView.dt.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":893
 * 
 *     property datetime:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.data_portal.get_last_traded_dt(
 *                 self.asset,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_8datetime_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_8datetime_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_8datetime___get__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_8datetime___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":894
 *     property datetime:
 *         def __get__(self):
 *             return self.data_portal.get_last_traded_dt(             # <<<<<<<<<<<<<<
 *                 self.asset,
 *                 self.simulation_dt_func(),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_last_traded_dt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":896
 *             return self.data_portal.get_last_traded_dt(
 *                 self.asset,
 *                 self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *                 self.data_frequency)
 * 
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_4 = __pyx_v_self->simulation_dt_func; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (__pyx_t_5) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "catalyst/_protocol.pyx":897
 *                 self.asset,
 *                 self.simulation_dt_func(),
 *                 self.data_frequency)             # <<<<<<<<<<<<<<
 * 
 *     property current_dt:
 */
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self->asset, __pyx_t_3, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_self->asset, __pyx_t_3, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->asset);
    __Pyx_GIVEREF(__pyx_v_self->asset);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_self->asset);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->data_frequency);
    __Pyx_GIVEREF(__pyx_v_self->data_frequency);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_v_self->data_frequency);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":893
 * 
 *     property datetime:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.data_portal.get_last_traded_dt(
 *                 self.asset,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.SidView.datetime.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":900
 * 
 *     property current_dt:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.simulation_dt_func()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_10current_dt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_10current_dt_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_10current_dt___get__(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_10current_dt___get__(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "catalyst/_protocol.pyx":901
 *     property current_dt:
 *         def __get__(self):
 *             return self.simulation_dt_func()             # <<<<<<<<<<<<<<
 * 
 *     def mavg(self, num_minutes):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_2 = __pyx_v_self->simulation_dt_func; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":900
 * 
 *     property current_dt:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.simulation_dt_func()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("catalyst._protocol.SidView.current_dt.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":903
 *             return self.simulation_dt_func()
 * 
 *     def mavg(self, num_minutes):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `mavg` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_9mavg(PyObject *__pyx_v_self, PyObject *__pyx_v_num_minutes); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_9mavg(PyObject *__pyx_v_self, PyObject *__pyx_v_num_minutes) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mavg (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_8mavg(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self), ((PyObject *)__pyx_v_num_minutes));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_8mavg(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_num_minutes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("mavg", 0);

  /* "catalyst/_protocol.pyx":904
 * 
 *     def mavg(self, num_minutes):
 *         self._warn_deprecated("The `mavg` method is deprecated.")             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "mavg", self.simulation_dt_func(),
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_SidView *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_The_mavg_method_is_deprecated); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":905
 *     def mavg(self, num_minutes):
 *         self._warn_deprecated("The `mavg` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "mavg", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_simple_transform); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "catalyst/_protocol.pyx":906
 *         self._warn_deprecated("The `mavg` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "mavg", self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *             self.data_frequency, bars=num_minutes
 *         )
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_3 = __pyx_v_self->simulation_dt_func; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "catalyst/_protocol.pyx":905
 *     def mavg(self, num_minutes):
 *         self._warn_deprecated("The `mavg` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "mavg", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->asset);
  __Pyx_GIVEREF(__pyx_v_self->asset);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->asset);
  __Pyx_INCREF(__pyx_n_s_mavg);
  __Pyx_GIVEREF(__pyx_n_s_mavg);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_mavg);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->data_frequency);
  __Pyx_GIVEREF(__pyx_v_self->data_frequency);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->data_frequency);
  __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":907
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "mavg", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":908
 *             self.asset, "mavg", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 *         )             # <<<<<<<<<<<<<<
 * 
 *     def stddev(self, num_minutes):
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bars, __pyx_v_num_minutes) < 0) __PYX_ERR(0, 907, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":905
 *     def mavg(self, num_minutes):
 *         self._warn_deprecated("The `mavg` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "mavg", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":903
 *             return self.simulation_dt_func()
 * 
 *     def mavg(self, num_minutes):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `mavg` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.SidView.mavg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":910
 *         )
 * 
 *     def stddev(self, num_minutes):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `stddev` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_11stddev(PyObject *__pyx_v_self, PyObject *__pyx_v_num_minutes); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_11stddev(PyObject *__pyx_v_self, PyObject *__pyx_v_num_minutes) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stddev (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_10stddev(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self), ((PyObject *)__pyx_v_num_minutes));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_10stddev(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_num_minutes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("stddev", 0);

  /* "catalyst/_protocol.pyx":911
 * 
 *     def stddev(self, num_minutes):
 *         self._warn_deprecated("The `stddev` method is deprecated.")             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "stddev", self.simulation_dt_func(),
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_SidView *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_The_stddev_method_is_deprecated); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":912
 *     def stddev(self, num_minutes):
 *         self._warn_deprecated("The `stddev` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "stddev", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_simple_transform); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "catalyst/_protocol.pyx":913
 *         self._warn_deprecated("The `stddev` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "stddev", self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *             self.data_frequency, bars=num_minutes
 *         )
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_3 = __pyx_v_self->simulation_dt_func; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 913, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 913, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "catalyst/_protocol.pyx":912
 *     def stddev(self, num_minutes):
 *         self._warn_deprecated("The `stddev` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "stddev", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->asset);
  __Pyx_GIVEREF(__pyx_v_self->asset);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->asset);
  __Pyx_INCREF(__pyx_n_s_stddev);
  __Pyx_GIVEREF(__pyx_n_s_stddev);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_stddev);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->data_frequency);
  __Pyx_GIVEREF(__pyx_v_self->data_frequency);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->data_frequency);
  __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":914
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "stddev", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":915
 *             self.asset, "stddev", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 *         )             # <<<<<<<<<<<<<<
 * 
 *     def vwap(self, num_minutes):
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bars, __pyx_v_num_minutes) < 0) __PYX_ERR(0, 914, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":912
 *     def stddev(self, num_minutes):
 *         self._warn_deprecated("The `stddev` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "stddev", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":910
 *         )
 * 
 *     def stddev(self, num_minutes):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `stddev` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.SidView.stddev", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":917
 *         )
 * 
 *     def vwap(self, num_minutes):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `vwap` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_13vwap(PyObject *__pyx_v_self, PyObject *__pyx_v_num_minutes); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_13vwap(PyObject *__pyx_v_self, PyObject *__pyx_v_num_minutes) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vwap (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_12vwap(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self), ((PyObject *)__pyx_v_num_minutes));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_12vwap(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_num_minutes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("vwap", 0);

  /* "catalyst/_protocol.pyx":918
 * 
 *     def vwap(self, num_minutes):
 *         self._warn_deprecated("The `vwap` method is deprecated.")             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "vwap", self.simulation_dt_func(),
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_SidView *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_The_vwap_method_is_deprecated); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":919
 *     def vwap(self, num_minutes):
 *         self._warn_deprecated("The `vwap` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "vwap", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_simple_transform); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "catalyst/_protocol.pyx":920
 *         self._warn_deprecated("The `vwap` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "vwap", self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *             self.data_frequency, bars=num_minutes
 *         )
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_3 = __pyx_v_self->simulation_dt_func; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "catalyst/_protocol.pyx":919
 *     def vwap(self, num_minutes):
 *         self._warn_deprecated("The `vwap` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "vwap", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->asset);
  __Pyx_GIVEREF(__pyx_v_self->asset);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->asset);
  __Pyx_INCREF(__pyx_n_s_vwap);
  __Pyx_GIVEREF(__pyx_n_s_vwap);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_vwap);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->data_frequency);
  __Pyx_GIVEREF(__pyx_v_self->data_frequency);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->data_frequency);
  __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":921
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "vwap", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":922
 *             self.asset, "vwap", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 *         )             # <<<<<<<<<<<<<<
 * 
 *     def returns(self):
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bars, __pyx_v_num_minutes) < 0) __PYX_ERR(0, 921, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":919
 *     def vwap(self, num_minutes):
 *         self._warn_deprecated("The `vwap` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "vwap", self.simulation_dt_func(),
 *             self.data_frequency, bars=num_minutes
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":917
 *         )
 * 
 *     def vwap(self, num_minutes):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `vwap` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.SidView.vwap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":924
 *         )
 * 
 *     def returns(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `returns` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

/* Python wrapper */
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_15returns(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8catalyst_9_protocol_7SidView_15returns(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("returns (wrapper)", 0);
  __pyx_r = __pyx_pf_8catalyst_9_protocol_7SidView_14returns(((struct __pyx_obj_8catalyst_9_protocol_SidView *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8catalyst_9_protocol_7SidView_14returns(struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("returns", 0);

  /* "catalyst/_protocol.pyx":925
 * 
 *     def returns(self):
 *         self._warn_deprecated("The `returns` method is deprecated.")             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "returns", self.simulation_dt_func(),
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8catalyst_9_protocol_SidView *)__pyx_v_self->__pyx_vtab)->_warn_deprecated(__pyx_v_self, __pyx_kp_s_The_returns_method_is_deprecated); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":926
 *     def returns(self):
 *         self._warn_deprecated("The `returns` method is deprecated.")
 *         return self.data_portal.get_simple_transform(             # <<<<<<<<<<<<<<
 *             self.asset, "returns", self.simulation_dt_func(),
 *             self.data_frequency
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data_portal, __pyx_n_s_get_simple_transform); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":927
 *         self._warn_deprecated("The `returns` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "returns", self.simulation_dt_func(),             # <<<<<<<<<<<<<<
 *             self.data_frequency
 *         )
 */
  __Pyx_INCREF(__pyx_v_self->simulation_dt_func);
  __pyx_t_4 = __pyx_v_self->simulation_dt_func; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (__pyx_t_5) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 927, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 927, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "catalyst/_protocol.pyx":928
 *         return self.data_portal.get_simple_transform(
 *             self.asset, "returns", self.simulation_dt_func(),
 *             self.data_frequency             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_self->asset, __pyx_n_s_returns, __pyx_t_3, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_self->asset, __pyx_n_s_returns, __pyx_t_3, __pyx_v_self->data_frequency};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->asset);
    __Pyx_GIVEREF(__pyx_v_self->asset);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_self->asset);
    __Pyx_INCREF(__pyx_n_s_returns);
    __Pyx_GIVEREF(__pyx_n_s_returns);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_n_s_returns);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->data_frequency);
    __Pyx_GIVEREF(__pyx_v_self->data_frequency);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_6, __pyx_v_self->data_frequency);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "catalyst/_protocol.pyx":924
 *         )
 * 
 *     def returns(self):             # <<<<<<<<<<<<<<
 *         self._warn_deprecated("The `returns` method is deprecated.")
 *         return self.data_portal.get_simple_transform(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("catalyst._protocol.SidView.returns", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "catalyst/_protocol.pyx":931
 *         )
 * 
 *     cdef _warn_deprecated(self, msg):             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             msg,
 */

static PyObject *__pyx_f_8catalyst_9_protocol_7SidView__warn_deprecated(CYTHON_UNUSED struct __pyx_obj_8catalyst_9_protocol_SidView *__pyx_v_self, PyObject *__pyx_v_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_warn_deprecated", 0);

  /* "catalyst/_protocol.pyx":932
 * 
 *     cdef _warn_deprecated(self, msg):
 *         warnings.warn(             # <<<<<<<<<<<<<<
 *             msg,
 *             category=ZiplineDeprecationWarning,
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":933
 *     cdef _warn_deprecated(self, msg):
 *         warnings.warn(
 *             msg,             # <<<<<<<<<<<<<<
 *             category=ZiplineDeprecationWarning,
 *             stacklevel=1
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_msg);
  __Pyx_GIVEREF(__pyx_v_msg);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_msg);

  /* "catalyst/_protocol.pyx":934
 *         warnings.warn(
 *             msg,
 *             category=ZiplineDeprecationWarning,             # <<<<<<<<<<<<<<
 *             stacklevel=1
 *         )
 */
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ZiplineDeprecationWarning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_category, __pyx_t_4) < 0) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_1) < 0) __PYX_ERR(0, 934, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":932
 * 
 *     cdef _warn_deprecated(self, msg):
 *         warnings.warn(             # <<<<<<<<<<<<<<
 *             msg,
 *             category=ZiplineDeprecationWarning,
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "catalyst/_protocol.pyx":931
 *         )
 * 
 *     cdef _warn_deprecated(self, msg):             # <<<<<<<<<<<<<<
 *         warnings.warn(
 *             msg,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("catalyst._protocol.SidView._warn_deprecated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_8catalyst_9_protocol_check_parameters(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8catalyst_9_protocol_check_parameters *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8catalyst_9_protocol_check_parameters *)o);
  p->keyword_names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->keys_to_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8catalyst_9_protocol_check_parameters(PyObject *o) {
  struct __pyx_obj_8catalyst_9_protocol_check_parameters *p = (struct __pyx_obj_8catalyst_9_protocol_check_parameters *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->keyword_names);
  Py_CLEAR(p->types);
  Py_CLEAR(p->keys_to_types);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8catalyst_9_protocol_check_parameters(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8catalyst_9_protocol_check_parameters *p = (struct __pyx_obj_8catalyst_9_protocol_check_parameters *)o;
  if (p->keyword_names) {
    e = (*v)(p->keyword_names, a); if (e) return e;
  }
  if (p->types) {
    e = (*v)(p->types, a); if (e) return e;
  }
  if (p->keys_to_types) {
    e = (*v)(p->keys_to_types, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8catalyst_9_protocol_check_parameters(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8catalyst_9_protocol_check_parameters *p = (struct __pyx_obj_8catalyst_9_protocol_check_parameters *)o;
  tmp = ((PyObject*)p->keyword_names);
  p->keyword_names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->types);
  p->types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->keys_to_types);
  p->keys_to_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8catalyst_9_protocol_check_parameters[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8catalyst_9_protocol_check_parameters = {
  PyVarObject_HEAD_INIT(0, 0)
  "catalyst._protocol.check_parameters", /*tp_name*/
  sizeof(struct __pyx_obj_8catalyst_9_protocol_check_parameters), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8catalyst_9_protocol_check_parameters, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  __pyx_pw_8catalyst_9_protocol_16check_parameters_3__call__, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    Asserts that the keywords passed into the wrapped function are included\n    in those passed into this decorator. If not, raise a TypeError with a\n    meaningful message, unlike the one Cython returns by default.\n\n    Also asserts that the arguments passed into the wrapped function are\n    consistent with the types passed into this decorator. If not, raise a\n    TypeError with a meaningful message.\n    ", /*tp_doc*/
  __pyx_tp_traverse_8catalyst_9_protocol_check_parameters, /*tp_traverse*/
  __pyx_tp_clear_8catalyst_9_protocol_check_parameters, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8catalyst_9_protocol_check_parameters, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8catalyst_9_protocol_16check_parameters_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8catalyst_9_protocol_check_parameters, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8catalyst_9_protocol_BarData __pyx_vtable_8catalyst_9_protocol_BarData;

static PyObject *__pyx_tp_new_8catalyst_9_protocol_BarData(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8catalyst_9_protocol_BarData *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8catalyst_9_protocol_BarData *)o);
  p->__pyx_vtab = __pyx_vtabptr_8catalyst_9_protocol_BarData;
  p->data_portal = Py_None; Py_INCREF(Py_None);
  p->simulation_dt_func = Py_None; Py_INCREF(Py_None);
  p->data_frequency = Py_None; Py_INCREF(Py_None);
  p->restrictions = Py_None; Py_INCREF(Py_None);
  p->_views = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_universe_func = Py_None; Py_INCREF(Py_None);
  p->_last_calculated_universe = Py_None; Py_INCREF(Py_None);
  p->_universe_last_updated_at = Py_None; Py_INCREF(Py_None);
  p->_daily_mode = ((PyBoolObject *)Py_None); Py_INCREF(Py_None);
  p->_trading_calendar = Py_None; Py_INCREF(Py_None);
  p->_is_restricted = Py_None; Py_INCREF(Py_None);
  p->_adjust_minutes = ((PyBoolObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8catalyst_9_protocol_BarData(PyObject *o) {
  struct __pyx_obj_8catalyst_9_protocol_BarData *p = (struct __pyx_obj_8catalyst_9_protocol_BarData *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->data_portal);
  Py_CLEAR(p->simulation_dt_func);
  Py_CLEAR(p->data_frequency);
  Py_CLEAR(p->restrictions);
  Py_CLEAR(p->_views);
  Py_CLEAR(p->_universe_func);
  Py_CLEAR(p->_last_calculated_universe);
  Py_CLEAR(p->_universe_last_updated_at);
  Py_CLEAR(p->_daily_mode);
  Py_CLEAR(p->_trading_calendar);
  Py_CLEAR(p->_is_restricted);
  Py_CLEAR(p->_adjust_minutes);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8catalyst_9_protocol_BarData(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8catalyst_9_protocol_BarData *p = (struct __pyx_obj_8catalyst_9_protocol_BarData *)o;
  if (p->data_portal) {
    e = (*v)(p->data_portal, a); if (e) return e;
  }
  if (p->simulation_dt_func) {
    e = (*v)(p->simulation_dt_func, a); if (e) return e;
  }
  if (p->data_frequency) {
    e = (*v)(p->data_frequency, a); if (e) return e;
  }
  if (p->restrictions) {
    e = (*v)(p->restrictions, a); if (e) return e;
  }
  if (p->_views) {
    e = (*v)(p->_views, a); if (e) return e;
  }
  if (p->_universe_func) {
    e = (*v)(p->_universe_func, a); if (e) return e;
  }
  if (p->_last_calculated_universe) {
    e = (*v)(p->_last_calculated_universe, a); if (e) return e;
  }
  if (p->_universe_last_updated_at) {
    e = (*v)(p->_universe_last_updated_at, a); if (e) return e;
  }
  if (p->_daily_mode) {
    e = (*v)(((PyObject*)p->_daily_mode), a); if (e) return e;
  }
  if (p->_trading_calendar) {
    e = (*v)(p->_trading_calendar, a); if (e) return e;
  }
  if (p->_is_restricted) {
    e = (*v)(p->_is_restricted, a); if (e) return e;
  }
  if (p->_adjust_minutes) {
    e = (*v)(((PyObject*)p->_adjust_minutes), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8catalyst_9_protocol_BarData(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8catalyst_9_protocol_BarData *p = (struct __pyx_obj_8catalyst_9_protocol_BarData *)o;
  tmp = ((PyObject*)p->data_portal);
  p->data_portal = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->simulation_dt_func);
  p->simulation_dt_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_frequency);
  p->data_frequency = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->restrictions);
  p->restrictions = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_views);
  p->_views = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_universe_func);
  p->_universe_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_calculated_universe);
  p->_last_calculated_universe = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_universe_last_updated_at);
  p->_universe_last_updated_at = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_daily_mode);
  p->_daily_mode = ((PyBoolObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_trading_calendar);
  p->_trading_calendar = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_is_restricted);
  p->_is_restricted = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_adjust_minutes);
  p->_adjust_minutes = ((PyBoolObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_8catalyst_9_protocol_BarData(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7BarData_current_dt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7BarData_10current_dt_1__get__(o);
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7BarData_fetcher_assets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7BarData_14fetcher_assets_1__get__(o);
}

static int __pyx_setprop_8catalyst_9_protocol_7BarData__handle_non_market_minutes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8catalyst_9_protocol_7BarData_26_handle_non_market_minutes_1__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7BarData_current_session(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7BarData_15current_session_1__get__(o);
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7BarData_current_session_minutes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7BarData_23current_session_minutes_1__get__(o);
}

static PyMethodDef __pyx_methods_8catalyst_9_protocol_BarData[] = {
  {"current", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_3current, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8catalyst_9_protocol_7BarData_2current},
  {"current_chain", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_5current_chain, METH_O, 0},
  {"can_trade", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_7can_trade, METH_O, __pyx_doc_8catalyst_9_protocol_7BarData_6can_trade},
  {"is_stale", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_9is_stale, METH_O, __pyx_doc_8catalyst_9_protocol_7BarData_8is_stale},
  {"history", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_11history, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8catalyst_9_protocol_7BarData_10history},
  {"items", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_18items, METH_NOARGS, 0},
  {"iteritems", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_20iteritems, METH_NOARGS, 0},
  {"keys", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_25keys, METH_NOARGS, 0},
  {"iterkeys", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7BarData_27iterkeys, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8catalyst_9_protocol_BarData[] = {
  {(char *)"current_dt", __pyx_getprop_8catalyst_9_protocol_7BarData_current_dt, 0, (char *)0, 0},
  {(char *)"fetcher_assets", __pyx_getprop_8catalyst_9_protocol_7BarData_fetcher_assets, 0, (char *)0, 0},
  {(char *)"_handle_non_market_minutes", 0, __pyx_setprop_8catalyst_9_protocol_7BarData__handle_non_market_minutes, (char *)0, 0},
  {(char *)"current_session", __pyx_getprop_8catalyst_9_protocol_7BarData_current_session, 0, (char *)0, 0},
  {(char *)"current_session_minutes", __pyx_getprop_8catalyst_9_protocol_7BarData_current_session_minutes, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_BarData = {
  __pyx_pw_8catalyst_9_protocol_7BarData_23__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_8catalyst_9_protocol_BarData, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_8catalyst_9_protocol_7BarData_16__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_BarData = {
  __pyx_pw_8catalyst_9_protocol_7BarData_23__len__, /*mp_length*/
  __pyx_pw_8catalyst_9_protocol_7BarData_29__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_8catalyst_9_protocol_BarData = {
  PyVarObject_HEAD_INIT(0, 0)
  "catalyst._protocol.BarData", /*tp_name*/
  sizeof(struct __pyx_obj_8catalyst_9_protocol_BarData), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8catalyst_9_protocol_BarData, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_BarData, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_BarData, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    Provides methods to access spot value or history windows of price data.\n    Also provides some utility methods to determine if an asset is alive,\n    has recent trade data, etc.\n\n    This is what is passed as ``data`` to the ``handle_data`` function.\n\n    Parameters\n    ----------\n    data_portal : DataPortal\n        Provider for bar pricing data.\n    simulation_dt_func : callable\n        Function which returns the current simulation time.\n        This is usually bound to a method of TradingSimulation.\n    data_frequency : {'minute', 'daily'}\n        The frequency of the bar data; i.e. whether the data is\n        daily or minute bars\n    restrictions : catalyst.finance.asset_restrictions.Restrictions\n        Object that combines and returns restricted list information from\n        multiple sources\n    universe_func : callable, optional\n        Function which returns the current 'universe'.  This is for\n        backwards compatibility with older API concepts.\n    ", /*tp_doc*/
  __pyx_tp_traverse_8catalyst_9_protocol_BarData, /*tp_traverse*/
  __pyx_tp_clear_8catalyst_9_protocol_BarData, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8catalyst_9_protocol_7BarData_13__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8catalyst_9_protocol_BarData, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8catalyst_9_protocol_BarData, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8catalyst_9_protocol_7BarData_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8catalyst_9_protocol_BarData, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8catalyst_9_protocol_SidView __pyx_vtable_8catalyst_9_protocol_SidView;

static PyObject *__pyx_tp_new_8catalyst_9_protocol_SidView(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8catalyst_9_protocol_SidView *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8catalyst_9_protocol_SidView *)o);
  p->__pyx_vtab = __pyx_vtabptr_8catalyst_9_protocol_SidView;
  p->asset = Py_None; Py_INCREF(Py_None);
  p->data_portal = Py_None; Py_INCREF(Py_None);
  p->simulation_dt_func = Py_None; Py_INCREF(Py_None);
  p->data_frequency = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8catalyst_9_protocol_SidView(PyObject *o) {
  struct __pyx_obj_8catalyst_9_protocol_SidView *p = (struct __pyx_obj_8catalyst_9_protocol_SidView *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->asset);
  Py_CLEAR(p->data_portal);
  Py_CLEAR(p->simulation_dt_func);
  Py_CLEAR(p->data_frequency);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8catalyst_9_protocol_SidView(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8catalyst_9_protocol_SidView *p = (struct __pyx_obj_8catalyst_9_protocol_SidView *)o;
  if (p->asset) {
    e = (*v)(p->asset, a); if (e) return e;
  }
  if (p->data_portal) {
    e = (*v)(p->data_portal, a); if (e) return e;
  }
  if (p->simulation_dt_func) {
    e = (*v)(p->simulation_dt_func, a); if (e) return e;
  }
  if (p->data_frequency) {
    e = (*v)(p->data_frequency, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8catalyst_9_protocol_SidView(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8catalyst_9_protocol_SidView *p = (struct __pyx_obj_8catalyst_9_protocol_SidView *)o;
  tmp = ((PyObject*)p->asset);
  p->asset = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_portal);
  p->data_portal = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->simulation_dt_func);
  p->simulation_dt_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_frequency);
  p->data_frequency = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_8catalyst_9_protocol_SidView(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_tp_getattro_8catalyst_9_protocol_SidView(PyObject *o, PyObject *n) {
  PyObject *v = PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_pw_8catalyst_9_protocol_7SidView_3__getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7SidView_sid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7SidView_3sid_1__get__(o);
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7SidView_dt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7SidView_2dt_1__get__(o);
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7SidView_datetime(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7SidView_8datetime_1__get__(o);
}

static PyObject *__pyx_getprop_8catalyst_9_protocol_7SidView_current_dt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8catalyst_9_protocol_7SidView_10current_dt_1__get__(o);
}

static PyMethodDef __pyx_methods_8catalyst_9_protocol_SidView[] = {
  {"__getattr__", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7SidView_3__getattr__, METH_O|METH_COEXIST, 0},
  {"mavg", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7SidView_9mavg, METH_O, 0},
  {"stddev", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7SidView_11stddev, METH_O, 0},
  {"vwap", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7SidView_13vwap, METH_O, 0},
  {"returns", (PyCFunction)__pyx_pw_8catalyst_9_protocol_7SidView_15returns, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8catalyst_9_protocol_SidView[] = {
  {(char *)"sid", __pyx_getprop_8catalyst_9_protocol_7SidView_sid, 0, (char *)0, 0},
  {(char *)"dt", __pyx_getprop_8catalyst_9_protocol_7SidView_dt, 0, (char *)0, 0},
  {(char *)"datetime", __pyx_getprop_8catalyst_9_protocol_7SidView_datetime, 0, (char *)0, 0},
  {(char *)"current_dt", __pyx_getprop_8catalyst_9_protocol_7SidView_current_dt, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_SidView = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_8catalyst_9_protocol_SidView, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_8catalyst_9_protocol_7SidView_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_SidView = {
  0, /*mp_length*/
  __pyx_pw_8catalyst_9_protocol_7SidView_7__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_8catalyst_9_protocol_SidView = {
  PyVarObject_HEAD_INIT(0, 0)
  "catalyst._protocol.SidView", /*tp_name*/
  sizeof(struct __pyx_obj_8catalyst_9_protocol_SidView), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8catalyst_9_protocol_SidView, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_SidView, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_SidView, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_8catalyst_9_protocol_SidView, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8catalyst_9_protocol_SidView, /*tp_traverse*/
  __pyx_tp_clear_8catalyst_9_protocol_SidView, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8catalyst_9_protocol_SidView, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8catalyst_9_protocol_SidView, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8catalyst_9_protocol_7SidView_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8catalyst_9_protocol_SidView, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct____call__[8];
static int __pyx_freecount_8catalyst_9_protocol___pyx_scope_struct____call__ = 0;

static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct____call__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct____call__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__)))) {
    o = (PyObject*)__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct____call__[--__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct____call__];
    memset(o, 0, sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct____call__(PyObject *o) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_func);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct____call__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__)))) {
    __pyx_freelist_8catalyst_9_protocol___pyx_scope_struct____call__[__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct____call__++] = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct____call__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *)o;
  if (p->__pyx_v_func) {
    e = (*v)(p->__pyx_v_func, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct____call__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__ *)o;
  tmp = ((PyObject*)p->__pyx_v_func);
  p->__pyx_v_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8catalyst_9_protocol_check_parameters *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8catalyst_9_protocol___pyx_scope_struct____call__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "catalyst._protocol.__pyx_scope_struct____call__", /*tp_name*/
  sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct____call__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct____call__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct____call__, /*tp_traverse*/
  __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct____call__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct____call__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes[8];
static int __pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes = 0;

static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes)))) {
    o = (PyObject*)__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes[--__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes];
    memset(o, 0, sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes(PyObject *o) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_bar_data);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes)))) {
    __pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes[__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes++] = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *)o;
  if (p->__pyx_v_bar_data) {
    e = (*v)(p->__pyx_v_bar_data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes *)o;
  tmp = ((PyObject*)p->__pyx_v_bar_data);
  p->__pyx_v_bar_data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes = {
  PyVarObject_HEAD_INIT(0, 0)
  "catalyst._protocol.__pyx_scope_struct_1_handle_non_market_minutes", /*tp_name*/
  sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes, /*tp_traverse*/
  __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_2___iter__[8];
static int __pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_2___iter__ = 0;

static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_2___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_2___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__)))) {
    o = (PyObject*)__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_2___iter__[--__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_2___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct_2___iter__(PyObject *o) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_asset);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_2___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__)))) {
    __pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_2___iter__[__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_2___iter__++] = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct_2___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *)o;
  if (p->__pyx_v_asset) {
    e = (*v)(p->__pyx_v_asset, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct_2___iter__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__ *)o;
  tmp = ((PyObject*)p->__pyx_v_asset);
  p->__pyx_v_asset = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8catalyst_9_protocol_BarData *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_0);
  p->__pyx_t_0 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8catalyst_9_protocol___pyx_scope_struct_2___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "catalyst._protocol.__pyx_scope_struct_2___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_2___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct_2___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct_2___iter__, /*tp_traverse*/
  __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct_2___iter__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_2___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_3_iteritems[8];
static int __pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_3_iteritems = 0;

static PyObject *__pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_3_iteritems(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_3_iteritems > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems)))) {
    o = (PyObject*)__pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_3_iteritems[--__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_3_iteritems];
    memset(o, 0, sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct_3_iteritems(PyObject *o) {
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_asset);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_3_iteritems < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems)))) {
    __pyx_freelist_8catalyst_9_protocol___pyx_scope_struct_3_iteritems[__pyx_freecount_8catalyst_9_protocol___pyx_scope_struct_3_iteritems++] = ((struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct_3_iteritems(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *)o;
  if (p->__pyx_v_asset) {
    e = (*v)(p->__pyx_v_asset, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct_3_iteritems(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *p = (struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems *)o;
  tmp = ((PyObject*)p->__pyx_v_asset);
  p->__pyx_v_asset = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8catalyst_9_protocol_BarData *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_0);
  p->__pyx_t_0 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8catalyst_9_protocol___pyx_scope_struct_3_iteritems = {
  PyVarObject_HEAD_INIT(0, 0)
  "catalyst._protocol.__pyx_scope_struct_3_iteritems", /*tp_name*/
  sizeof(struct __pyx_obj_8catalyst_9_protocol___pyx_scope_struct_3_iteritems), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8catalyst_9_protocol___pyx_scope_struct_3_iteritems, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8catalyst_9_protocol___pyx_scope_struct_3_iteritems, /*tp_traverse*/
  __pyx_tp_clear_8catalyst_9_protocol___pyx_scope_struct_3_iteritems, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8catalyst_9_protocol___pyx_scope_struct_3_iteritems, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_protocol",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_Asset, __pyx_k_Asset, sizeof(__pyx_k_Asset), 0, 0, 1, 1},
  {&__pyx_n_s_AssetConvertible, __pyx_k_AssetConvertible, sizeof(__pyx_k_AssetConvertible), 0, 0, 1, 1},
  {&__pyx_n_s_BarData___iter, __pyx_k_BarData___iter, sizeof(__pyx_k_BarData___iter), 0, 0, 1, 1},
  {&__pyx_n_s_BarData_iteritems, __pyx_k_BarData_iteritems, sizeof(__pyx_k_BarData_iteritems), 0, 0, 1, 1},
  {&__pyx_kp_s_Checking_whether_an_asset_is_in, __pyx_k_Checking_whether_an_asset_is_in, sizeof(__pyx_k_Checking_whether_an_asset_is_in), 0, 0, 1, 0},
  {&__pyx_n_s_ContinuousFuture, __pyx_k_ContinuousFuture, sizeof(__pyx_k_ContinuousFuture), 0, 0, 1, 1},
  {&__pyx_n_s_DataFrame, __pyx_k_DataFrame, sizeof(__pyx_k_DataFrame), 0, 0, 1, 1},
  {&__pyx_kp_s_Expected_s_argument_to_be_of_typ, __pyx_k_Expected_s_argument_to_be_of_typ, sizeof(__pyx_k_Expected_s_argument_to_be_of_typ), 0, 0, 1, 0},
  {&__pyx_n_s_Future, __pyx_k_Future, sizeof(__pyx_k_Future), 0, 0, 1, 1},
  {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
  {&__pyx_kp_s_Iterating_over_the_assets_in_dat, __pyx_k_Iterating_over_the_assets_in_dat, sizeof(__pyx_k_Iterating_over_the_assets_in_dat), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_NaT, __pyx_k_NaT, sizeof(__pyx_k_NaT), 0, 0, 1, 1},
  {&__pyx_n_s_PY2, __pyx_k_PY2, sizeof(__pyx_k_PY2), 0, 0, 1, 1},
  {&__pyx_n_s_Panel, __pyx_k_Panel, sizeof(__pyx_k_Panel), 0, 0, 1, 1},
  {&__pyx_n_s_PricingDataAssociable, __pyx_k_PricingDataAssociable, sizeof(__pyx_k_PricingDataAssociable), 0, 0, 1, 1},
  {&__pyx_n_s_Series, __pyx_k_Series, sizeof(__pyx_k_Series), 0, 0, 1, 1},
  {&__pyx_kp_s_The_mavg_method_is_deprecated, __pyx_k_The_mavg_method_is_deprecated, sizeof(__pyx_k_The_mavg_method_is_deprecated), 0, 0, 1, 0},
  {&__pyx_kp_s_The_returns_method_is_deprecated, __pyx_k_The_returns_method_is_deprecated, sizeof(__pyx_k_The_returns_method_is_deprecated), 0, 0, 1, 0},
  {&__pyx_kp_s_The_stddev_method_is_deprecated, __pyx_k_The_stddev_method_is_deprecated, sizeof(__pyx_k_The_stddev_method_is_deprecated), 0, 0, 1, 0},
  {&__pyx_kp_s_The_vwap_method_is_deprecated, __pyx_k_The_vwap_method_is_deprecated, sizeof(__pyx_k_The_vwap_method_is_deprecated), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Users_fredfortier_Code_enigma_c, __pyx_k_Users_fredfortier_Code_enigma_c, sizeof(__pyx_k_Users_fredfortier_Code_enigma_c), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_ZiplineDeprecationWarning, __pyx_k_ZiplineDeprecationWarning, sizeof(__pyx_k_ZiplineDeprecationWarning), 0, 0, 1, 1},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_n_s_arg, __pyx_k_arg, sizeof(__pyx_k_arg), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_assert_keywords_and_call, __pyx_k_assert_keywords_and_call, sizeof(__pyx_k_assert_keywords_and_call), 0, 0, 1, 1},
  {&__pyx_n_s_asset, __pyx_k_asset, sizeof(__pyx_k_asset), 0, 0, 1, 1},
  {&__pyx_n_s_asset_finder, __pyx_k_asset_finder, sizeof(__pyx_k_asset_finder), 0, 0, 1, 1},
  {&__pyx_n_s_assets, __pyx_k_assets, sizeof(__pyx_k_assets), 0, 0, 1, 1},
  {&__pyx_n_s_auto_close_date, __pyx_k_auto_close_date, sizeof(__pyx_k_auto_close_date), 0, 0, 1, 1},
  {&__pyx_n_s_bar_count, __pyx_k_bar_count, sizeof(__pyx_k_bar_count), 0, 0, 1, 1},
  {&__pyx_n_s_bar_data, __pyx_k_bar_data, sizeof(__pyx_k_bar_data), 0, 0, 1, 1},
  {&__pyx_n_s_bars, __pyx_k_bars, sizeof(__pyx_k_bars), 0, 0, 1, 1},
  {&__pyx_n_s_call___locals_assert_keywords, __pyx_k_call___locals_assert_keywords, sizeof(__pyx_k_call___locals_assert_keywords), 0, 0, 1, 1},
  {&__pyx_n_s_can_trade, __pyx_k_can_trade, sizeof(__pyx_k_can_trade), 0, 0, 1, 1},
  {&__pyx_n_s_catalyst__protocol, __pyx_k_catalyst__protocol, sizeof(__pyx_k_catalyst__protocol), 0, 0, 1, 1},
  {&__pyx_n_s_catalyst_assets, __pyx_k_catalyst_assets, sizeof(__pyx_k_catalyst_assets), 0, 0, 1, 1},
  {&__pyx_n_s_catalyst_assets_continuous_futur, __pyx_k_catalyst_assets_continuous_futur, sizeof(__pyx_k_catalyst_assets_continuous_futur), 0, 0, 1, 1},
  {&__pyx_n_s_catalyst_catalyst_warnings, __pyx_k_catalyst_catalyst_warnings, sizeof(__pyx_k_catalyst_catalyst_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_category, __pyx_k_category, sizeof(__pyx_k_category), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_close_price, __pyx_k_close_price, sizeof(__pyx_k_close_price), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_contains, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {&__pyx_n_s_contextlib, __pyx_k_contextlib, sizeof(__pyx_k_contextlib), 0, 0, 1, 1},
  {&__pyx_n_s_contextmanager, __pyx_k_contextmanager, sizeof(__pyx_k_contextmanager), 0, 0, 1, 1},
  {&__pyx_n_s_continuous_future, __pyx_k_continuous_future, sizeof(__pyx_k_continuous_future), 0, 0, 1, 1},
  {&__pyx_n_s_current, __pyx_k_current, sizeof(__pyx_k_current), 0, 0, 1, 1},
  {&__pyx_n_s_current_chain, __pyx_k_current_chain, sizeof(__pyx_k_current_chain), 0, 0, 1, 1},
  {&__pyx_n_s_current_session, __pyx_k_current_session, sizeof(__pyx_k_current_session), 0, 0, 1, 1},
  {&__pyx_n_s_daily, __pyx_k_daily, sizeof(__pyx_k_daily), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_frequency, __pyx_k_data_frequency, sizeof(__pyx_k_data_frequency), 0, 0, 1, 1},
  {&__pyx_n_s_data_portal, __pyx_k_data_portal, sizeof(__pyx_k_data_portal), 0, 0, 1, 1},
  {&__pyx_kp_s_data_sid_N_is_deprecated_Use_da, __pyx_k_data_sid_N_is_deprecated_Use_da, sizeof(__pyx_k_data_sid_N_is_deprecated_Use_da), 0, 0, 1, 0},
  {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
  {&__pyx_n_s_dec, __pyx_k_dec, sizeof(__pyx_k_dec), 0, 0, 1, 1},
  {&__pyx_n_s_direction, __pyx_k_direction, sizeof(__pyx_k_direction), 0, 0, 1, 1},
  {&__pyx_n_s_dt, __pyx_k_dt, sizeof(__pyx_k_dt), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_expected_type, __pyx_k_expected_type, sizeof(__pyx_k_expected_type), 0, 0, 1, 1},
  {&__pyx_n_s_expected_type_name, __pyx_k_expected_type_name, sizeof(__pyx_k_expected_type_name), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_field, __pyx_k_field, sizeof(__pyx_k_field), 0, 0, 1, 1},
  {&__pyx_n_s_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 0, 1, 1},
  {&__pyx_n_s_frequency, __pyx_k_frequency, sizeof(__pyx_k_frequency), 0, 0, 1, 1},
  {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
  {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
  {&__pyx_n_s_get_adjusted_value, __pyx_k_get_adjusted_value, sizeof(__pyx_k_get_adjusted_value), 0, 0, 1, 1},
  {&__pyx_n_s_get_adjustments, __pyx_k_get_adjustments, sizeof(__pyx_k_get_adjustments), 0, 0, 1, 1},
  {&__pyx_n_s_get_current_future_chain, __pyx_k_get_current_future_chain, sizeof(__pyx_k_get_current_future_chain), 0, 0, 1, 1},
  {&__pyx_n_s_get_fetcher_assets, __pyx_k_get_fetcher_assets, sizeof(__pyx_k_get_fetcher_assets), 0, 0, 1, 1},
  {&__pyx_n_s_get_history_window, __pyx_k_get_history_window, sizeof(__pyx_k_get_history_window), 0, 0, 1, 1},
  {&__pyx_n_s_get_last_traded_dt, __pyx_k_get_last_traded_dt, sizeof(__pyx_k_get_last_traded_dt), 0, 0, 1, 1},
  {&__pyx_n_s_get_simple_transform, __pyx_k_get_simple_transform, sizeof(__pyx_k_get_simple_transform), 0, 0, 1, 1},
  {&__pyx_n_s_get_spot_value, __pyx_k_get_spot_value, sizeof(__pyx_k_get_spot_value), 0, 0, 1, 1},
  {&__pyx_n_s_getattr, __pyx_k_getattr, sizeof(__pyx_k_getattr), 0, 0, 1, 1},
  {&__pyx_n_s_handle_non_market_minutes, __pyx_k_handle_non_market_minutes, sizeof(__pyx_k_handle_non_market_minutes), 0, 0, 1, 1},
  {&__pyx_n_s_handle_non_market_minutes_2, __pyx_k_handle_non_market_minutes_2, sizeof(__pyx_k_handle_non_market_minutes_2), 0, 0, 1, 1},
  {&__pyx_n_s_history, __pyx_k_history, sizeof(__pyx_k_history), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_is_alive_for_session, __pyx_k_is_alive_for_session, sizeof(__pyx_k_is_alive_for_session), 0, 0, 1, 1},
  {&__pyx_n_s_is_exchange_open, __pyx_k_is_exchange_open, sizeof(__pyx_k_is_exchange_open), 0, 0, 1, 1},
  {&__pyx_n_s_is_open_on_minute, __pyx_k_is_open_on_minute, sizeof(__pyx_k_is_open_on_minute), 0, 0, 1, 1},
  {&__pyx_n_s_is_restricted, __pyx_k_is_restricted, sizeof(__pyx_k_is_restricted), 0, 0, 1, 1},
  {&__pyx_n_s_is_stale, __pyx_k_is_stale, sizeof(__pyx_k_is_stale), 0, 0, 1, 1},
  {&__pyx_n_s_isnan, __pyx_k_isnan, sizeof(__pyx_k_isnan), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_keyword, __pyx_k_keyword, sizeof(__pyx_k_keyword), 0, 0, 1, 1},
  {&__pyx_n_s_keyword_names, __pyx_k_keyword_names, sizeof(__pyx_k_keyword_names), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_last_traded, __pyx_k_last_traded, sizeof(__pyx_k_last_traded), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mavg, __pyx_k_mavg, sizeof(__pyx_k_mavg), 0, 0, 1, 1},
  {&__pyx_n_s_minute_to_session_label, __pyx_k_minute_to_session_label, sizeof(__pyx_k_minute_to_session_label), 0, 0, 1, 1},
  {&__pyx_n_s_minutes_for_session, __pyx_k_minutes_for_session, sizeof(__pyx_k_minutes_for_session), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_next, __pyx_k_next, sizeof(__pyx_k_next), 0, 0, 1, 1},
  {&__pyx_n_s_next_open, __pyx_k_next_open, sizeof(__pyx_k_next_open), 0, 0, 1, 1},
  {&__pyx_n_s_no_wraps_py2, __pyx_k_no_wraps_py2, sizeof(__pyx_k_no_wraps_py2), 0, 0, 1, 1},
  {&__pyx_n_s_no_wraps_py2_locals_dec, __pyx_k_no_wraps_py2_locals_dec, sizeof(__pyx_k_no_wraps_py2_locals_dec), 0, 0, 1, 1},
  {&__pyx_n_s_normalize_date, __pyx_k_normalize_date, sizeof(__pyx_k_normalize_date), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_open_price, __pyx_k_open_price, sizeof(__pyx_k_open_price), 0, 0, 1, 1},
  {&__pyx_kp_s_or_iterable_of_type, __pyx_k_or_iterable_of_type, sizeof(__pyx_k_or_iterable_of_type), 0, 0, 1, 0},
  {&__pyx_n_s_pandas, __pyx_k_pandas, sizeof(__pyx_k_pandas), 0, 0, 1, 1},
  {&__pyx_n_s_pandas_tslib, __pyx_k_pandas_tslib, sizeof(__pyx_k_pandas_tslib), 0, 0, 1, 1},
  {&__pyx_n_s_pd, __pyx_k_pd, sizeof(__pyx_k_pd), 0, 0, 1, 1},
  {&__pyx_n_s_previous_minute, __pyx_k_previous_minute, sizeof(__pyx_k_previous_minute), 0, 0, 1, 1},
  {&__pyx_n_s_price, __pyx_k_price, sizeof(__pyx_k_price), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_restrictions, __pyx_k_restrictions, sizeof(__pyx_k_restrictions), 0, 0, 1, 1},
  {&__pyx_n_s_retrieve_asset, __pyx_k_retrieve_asset, sizeof(__pyx_k_retrieve_asset), 0, 0, 1, 1},
  {&__pyx_n_s_returns, __pyx_k_returns, sizeof(__pyx_k_returns), 0, 0, 1, 1},
  {&__pyx_kp_s_s_got_an_unexpected_keyword_arg, __pyx_k_s_got_an_unexpected_keyword_arg, sizeof(__pyx_k_s_got_an_unexpected_keyword_arg), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_sid, __pyx_k_sid, sizeof(__pyx_k_sid), 0, 0, 1, 1},
  {&__pyx_n_s_simulation_dt_func, __pyx_k_simulation_dt_func, sizeof(__pyx_k_simulation_dt_func), 0, 0, 1, 1},
  {&__pyx_n_s_six, __pyx_k_six, sizeof(__pyx_k_six), 0, 0, 1, 1},
  {&__pyx_n_s_stacklevel, __pyx_k_stacklevel, sizeof(__pyx_k_stacklevel), 0, 0, 1, 1},
  {&__pyx_n_s_stddev, __pyx_k_stddev, sizeof(__pyx_k_stddev), 0, 0, 1, 1},
  {&__pyx_n_s_string_types, __pyx_k_string_types, sizeof(__pyx_k_string_types), 0, 0, 1, 1},
  {&__pyx_n_s_symbol, __pyx_k_symbol, sizeof(__pyx_k_symbol), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_trading_calendar, __pyx_k_trading_calendar, sizeof(__pyx_k_trading_calendar), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {&__pyx_n_s_universe_func, __pyx_k_universe_func, sizeof(__pyx_k_universe_func), 0, 0, 1, 1},
  {&__pyx_n_s_volume, __pyx_k_volume, sizeof(__pyx_k_volume), 0, 0, 1, 1},
  {&__pyx_n_s_vwap, __pyx_k_vwap, sizeof(__pyx_k_vwap), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_wraps, __pyx_k_wraps, sizeof(__pyx_k_wraps), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 68, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 82, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 209, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 212, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "catalyst/_protocol.pyx":43
 * if PY2:
 *     def no_wraps_py2(f):
 *         def dec(g):             # <<<<<<<<<<<<<<
 *             return g
 *         return dec
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_n_s_g); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple_, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_fredfortier_Code_enigma_c, __pyx_n_s_dec, 43, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(0, 43, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":72
 *     def __call__(self, func):
 *         @no_wraps_py2(func)
 *         def assert_keywords_and_call(*args, **kwargs):             # <<<<<<<<<<<<<<
 *             cdef short i
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(9, __pyx_n_s_args, __pyx_n_s_kwargs, __pyx_n_s_i, __pyx_n_s_field, __pyx_n_s_arg, __pyx_n_s_expected_type, __pyx_n_s_expected_type_name, __pyx_n_s_keyword, __pyx_n_s_type); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __pyx_codeobj__6 = (PyObject*)__Pyx_PyCode_New(0, 0, 9, 0, CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__5, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_fredfortier_Code_enigma_c, __pyx_n_s_assert_keywords_and_call, 72, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__6)) __PYX_ERR(0, 72, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":42
 * # with python3 so it is not a big deal.
 * if PY2:
 *     def no_wraps_py2(f):             # <<<<<<<<<<<<<<
 *         def dec(g):
 *             return g
 */
  __pyx_tuple__7 = PyTuple_Pack(3, __pyx_n_s_f, __pyx_n_s_dec, __pyx_n_s_dec); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_codeobj__8 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__7, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_fredfortier_Code_enigma_c, __pyx_n_s_no_wraps_py2, 42, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__8)) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":126
 * 
 * @contextmanager
 * def handle_non_market_minutes(bar_data):             # <<<<<<<<<<<<<<
 *     try:
 *         bar_data._handle_non_market_minutes = True
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_n_s_bar_data); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__9, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_fredfortier_Code_enigma_c, __pyx_n_s_handle_non_market_minutes, 126, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(0, 126, __pyx_L1_error)

  /* "catalyst/_protocol.pyx":252
 *         return dt
 * 
 *     @check_parameters(('assets', 'fields'),             # <<<<<<<<<<<<<<
 *                       ((Asset, ContinuousFuture) + string_types, string_types))
 *     def current(self, assets, fields):
 */
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_n_s_assets, __pyx_n_s_fields); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "catalyst/_protocol.pyx":429
 *                 return pd.DataFrame(data)
 * 
 *     @check_parameters(('continuous_future',),             # <<<<<<<<<<<<<<
 *                       (ContinuousFuture,))
 *     def current_chain(self, continuous_future):
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_n_s_continuous_future); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "catalyst/_protocol.pyx":436
 *             self.simulation_dt_func())
 * 
 *     @check_parameters(('assets',), (Asset,))             # <<<<<<<<<<<<<<
 *     def can_trade(self, assets):
 *         """
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_n_s_assets); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "catalyst/_protocol.pyx":526
 *         )
 * 
 *     @check_parameters(('assets',), (Asset,))             # <<<<<<<<<<<<<<
 *     def is_stale(self, assets):
 *         """
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_n_s_assets); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "catalyst/_protocol.pyx":588
 *             return not (last_traded_dt is pd.NaT)
 * 
 *     @check_parameters(('assets', 'fields', 'bar_count',             # <<<<<<<<<<<<<<
 *                        'frequency'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types,
 */
  __pyx_tuple__15 = PyTuple_Pack(4, __pyx_n_s_assets, __pyx_n_s_fields, __pyx_n_s_bar_count, __pyx_n_s_frequency); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_protocol(void); /*proto*/
PyMODINIT_FUNC init_protocol(void)
#else
PyMODINIT_FUNC PyInit__protocol(void); /*proto*/
PyMODINIT_FUNC PyInit__protocol(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__protocol(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_protocol", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_catalyst___protocol) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "catalyst._protocol")) {
      if (unlikely(PyDict_SetItemString(modules, "catalyst._protocol", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_8catalyst_9_protocol_check_parameters) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_type_8catalyst_9_protocol_check_parameters.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "check_parameters", (PyObject *)&__pyx_type_8catalyst_9_protocol_check_parameters) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_ptype_8catalyst_9_protocol_check_parameters = &__pyx_type_8catalyst_9_protocol_check_parameters;
  __pyx_vtabptr_8catalyst_9_protocol_BarData = &__pyx_vtable_8catalyst_9_protocol_BarData;
  __pyx_vtable_8catalyst_9_protocol_BarData._get_equity_price_view = (PyObject *(*)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *))__pyx_f_8catalyst_9_protocol_7BarData__get_equity_price_view;
  __pyx_vtable_8catalyst_9_protocol_BarData._create_sid_view = (PyObject *(*)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *))__pyx_f_8catalyst_9_protocol_7BarData__create_sid_view;
  __pyx_vtable_8catalyst_9_protocol_BarData._get_current_minute = (PyObject *(*)(struct __pyx_obj_8catalyst_9_protocol_BarData *))__pyx_f_8catalyst_9_protocol_7BarData__get_current_minute;
  __pyx_vtable_8catalyst_9_protocol_BarData._can_trade_for_asset = (PyBoolObject *(*)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_8catalyst_9_protocol_7BarData__can_trade_for_asset;
  __pyx_vtable_8catalyst_9_protocol_BarData._is_stale_for_asset = (PyBoolObject *(*)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_8catalyst_9_protocol_7BarData__is_stale_for_asset;
  __pyx_vtable_8catalyst_9_protocol_BarData._calculate_universe = (PyObject *(*)(struct __pyx_obj_8catalyst_9_protocol_BarData *))__pyx_f_8catalyst_9_protocol_7BarData__calculate_universe;
  __pyx_vtable_8catalyst_9_protocol_BarData._warn_deprecated = (PyObject *(*)(struct __pyx_obj_8catalyst_9_protocol_BarData *, PyObject *))__pyx_f_8catalyst_9_protocol_7BarData__warn_deprecated;
  if (PyType_Ready(&__pyx_type_8catalyst_9_protocol_BarData) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_type_8catalyst_9_protocol_BarData.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_8catalyst_9_protocol_BarData.tp_dict, __pyx_vtabptr_8catalyst_9_protocol_BarData) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "BarData", (PyObject *)&__pyx_type_8catalyst_9_protocol_BarData) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_ptype_8catalyst_9_protocol_BarData = &__pyx_type_8catalyst_9_protocol_BarData;
  __pyx_vtabptr_8catalyst_9_protocol_SidView = &__pyx_vtable_8catalyst_9_protocol_SidView;
  __pyx_vtable_8catalyst_9_protocol_SidView._warn_deprecated = (PyObject *(*)(struct __pyx_obj_8catalyst_9_protocol_SidView *, PyObject *))__pyx_f_8catalyst_9_protocol_7SidView__warn_deprecated;
  if (PyType_Ready(&__pyx_type_8catalyst_9_protocol_SidView) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __pyx_type_8catalyst_9_protocol_SidView.tp_print = 0;
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_8catalyst_9_protocol_SidView, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 826, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_8catalyst_9_protocol_7SidView___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_8catalyst_9_protocol_7SidView___init__.doc = __pyx_doc_8catalyst_9_protocol_7SidView___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_8catalyst_9_protocol_7SidView___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_8catalyst_9_protocol_SidView.tp_dict, __pyx_vtabptr_8catalyst_9_protocol_SidView) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "SidView", (PyObject *)&__pyx_type_8catalyst_9_protocol_SidView) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __pyx_ptype_8catalyst_9_protocol_SidView = &__pyx_type_8catalyst_9_protocol_SidView;
  if (PyType_Ready(&__pyx_type_8catalyst_9_protocol___pyx_scope_struct____call__) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_type_8catalyst_9_protocol___pyx_scope_struct____call__.tp_print = 0;
  __pyx_ptype_8catalyst_9_protocol___pyx_scope_struct____call__ = &__pyx_type_8catalyst_9_protocol___pyx_scope_struct____call__;
  if (PyType_Ready(&__pyx_type_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __pyx_type_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes.tp_print = 0;
  __pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes = &__pyx_type_8catalyst_9_protocol___pyx_scope_struct_1_handle_non_market_minutes;
  if (PyType_Ready(&__pyx_type_8catalyst_9_protocol___pyx_scope_struct_2___iter__) < 0) __PYX_ERR(0, 778, __pyx_L1_error)
  __pyx_type_8catalyst_9_protocol___pyx_scope_struct_2___iter__.tp_print = 0;
  __pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_2___iter__ = &__pyx_type_8catalyst_9_protocol___pyx_scope_struct_2___iter__;
  if (PyType_Ready(&__pyx_type_8catalyst_9_protocol___pyx_scope_struct_3_iteritems) < 0) __PYX_ERR(0, 795, __pyx_L1_error)
  __pyx_type_8catalyst_9_protocol___pyx_scope_struct_3_iteritems.tp_print = 0;
  __pyx_ptype_8catalyst_9_protocol___pyx_scope_struct_3_iteritems = &__pyx_type_8catalyst_9_protocol___pyx_scope_struct_3_iteritems;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(1, 9, __pyx_L1_error)
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) __PYX_ERR(2, 8, __pyx_L1_error)
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) __PYX_ERR(3, 15, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "catalyst/_protocol.pyx":15
 * # See the License for the specific language governing permissions and
 * # limitations under the License.
 * import warnings             # <<<<<<<<<<<<<<
 * from contextlib import contextmanager
 * from functools import wraps
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":16
 * # limitations under the License.
 * import warnings
 * from contextlib import contextmanager             # <<<<<<<<<<<<<<
 * from functools import wraps
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_contextmanager);
  __Pyx_GIVEREF(__pyx_n_s_contextmanager);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_contextmanager);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_contextlib, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_contextmanager); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_contextmanager, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":17
 * import warnings
 * from contextlib import contextmanager
 * from functools import wraps             # <<<<<<<<<<<<<<
 * 
 * from pandas.tslib import normalize_date
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_wraps);
  __Pyx_GIVEREF(__pyx_n_s_wraps);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_wraps);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_functools, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_wraps); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_wraps, __pyx_t_2) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":19
 * from functools import wraps
 * 
 * from pandas.tslib import normalize_date             # <<<<<<<<<<<<<<
 * import pandas as pd
 * import numpy as np
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_normalize_date);
  __Pyx_GIVEREF(__pyx_n_s_normalize_date);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_normalize_date);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pandas_tslib, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_normalize_date); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_normalize_date, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":20
 * 
 * from pandas.tslib import normalize_date
 * import pandas as pd             # <<<<<<<<<<<<<<
 * import numpy as np
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pandas, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pd, __pyx_t_2) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":21
 * from pandas.tslib import normalize_date
 * import pandas as pd
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * from six import iteritems, PY2, string_types
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":23
 * import numpy as np
 * 
 * from six import iteritems, PY2, string_types             # <<<<<<<<<<<<<<
 * from cpython cimport bool
 * from collections import Iterable
 */
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_iteritems);
  __Pyx_GIVEREF(__pyx_n_s_iteritems);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_iteritems);
  __Pyx_INCREF(__pyx_n_s_PY2);
  __Pyx_GIVEREF(__pyx_n_s_PY2);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_PY2);
  __Pyx_INCREF(__pyx_n_s_string_types);
  __Pyx_GIVEREF(__pyx_n_s_string_types);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_string_types);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_six, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_iteritems); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_iteritems, __pyx_t_2) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_PY2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PY2, __pyx_t_2) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_string_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_string_types, __pyx_t_2) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":25
 * from six import iteritems, PY2, string_types
 * from cpython cimport bool
 * from collections import Iterable             # <<<<<<<<<<<<<<
 * 
 * from catalyst.assets import (Asset,
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Iterable);
  __Pyx_GIVEREF(__pyx_n_s_Iterable);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Iterable);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Iterable, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":27
 * from collections import Iterable
 * 
 * from catalyst.assets import (Asset,             # <<<<<<<<<<<<<<
 *                             AssetConvertible,
 *                             PricingDataAssociable,
 */
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Asset);
  __Pyx_GIVEREF(__pyx_n_s_Asset);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Asset);
  __Pyx_INCREF(__pyx_n_s_AssetConvertible);
  __Pyx_GIVEREF(__pyx_n_s_AssetConvertible);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_AssetConvertible);
  __Pyx_INCREF(__pyx_n_s_PricingDataAssociable);
  __Pyx_GIVEREF(__pyx_n_s_PricingDataAssociable);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_PricingDataAssociable);
  __Pyx_INCREF(__pyx_n_s_Future);
  __Pyx_GIVEREF(__pyx_n_s_Future);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_Future);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_catalyst_assets, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Asset, __pyx_t_2) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_AssetConvertible); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AssetConvertible, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_PricingDataAssociable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PricingDataAssociable, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Future); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Future, __pyx_t_2) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":31
 *                             PricingDataAssociable,
 *                             Future)
 * from catalyst.assets.continuous_futures import ContinuousFuture             # <<<<<<<<<<<<<<
 * from catalyst.catalyst_warnings import ZiplineDeprecationWarning
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ContinuousFuture);
  __Pyx_GIVEREF(__pyx_n_s_ContinuousFuture);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ContinuousFuture);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_catalyst_assets_continuous_futur, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ContinuousFuture); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ContinuousFuture, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":32
 *                             Future)
 * from catalyst.assets.continuous_futures import ContinuousFuture
 * from catalyst.catalyst_warnings import ZiplineDeprecationWarning             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ZiplineDeprecationWarning);
  __Pyx_GIVEREF(__pyx_n_s_ZiplineDeprecationWarning);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ZiplineDeprecationWarning);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_catalyst_catalyst_warnings, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ZiplineDeprecationWarning); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ZiplineDeprecationWarning, __pyx_t_2) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":41
 * # Wraps doesn't work for method objects in python2. Docs should be generated
 * # with python3 so it is not a big deal.
 * if PY2:             # <<<<<<<<<<<<<<
 *     def no_wraps_py2(f):
 *         def dec(g):
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_PY2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "catalyst/_protocol.pyx":42
 * # with python3 so it is not a big deal.
 * if PY2:
 *     def no_wraps_py2(f):             # <<<<<<<<<<<<<<
 *         def dec(g):
 *             return g
 */
    __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8catalyst_9_protocol_1no_wraps_py2, NULL, __pyx_n_s_catalyst__protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_no_wraps_py2, __pyx_t_1) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "catalyst/_protocol.pyx":41
 * # Wraps doesn't work for method objects in python2. Docs should be generated
 * # with python3 so it is not a big deal.
 * if PY2:             # <<<<<<<<<<<<<<
 *     def no_wraps_py2(f):
 *         def dec(g):
 */
    goto __pyx_L2;
  }

  /* "catalyst/_protocol.pyx":47
 *         return dec
 * else:
 *     no_wraps_py2 = wraps             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_wraps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_no_wraps_py2, __pyx_t_1) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L2:;

  /* "catalyst/_protocol.pyx":125
 * 
 * 
 * @contextmanager             # <<<<<<<<<<<<<<
 * def handle_non_market_minutes(bar_data):
 *     try:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_contextmanager); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":126
 * 
 * @contextmanager
 * def handle_non_market_minutes(bar_data):             # <<<<<<<<<<<<<<
 *     try:
 *         bar_data._handle_non_market_minutes = True
 */
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_8catalyst_9_protocol_3handle_non_market_minutes, NULL, __pyx_n_s_catalyst__protocol); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_handle_non_market_minutes, __pyx_t_1) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "catalyst/_protocol.pyx":253
 * 
 *     @check_parameters(('assets', 'fields'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types))             # <<<<<<<<<<<<<<
 *     def current(self, assets, fields):
 *         """
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_Asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ContinuousFuture); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_string_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = PyNumber_Add(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_string_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;

  /* "catalyst/_protocol.pyx":252
 *         return dt
 * 
 *     @check_parameters(('assets', 'fields'),             # <<<<<<<<<<<<<<
 *                       ((Asset, ContinuousFuture) + string_types, string_types))
 *     def current(self, assets, fields):
 */
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8catalyst_9_protocol_check_parameters), __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "catalyst/_protocol.pyx":254
 *     @check_parameters(('assets', 'fields'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types))
 *     def current(self, assets, fields):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the current value of the given assets for the given fields
 */
  __pyx_t_6 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData, __pyx_n_s_current); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData->tp_dict, __pyx_n_s_current, __pyx_t_1) < 0) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_8catalyst_9_protocol_BarData);

  /* "catalyst/_protocol.pyx":430
 * 
 *     @check_parameters(('continuous_future',),
 *                       (ContinuousFuture,))             # <<<<<<<<<<<<<<
 *     def current_chain(self, continuous_future):
 *         return self.data_portal.get_current_future_chain(
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ContinuousFuture); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "catalyst/_protocol.pyx":429
 *                 return pd.DataFrame(data)
 * 
 *     @check_parameters(('continuous_future',),             # <<<<<<<<<<<<<<
 *                       (ContinuousFuture,))
 *     def current_chain(self, continuous_future):
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8catalyst_9_protocol_check_parameters), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "catalyst/_protocol.pyx":431
 *     @check_parameters(('continuous_future',),
 *                       (ContinuousFuture,))
 *     def current_chain(self, continuous_future):             # <<<<<<<<<<<<<<
 *         return self.data_portal.get_current_future_chain(
 *             continuous_future,
 */
  __pyx_t_4 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData, __pyx_n_s_current_chain); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData->tp_dict, __pyx_n_s_current_chain, __pyx_t_1) < 0) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_8catalyst_9_protocol_BarData);

  /* "catalyst/_protocol.pyx":436
 *             self.simulation_dt_func())
 * 
 *     @check_parameters(('assets',), (Asset,))             # <<<<<<<<<<<<<<
 *     def can_trade(self, assets):
 *         """
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8catalyst_9_protocol_check_parameters), __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "catalyst/_protocol.pyx":437
 * 
 *     @check_parameters(('assets',), (Asset,))
 *     def can_trade(self, assets):             # <<<<<<<<<<<<<<
 *         """
 *         For the given asset or iterable of assets, returns true if all of the
 */
  __pyx_t_5 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData, __pyx_n_s_can_trade); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData->tp_dict, __pyx_n_s_can_trade, __pyx_t_1) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_8catalyst_9_protocol_BarData);

  /* "catalyst/_protocol.pyx":526
 *         )
 * 
 *     @check_parameters(('assets',), (Asset,))             # <<<<<<<<<<<<<<
 *     def is_stale(self, assets):
 *         """
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_Asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8catalyst_9_protocol_check_parameters), __pyx_t_2, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":527
 * 
 *     @check_parameters(('assets',), (Asset,))
 *     def is_stale(self, assets):             # <<<<<<<<<<<<<<
 *         """
 *         For the given asset or iterable of assets, returns true if the asset
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData, __pyx_n_s_is_stale); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 526, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData->tp_dict, __pyx_n_s_is_stale, __pyx_t_1) < 0) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_8catalyst_9_protocol_BarData);

  /* "catalyst/_protocol.pyx":590
 *     @check_parameters(('assets', 'fields', 'bar_count',
 *                        'frequency'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types,             # <<<<<<<<<<<<<<
 *                        int,
 *                        string_types))
 */
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_Asset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ContinuousFuture); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __pyx_t_6 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_string_types); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyNumber_Add(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_string_types); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "catalyst/_protocol.pyx":592
 *                       ((Asset, ContinuousFuture) + string_types, string_types,
 *                        int,
 *                        string_types))             # <<<<<<<<<<<<<<
 *     def history(self, assets, fields, bar_count, frequency):
 *         """
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_string_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "catalyst/_protocol.pyx":590
 *     @check_parameters(('assets', 'fields', 'bar_count',
 *                        'frequency'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types,             # <<<<<<<<<<<<<<
 *                        int,
 *                        string_types))
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_5, 2, ((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_2);
  __pyx_t_6 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":588
 *             return not (last_traded_dt is pd.NaT)
 * 
 *     @check_parameters(('assets', 'fields', 'bar_count',             # <<<<<<<<<<<<<<
 *                        'frequency'),
 *                       ((Asset, ContinuousFuture) + string_types, string_types,
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8catalyst_9_protocol_check_parameters), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "catalyst/_protocol.pyx":593
 *                        int,
 *                        string_types))
 *     def history(self, assets, fields, bar_count, frequency):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a window of data for the given assets and fields.
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData, __pyx_n_s_history); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_8catalyst_9_protocol_BarData->tp_dict, __pyx_n_s_history, __pyx_t_1) < 0) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_8catalyst_9_protocol_BarData);

  /* "catalyst/_protocol.pyx":1
 * #             # <<<<<<<<<<<<<<
 * # Copyright 2016 Quantopian, Inc.
 * #
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init catalyst._protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init catalyst._protocol");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* ExtTypeTest */
  static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
  static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* RaiseArgTupleInvalid */
      static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
      static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
      static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCall */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* KeywordStringCheck */
      static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* IterFinish */
      static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyFunctionFastCall */
      #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyObjectCallMethO */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
        #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* PyObjectCallOneArg */
        #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallMethod0 */
          static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto bad;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *function = PyMethod_GET_FUNCTION(method);
            result = __Pyx_PyObject_CallOneArg(function, self);
            Py_DECREF(method);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
          static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
          static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
          static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
          static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
          static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
          static CYTHON_INLINE int __Pyx_unpack_tuple2(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                             int is_tuple, int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    if (!is_tuple && unlikely(!PyTuple_Check(tuple))) {
        iternextfunc iternext;
        iter = PyObject_GetIter(tuple);
        if (unlikely(!iter)) goto bad;
        if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
        iternext = Py_TYPE(iter)->tp_iternext;
        value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
        value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
        if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
        Py_DECREF(iter);
    } else {
        if (!has_known_size && unlikely(PyTuple_GET_SIZE(tuple) != 2)) {
            __Pyx_UnpackTupleError(tuple, 2);
            goto bad;
        }
#if CYTHON_COMPILING_IN_PYPY
        value1 = PySequence_ITEM(tuple, 0);
        if (unlikely(!value1)) goto bad;
        value2 = PySequence_ITEM(tuple, 1);
        if (unlikely(!value2)) goto bad;
#else
        value1 = PyTuple_GET_ITEM(tuple, 0);
        value2 = PyTuple_GET_ITEM(tuple, 1);
        Py_INCREF(value1);
        Py_INCREF(value2);
#endif
        if (decref_tuple) { Py_DECREF(tuple); }
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
          static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
#if !CYTHON_COMPILING_IN_PYPY
    if (is_dict) {
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#endif
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* None */
          static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyErrFetchRestore */
          #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
          #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* SliceTupleAndList */
            #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject** CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyList_New(0);
    dest = PyList_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyListObject*)src)->ob_item + start,
        ((PyListObject*)dest)->ob_item,
        length);
    return dest;
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyTuple_New(0);
    dest = PyTuple_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyTupleObject*)src)->ob_item + start,
        ((PyTupleObject*)dest)->ob_item,
        length);
    return dest;
}
#endif

/* GetItemInt */
            static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* StringJoin */
            #if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* BytesEquals */
            static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
            static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* GetException */
            #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
              #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* SaveResetException */
              #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
              #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* SetVTable */
              static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* Import */
              static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
              static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetNameInClass */
              static PyObject *__Pyx_GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result)
        result = __Pyx_GetModuleGlobalName(name);
    return result;
}

/* CodeObjectCache */
              static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
              #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
              #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
              static CYTHON_INLINE PyObject* __Pyx_PyInt_From_short(short value) {
    const short neg_one = (short) -1, const_zero = (short) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(short) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(short) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(short) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(short) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(short),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
              static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
              static CYTHON_INLINE short __Pyx_PyInt_As_short(PyObject *x) {
    const short neg_one = (short) -1, const_zero = (short) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(short) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(short, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (short) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (short) 0;
                case  1: __PYX_VERIFY_RETURN_INT(short, digit, digits[0])
                case 2:
                    if (8 * sizeof(short) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 2 * PyLong_SHIFT) {
                            return (short) (((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(short) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 3 * PyLong_SHIFT) {
                            return (short) (((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(short) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 4 * PyLong_SHIFT) {
                            return (short) (((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (short) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(short) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(short) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (short) 0;
                case -1: __PYX_VERIFY_RETURN_INT(short, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(short,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(short) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(short) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                            return (short) ((((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(short) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                            return (short) ((((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 4 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(short) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 4 * PyLong_SHIFT) {
                            return (short) ((((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(short) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(short) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            short val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (short) -1;
        }
    } else {
        short val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (short) -1;
        val = __Pyx_PyInt_As_short(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to short");
    return (short) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to short");
    return (short) -1;
}

/* CIntFromPy */
              static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
              static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* PyObjectCallMethod1 */
              static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto done;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyCFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            args = PyTuple_New(2);
            if (unlikely(!args)) goto done;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
done:
    Py_XDECREF(method);
    return result;
}

/* CoroutineBase */
              #include <structmember.h>
#include <frameobject.h>
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
#endif
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__pyx_CoroutineObject *self) {
    PyObject *exc_type = self->exc_type;
    PyObject *exc_value = self->exc_value;
    PyObject *exc_traceback = self->exc_traceback;
    self->exc_type = NULL;
    self->exc_value = NULL;
    self->exc_traceback = NULL;
    Py_XDECREF(exc_type);
    Py_XDECREF(exc_value);
    Py_XDECREF(exc_traceback);
}
static CYTHON_INLINE
int __Pyx_Coroutine_CheckRunning(__pyx_CoroutineObject *gen) {
    if (unlikely(gen->is_running)) {
        PyErr_SetString(PyExc_ValueError,
                        "generator already executing");
        return 1;
    }
    return 0;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value) {
    PyObject *retval;
    __Pyx_PyThreadState_declare
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            PyErr_SetString(PyExc_TypeError,
                            "can't send non-None value to a "
                            "just-started generator");
            return NULL;
        }
    }
    if (unlikely(self->resume_label == -1)) {
        PyErr_SetNone(PyExc_StopIteration);
        return NULL;
    }
    __Pyx_PyThreadState_assign
    if (value) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(__pyx_tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = __pyx_tstate->frame;
        }
#endif
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    self->is_running = 1;
    retval = self->body((PyObject *) self, value);
    self->is_running = 0;
    if (retval) {
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_CLEAR(f->f_back);
        }
#endif
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject *retval) {
    if (unlikely(!retval && !PyErr_Occurred())) {
        PyErr_SetNone(PyExc_StopIteration);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen_FetchStopIterationValue(&val);
    ret = __Pyx_Coroutine_SendEx(gen, val);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value);
    }
    return __Pyx_Coroutine_MethodReturn(retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL);
    if (retval) {
        Py_DECREF(retval);
        PyErr_SetString(PyExc_RuntimeError,
                        "generator ignored GeneratorExit");
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (!raised_exception
        || raised_exception == PyExc_StopIteration
        || raised_exception == PyExc_GeneratorExit
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_GeneratorExit)
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_StopIteration))
    {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *typ;
    PyObject *tb = NULL;
    PyObject *val = NULL;
    PyObject *yf = gen->yieldfrom;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit)) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
            goto throw_here;
        }
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            ret = PyObject_CallObject(meth, args);
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
}
static int __Pyx_Coroutine_traverse(PyObject *self, visitproc visit, void *arg) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    Py_VISIT(gen->exc_type);
    Py_VISIT(gen->exc_value);
    Py_VISIT(gen->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    Py_CLEAR(gen->exc_type);
    Py_CLEAR(gen->exc_value);
    Py_CLEAR(gen->exc_traceback);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label > 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *res;
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label <= 0)
        return ;
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
    res = __Pyx_Coroutine_Close(self);
    if (res == NULL)
        PyErr_WriteUnraisable(self);
    else
        Py_DECREF(res);
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (gen == NULL)
        return NULL;
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->exc_type = NULL;
    gen->exc_value = NULL;
    gen->exc_traceback = NULL;
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
                  static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
                  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (!abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_VERSION_HEX >= 0x03030000) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_VERSION_HEX >= 0x03030000) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if (0) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
                  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if PY_VERSION_HEX >= 0x030400a1
    __Pyx_Coroutine_del,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = PyObject_GenericGetAttr;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
                  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* ModuleImport */
                  #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
                  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* InitStrings */
                  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
